# The function of this script overlaps with that of memecv.py svmlight,
# and can be a replacement.
# Features and major differences with memecv.py:
# 1. Deep directory structure (arguments designed for flexibility)
# 2. Supporting multiple compartments for a sequence

import sys
from memecv import sum, fastaReadAll, readFeatures, svmLearn, svmClassify, \
                   svmLightReadOutput, writeResult1Fold

USAGE = 'feat2acc_fold.py FOLD [SEQDIR] [FEATDIR] [ACCDIR] [PARAM] [SEED]'

COMPS = ['YeastCyto', 'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastMito',
         'YeastNuclear', 'YeastPero', 'YeastPM', 'YeastSecreted']

def feat2acc(comps, fold, randSeed, seqDir, featDir, resDir, svmOpt):
    '''Train and test a SVM based on previously generated features,
    either from MEME or HMM.
    Overlap and replace memecv.svmLightReuse by supporting multi-compartment
    for a single sequence and deep directory structure.'''
    result1Filename     = '%sYeast%d_%d.result' % (resDir,  randSeed, fold)
    testFeatFilename    = '%sYeast%dte%d.dat' % (featDir, randSeed, fold)
    trainFeatFile0      = '%sYeast%dtr%d.dat'   % (featDir, randSeed, fold)
    trainFeatFilename   = '%sYeast%dmc%d.dat'   % (resDir, randSeed, fold)
    modelFilename       = '%sYeast%d_%d.svm'    % (resDir, randSeed, fold)
    predFilename        = '%sYeast%d_%d.pred'   % (resDir, randSeed, fold)

    trainFiles = ['%s%s%str%s.fa'%(seqDir,comp,randSeed,fold) for comp in comps]
    testFiles  = ['%s%s%ste%s.fa'%(seqDir,comp,randSeed,fold) for comp in comps]
    trainSeqId = []; testSeqId = []; trainNSeq = []; testNSeq = []
    for seqFilename in trainFiles:
        idSeqs = fastaReadAll(seqFilename)
        trainSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
        trainNSeq.append(len(idSeqs))
    for seqFilename in testFiles:
        idSeqs = fastaReadAll(seqFilename)
        testSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
        testNSeq.append(len(idSeqs))
    print '%s => %s ...' % (trainFeatFile0, trainFeatFilename)
    trainFeat = readFeatures(trainNSeq, trainFeatFile0)
    (corr, nTestSeq, precision, recall) = trainTest(trainFeat, trainSeqId, \
            testNSeq, testSeqId, trainFeatFilename, testFeatFilename, \
            modelFilename, predFilename, svmOpt)
    writeResult1Fold(result1Filename, float(corr) / nTestSeq, precision, \
                     recall, fold, comps, \
                     'multi-compartment flat svm (%s)\n' % svmOpt)

def trainTest(trainFeat, trainSeqId, testNSeq, testSeqId, trainFeatFilename, \
              testFeatFilename, modelFilename, predFilename, svmOpt = ''):
    '''Train and test the given (multiclass) feature set.
    Overlap and replace memecv.trainTest; support multi-compartment for
    a single sequence, and remove the "general mode".
    The training feature file include all features of all comp-seq,
    with +1/-1 labels depending on the compartment (1-vs-all).
    The testing feature file include all features of all comp-seqs and
    null labels. Similar to memecv.trainTest, but only the reuse mode
    (reusing previous SVM-light feature files) is supported. Training feature
    files are first read, and then re-written with the labels changed;
    testing feature files are re-used, including the labels,
    so no need to read features first.
    To support multi-compartment for a single sequence, trainSeqId and
    testSeqId are required.'''
    # Remove overlap between training and testing
    trainIds = dict([(id, 1) for ids in trainSeqId for id in ids]) #not to test
    testIds = dict([(id, 1) for ids in testSeqId for id in ids]) #assert
    nUniqueTest = sum([(id not in trainIds) for id in testIds.keys()]) #assert
    # Train and predict
    for trainComp in range(len(trainFeat)):
        writeFeatures(trainFeat, trainSeqId, trainComp, trainFeatFilename)
        svmLearn(trainFeatFilename, modelFilename, svmOpt)
        svmClassify(testFeatFilename, modelFilename, predFilename)
        pred = svmLightReadOutput(predFilename)
        if trainComp == 0:
            predMax = pred; predArgMax = [0] * len(pred)
        else:
            for i in range(len(pred)):
                if pred[i] > predMax[i]:
                    predMax[i] = pred[i]; predArgMax[i] = trainComp
    # Evaluate the prediction
    nComp = len(testNSeq)
    cumu = 0
    predIds = [{} for comp in range(nComp)]
    filteredPred = [{} for comp in range(nComp)]
    for comp in range(nComp):
        cumu1 = cumu + testNSeq[comp]
        for id, argMax in zip(testSeqId[comp], predArgMax[cumu:cumu1]):
            if not id in trainIds:
                filteredPred[comp][id] = argMax
                predIds[argMax][id] = 1
        cumu = cumu1
    nPred = [len(idDict) for idDict in predIds]; nTestSeq = sum(nPred)
    corr = 0; precision = []; recall = []
    for comp in range(nComp):
        tp = sum([(argMax==comp) for id, argMax in filteredPred[comp].items()])
        if nPred[comp] > 0: precision.append(float(tp)/nPred[comp])
        else:               precision.append(0.0)
        if len(filteredPred[comp]) > 0:
            recall.append(float(tp)/len(filteredPred[comp]))#cause low recall
        else: recall.append(0.0)
        corr = corr + tp
    if nUniqueTest != nTestSeq:
        raise Exception('Incorrect nTestSeq (repeated prediction mismatch)' \
                        ' nTestSeq=%d (should be %d), #trainseq=%d' \
                        % (nTestSeq, nUniqueTest, len(trainIds)))
    print '%d unique sequences processed (training: %d, testing: %d)' \
          % (len(trainIds) + nUniqueTest, len(trainIds), nUniqueTest)
    return (corr, nTestSeq, precision, recall)

def writeFeatures(compFeat, compSeqId, trainComp, featFilename):
    '''Write features in SVM light or multiclass format.
    Overlap and replace memecv.writeFeatures by supporting multi-compartment
    for a single sequence.'''
    fFeat = open(featFilename, 'w')
    for compart in range(len(compFeat)):
        for id, featVec in zip(compSeqId[compart], compFeat[compart]):
            if compart != trainComp:
                if id in compSeqId[trainComp]: continue
                fFeat.write('-1')
            else:
                fFeat.write('+1')
            for iFeat in range(len(featVec)):
                fFeat.write('\t%d:%f' % (iFeat + 1, featVec[iFeat]))
            fFeat.write('\n')
    fFeat.close()

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2: print USAGE; sys.exit(1)
    fold = int(sys.argv[1])
    seqDir = ''; featDir = ''; resDir = ''; svmOpt = ''; randSeed = 1000
    if len(sys.argv) > 2: seqDir = sys.argv[2]
    if len(sys.argv) > 3: featDir = sys.argv[3]
    if len(sys.argv) > 4: resDir = sys.argv[4]
    if len(sys.argv) > 5: svmOpt = sys.argv[5]
    if len(sys.argv) > 6: randSeed = int(sys.argv[6])
    feat2acc(COMPS, fold, randSeed, seqDir, featDir, resDir, svmOpt)
