#!/usr/bin/python
#Train one-vs-all SVMs using motif features, either from MEME, gen HMM,
#or disc HMM. One sequence with multi-compartments is supported.
#Assume one feature file, e.g. training data for one fold in cross validation.

import sys
from memecv import sum, fastaReadAll, readFeatures, writeResult1Fold
from feat2acc_fold import COMPS, writeFeatures
from feat2acc_tree import testSVM

USAGE = '''feat2acc_test RESULTFILE FEATFILE PREDFILE MODELPAT SEQPAT [SVMOPT]
SVMOPT is for remark in result-file.'''

def test1(resultFile, featFile, predFile, modelFiles, seqFiles, comps, svmOpt):
    '''Evaluate the classification of a trained one-vs-all SVM using new
    testing motif features.'''
    seqId = []
    for seqFilename in seqFiles:
        idSeqs = fastaReadAll(seqFilename)
        seqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
    predArgMax = testSVM(featFile, modelFiles, predFile)
    (accuracy, precision, recall, id2pred) = evaluate(predArgMax, seqId)
    writeResult1Fold(resultFile, accuracy, precision, recall, -1, \
                     comps, 'multi-compartment svm (%s)\n' % svmOpt)
    writeCompPred(id2pred, seqId, predFile, comps)

def evaluate(predArgMax, testSeqId):
    '''Evaludate prediction based on a flat topology.
    Modified from feat2acc_tree.evaluadate().'''
    nComp = len(testSeqId)    
    corr = [0] * nComp; nPred = [0] * nComp
    precision = [.0] * nComp; recall = [.0] * nComp
    id2pred = {}; uniqueIds = [None for comp in range(nComp)]
    cumu = 0
    id2i = {}
    for comp in range(nComp):
        cumu1 = cumu + len(testSeqId[comp])
        ids = {}
        for i in range(cumu, cumu1):
            id = testSeqId[comp][i-cumu]
            ids[id] = 1
            if id not in id2pred: id2pred[id] = predArgMax[i]; id2i[id] = i
            else: assert id2pred[id] == predArgMax[i]
            #else: print '%s[%d~%d]: %d==%d' % (id, id2i[id], i, id2pred[id], predArgMax[i])
        uniqueIds[comp] = ids.keys()
        cumu = cumu1
    for c1 in range(nComp):
        for id in uniqueIds[c1]:
            argMax = id2pred[id]
            if argMax==c1: corr[c1] = corr[c1] + 1
        nPred[c1] = len(dict([(id,1) for id,argMax in id2pred.items() \
                              if argMax==c1]))
    nTestSeq = sum(nPred)
    for c1 in range(nComp):
        tp = float(corr[c1])
        if nPred[c1] > 0: precision[c1] = tp / nPred[c1]
        else:             precision[c1] = 0.0
        if len(uniqueIds[c1]) > 0: recall[c1] = tp / len(uniqueIds[c1])
        else:                      recall[c1] = 0.0
    accuracy = float(sum(corr)) / nTestSeq
    assert nTestSeq == len(dict([(id, 1) for ids in testSeqId for id in ids]))
    print '%d unique sequences processed' % nTestSeq
    return (accuracy, precision, recall, id2pred)

def writeCompPred(id2pred, seqId, predFile, comps):
    for comp, ids in zip(comps, seqId):
        fp = open(predFile + '.' + comp, 'w')
        for id in ids: fp.write('%s\t%s\n' % (id, comps[id2pred[id]]))
        fp.close()

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2: print USAGE; sys.exit(1)
    resultFile      = sys.argv[1]
    featFile        = sys.argv[2]
    predFile        = sys.argv[3]
    modelPattern    = sys.argv[4]
    seqPattern      = sys.argv[5]
    svmOpt = ''
    if len(sys.argv) > 6: svmOpt = sys.argv[6]
    modelFiles = [modelPattern % c for c in COMPS]
    seqFiles = [seqPattern % c for c in COMPS]
    test1(resultFile, featFile, predFile, modelFiles, seqFiles, COMPS, svmOpt)
