#!/usr/bin/python
# Classification under a tree topology,
# features including internal and leaf motifs.
# This program is similar to memecvtree3.py, but it supports the deep directory
# structure and multi-compartments.

import sys, time
from memecv import sum, fastaReadAll, readFeatures, memeFeatures, \
                   svmLearn, svmClassify, svmLightReadOutput
from memecvtree import writeResult1Fold
from feat2acc_fold import writeFeatures

USAGE = \
'''Classification under a tree topology using internal and leaf motifs.
Evaluate SVM classifier given feature file (one fold in cross validation).
This program is typically called by a script to have parallel runs.

Usage: feat2acc_tree.py FOLD [SEQDIR] [FEATDIR] [ACCDIR] [SVM_PARAM] [-b] [SEED]
Options:
    -b  2 models for binary branches
Example:
feat2acc_tree.py 0 ~/subloc/data/hallett2004cv/ features/ acc_leaf_t2c1/ "-t 2 -c 1" -b
'''

nodeList = ['YeastSecPw', 'YeastIntra', 'YeastOrga',
            'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastPM', 'YeastSecreted',
            'YeastCyto', 'YeastPero', 'YeastMito', 'YeastNuclear']
dfsCompList = nodeList[3:]

def feat2accTree(fold, randSeed, seqDir, featDir, accDir, binary1Model, svmOpt):
    # filenames (output)
    resultFile = '%sYeastTree%d_%d.result' % (accDir, randSeed, fold)
    rootTrFeatMcFile  = '%sYeastRoot%dmc%d.dat'  % (accDir, randSeed, fold)
    secPwTrFeatMcFile = '%sYeastSecPw%dmc%d.dat' % (accDir, randSeed, fold)
    intraTrFeatMcFile = '%sYeastIntra%dmc%d.dat' % (accDir, randSeed, fold)
    orgaTrFeatMcFile  = '%sYeastOrga%dmc%d.dat'  % (accDir, randSeed, fold)
    rootPredFile      = '%sYeastRoot%dpr%d.dat'  % (accDir, randSeed, fold)
    secPwPredFile     = '%sYeastSecPw%dpr%d.dat' % (accDir, randSeed, fold)
    intraPredFile     = '%sYeastIntra%dpr%d.dat' % (accDir, randSeed, fold)
    orgaPredFile      = '%sYeastOrga%dpr%d.dat'  % (accDir, randSeed, fold)
    if binary1Model:
        rootModelFiles = ['%sYeastSecPw%d_%d.svm' % (accDir, randSeed, fold)]
        intraModelFiles = ['%sYeastOrga%d_%d.svm' % (accDir, randSeed, fold)]
    else:
        rootModelFiles = ['%s%s%d_%d.svm' % (accDir, comp, randSeed, fold) \
                          for comp in nodeList[:2]]
        intraModelFiles = ['%s%s%d_%d.svm' % (accDir, comp, randSeed, fold) \
                           for comp in [nodeList[2], nodeList[-1]]]
    secPwModelFiles = ['%s%s%d_%d.svm' % (accDir, comp, randSeed, fold) \
                       for comp in dfsCompList[:5]]
    orgaModelFiles = ['%s%s%d_%d.svm' % (accDir, comp, randSeed, fold) \
                      for comp in dfsCompList[5:8]]

    # filenames (input)
    trainSeqFiles = ['%s%s%dtr%d.fa' % (seqDir, name, randSeed, fold) \
                     for name in dfsCompList]
    testSeqFiles  = ['%s%s%dte%d.fa' % (seqDir, name, randSeed, fold) \
                     for name in dfsCompList]
    trainFeatFile = '%sYeast%dtr%d.dat' % (featDir, randSeed, fold)
    testFeatFile  = '%sYeast%dte%d.dat' % (featDir, randSeed, fold)

    # initialization    
    # dfsTrFeat/dfsTeFeat: Depth-First-Search Train/Test Features
    trainSeqId = []; testSeqId = []; trainNSeq = []; testNSeq = []
    for seqFilename in trainSeqFiles:
        idSeqs = fastaReadAll(seqFilename)
        trainSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
        trainNSeq.append(len(idSeqs))
    for seqFilename in testSeqFiles:
        idSeqs = fastaReadAll(seqFilename)
        testSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
        testNSeq.append(len(idSeqs))
    dfsTrFeat = readFeatures(trainNSeq, trainFeatFile)
    dfsTeFeat = readFeatures(testNSeq , testFeatFile )
    (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat) = groupTree(dfsTrFeat)
    (rootTrSeqId,secPwTrSeqId,intraTrSeqId,orgaTrSeqId) = groupTree(trainSeqId)

    # train
    trainSVM(rootTrFeat, rootTrSeqId, rootTrFeatMcFile, rootModelFiles, svmOpt)
    trainSVM(secPwTrFeat,secPwTrSeqId,secPwTrFeatMcFile,secPwModelFiles,svmOpt)
    trainSVM(intraTrFeat,intraTrSeqId,intraTrFeatMcFile,intraModelFiles,svmOpt)
    trainSVM(orgaTrFeat, orgaTrSeqId, orgaTrFeatMcFile, orgaModelFiles, svmOpt)

    # test (classify)
    rootArgMax  = testSVM(testFeatFile, rootModelFiles,  rootPredFile )
    secPwArgMax = testSVM(testFeatFile, secPwModelFiles, secPwPredFile)
    intraArgMax = testSVM(testFeatFile, intraModelFiles, intraPredFile)
    orgaArgMax  = testSVM(testFeatFile, orgaModelFiles,  orgaPredFile )

    # evaluate and write result
    (accuracy, precision, recall) = evaluate(rootArgMax, secPwArgMax, \
            intraArgMax, orgaArgMax, trainSeqId, testSeqId)
    writeResult1Fold(resultFile, accuracy, precision, recall, fold, \
                     dfsCompList, 'multi-compartment tree svm (%s)\n' % svmOpt)

def groupTree(dfsTrFeat):
    'Group the leaf into branches for the tree hierarchy.'
    secPwTrFeat = dfsTrFeat[:5]
    orgaTrFeat = dfsTrFeat[5:8]
    intraTrFeat = [sum(orgaTrFeat), dfsTrFeat[8]]
    rootTrFeat = [sum(secPwTrFeat), sum(intraTrFeat)]
    return (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat)

def trainSVM(trainFeat, trainSeqId, trainFeatFile, modelFiles, svmOpt = ''):
    if len(modelFiles) == 1:
        writeBinFeatures(trainFeat, trainSeqId, 0, trainFeatFile)
        svmLearn(trainFeatFile, modelFiles[0], svmOpt)
        return
    for trainComp in range(len(trainFeat)):
        writeFeatures(trainFeat, trainSeqId, trainComp, trainFeatFile)
        svmLearn(trainFeatFile, modelFiles[trainComp], svmOpt)

def writeBinFeatures(compFeat, compSeqId, trainComp, featFilename):
    '''Write features in SVM light or multiclass format.
    Overlap and replace memecv.writeFeatures by supporting multi-compartment
    for a single sequence.'''
    if len(compFeat) != 2: raise Exception('writeBinFeatures not binary')
    fFeat = open(featFilename, 'w')
    for compart in range(len(compFeat)):
        for id, featVec in zip(compSeqId[compart], compFeat[compart]):
            if id in compSeqId[1 - compart]: continue
            if compart != trainComp:
                fFeat.write('-1')
            else:
                fFeat.write('+1')
            for iFeat in range(len(featVec)):
                fFeat.write('\t%d:%f' % (iFeat + 1, featVec[iFeat]))
            fFeat.write('\n')
    fFeat.close()

def testSVM(testFeatFile, modelFiles, predFile):
    '''testFeatFile remains the same for all target compartments;
    predFile is reused'''
    if len(modelFiles) == 1:
        svmClassify(testFeatFile, modelFiles[0], predFile)
        pred = svmLightReadOutput(predFile) #predFile reused
        predArgMax = [int(pred1 < 0.) for pred1 in pred] #+/- into 0/1
        return predArgMax
    for comp in range(len(modelFiles)):
        svmClassify(testFeatFile, modelFiles[comp], predFile)
        pred = svmLightReadOutput(predFile) #predFile reused
        if comp == 0: # first, init as max
            predMax = pred; predArgMax = [0] * len(pred)
        else:
            for i in range(len(pred)):
                if pred[i] > predMax[i]:
                    predMax[i] = pred[i]; predArgMax[i] = comp
    return predArgMax

def evaluate(rootArgMax, secPwArgMax, intraArgMax, orgaArgMax, trainSeqId, \
             testSeqId):
    # Remove overlap between training and testing
    trainIds = dict([(id, 1) for ids in trainSeqId for id in ids]) #not to test
    testIds = dict([(id, 1) for ids in testSeqId for id in ids]) #assert
    nUniqueTest = sum([(id not in trainIds) for id in testIds.keys()]) #assert
    # Initialization
    corr = [0] * 9; nPred = [0] * 9; precision = [.0] * 9; recall = [.0] * 9
    id2pred = {}; uniqueIds = [None for comp in range(9)]
    cumu = 0
    for comp in range(9):
        cumu1 = cumu + len(testSeqId[comp])
        ids = {}
        for i in range(cumu, cumu1):
            id = testSeqId[comp][i-cumu]
            if not id in trainIds:
                ids[id] = 1
                if id in id2pred:
                    assert id2pred[id] == (rootArgMax[i], secPwArgMax[i], \
                                           intraArgMax[i], orgaArgMax[i])
                id2pred[id] = (rootArgMax[i], secPwArgMax[i], intraArgMax[i],\
                               orgaArgMax[i])
        uniqueIds[comp] = ids.keys()
        cumu = cumu1
    print 'uniqueIds:'
    for uid in uniqueIds: print uid
    # ER,Golgi,Lyso,PM,Sec        
    for c1 in range(5):
        for id in uniqueIds[c1]:
            argMax = id2pred[id]
            if argMax[0]==0 and argMax[1]==c1: corr[c1] = corr[c1] + 1
        nPred[c1] = len(dict([(id,1) for id,argMax in id2pred.items() \
                              if argMax[0]==0 and argMax[1]==c1]))
    # Cyto,Mito,Pero
    for c1 in range(5, 8):
        for id in uniqueIds[c1]:
            argMax = id2pred[id]
            if argMax[0]==1 and argMax[2]==0 and argMax[3]==c1-5:
                corr[c1] = corr[c1] + 1
        nPred[c1] = len(dict([(id,1) for id,argMax in id2pred.items() \
                    if argMax[0]==1 and argMax[2]==0 and argMax[3]==c1-5]))
    # Nuclear
    for id in uniqueIds[8]:
        argMax = id2pred[id]
        if argMax[0]==1 and argMax[2]==1: corr[8] = corr[8] + 1
    nPred[8] = len(dict([(id,1) for id,argMax in id2pred.items() \
                          if argMax[0]==1 and argMax[2]==1]))
    nTestSeq = sum(nPred)
    print 'corr:'; print corr
    print 'nPred:'; print nPred
    print 'nTrue:'; print [len(uid) for uid in uniqueIds]
    for c1 in range(9):
        tp = float(corr[c1])
        if nPred[c1] > 0: precision[c1] = tp / nPred[c1]
        else:             precision[c1] = 0.0
        if len(uniqueIds[c1]) > 0: recall[c1] = tp / len(uniqueIds[c1])
        else:                      recall[c1] = 0.0
    accuracy = float(sum(corr)) / nTestSeq
    if nUniqueTest != nTestSeq:
        raise Exception('Incorrect nTestSeq (repeated prediction mismatch)' \
                        ' nTestSeq=%d (should be %d), #trainseq=%d' \
                        % (nTestSeq, nUniqueTest, len(trainIds)))
    print '%d unique sequences processed (training: %d, testing: %d)' \
          % (len(trainIds) + nUniqueTest, len(trainIds), nUniqueTest)
    return (accuracy, precision, recall)
    
# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2: print USAGE; sys.exit(1)
    fold = int(sys.argv[1])
    seqDir = ''; featDir = ''; resDir = ''; svmOpt = ''; randSeed = 1000
    if len(sys.argv) > 2: seqDir = sys.argv[2]
    if len(sys.argv) > 3: featDir = sys.argv[3]
    if len(sys.argv) > 4: resDir = sys.argv[4]
    if len(sys.argv) > 5: svmOpt = sys.argv[5]
    if len(sys.argv) > 6 and sys.argv[6].lower() == '-b': binary1Model = False
    else: binary1Model = True
    if len(sys.argv) > 7: randSeed = int(sys.argv[7])
    #t0 = time.time()
    feat2accTree(fold, randSeed, seqDir, featDir, resDir, binary1Model, svmOpt)
    #print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
