#!/usr/bin/python
# Train a SVM under a tree topology,
# features including internal and leaf motifs.
# This program supports the deep directory structure and multi-compartments.

import sys
from memecv import sum, fastaReadAll, readFeatures
from feat2acc_tree import nodeList, dfsCompList, groupTree, testSVM, evaluate, \
                          writeResult1Fold

USAGE = 'feat2acc_treete RESFILE FEATFILE PREDPAT MODELPAT SEQPAT [SVMOPT] [-b]'

def treeTest1(resultFile, testFeatFile, predPattern, modelPattern, testSeqFiles,\
              binary1Model, svmOpt):
    # filenames
    rootPredFile  = predPattern % 'YeastRoot'
    secPwPredFile = predPattern % 'YeastSecPw'
    intraPredFile = predPattern % 'YeastIntra'
    orgaPredFile  = predPattern % 'YeastOrga'
    if binary1Model:
        rootModelFiles = [modelPattern % 'YeastSecPw']
        intraModelFiles = [modelPattern % 'YeastOrga']
    else:
        rootModelFiles = [modelPattern % comp for comp in nodeList[:2]]
        intraModelFiles = [modelPattern % comp \
                           for comp in [nodeList[2], nodeList[-1]]]
    secPwModelFiles = [modelPattern % comp for comp in dfsCompList[:5]]
    orgaModelFiles = [modelPattern % comp for comp in dfsCompList[5:8]]

    # initialization
    # dfsTrFeat/dfsTeFeat: Depth-First-Search Train/Test Features
    testSeqId = []
    for seqFilename in testSeqFiles:
        idSeqs = fastaReadAll(seqFilename)
        testSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
    
    # test (classify)
    rootArgMax  = testSVM(testFeatFile, rootModelFiles,  rootPredFile )
    secPwArgMax = testSVM(testFeatFile, secPwModelFiles, secPwPredFile)
    intraArgMax = testSVM(testFeatFile, intraModelFiles, intraPredFile)
    orgaArgMax  = testSVM(testFeatFile, orgaModelFiles,  orgaPredFile )

    # evaluate and write result
    (accuracy, precision, recall, id2pred) = evaluate(rootArgMax, secPwArgMax,\
            intraArgMax, orgaArgMax, testSeqId)
    writeResult1Fold(resultFile, accuracy, precision, recall, -1, \
                     dfsCompList, 'multi-compartment tree svm (%s)\n' % svmOpt)
    writeCompPred(id2pred, testSeqId, resultFile)

def evaluate(rootArgMax, secPwArgMax, intraArgMax, orgaArgMax, testSeqId):
    # Remove overlap between training and testing
    testIds = dict([(id, 1) for ids in testSeqId for id in ids]) #assert
    nUniqueTest = len(testIds) #assert
    # Initialization
    corr = [0] * 9; nPred = [0] * 9; precision = [.0] * 9; recall = [.0] * 9
    id2pred = {}; uniqueIds = [None for comp in range(9)]
    cumu = 0
    id2i = {}
    for comp in range(9):
        cumu1 = cumu + len(testSeqId[comp])
        ids = {}
        for i in range(cumu, cumu1):
            id = testSeqId[comp][i-cumu]
            ids[id] = 1
            if id in id2pred:
                assert id2pred[id] == (rootArgMax[i], secPwArgMax[i], \
                       intraArgMax[i], orgaArgMax[i]), \
                       '%s[%d != %d]: %s != %s' % (id, id2i[id], i, \
                       id2pred[id], (rootArgMax[i],\
                       secPwArgMax[i], intraArgMax[i], orgaArgMax[i]))
            id2i[id] = i
            id2pred[id] = (rootArgMax[i], secPwArgMax[i], intraArgMax[i],\
                           orgaArgMax[i])
        uniqueIds[comp] = ids.keys()
        cumu = cumu1
    # print 'uniqueIds:'
    # for uid in uniqueIds: print uid
    # ER,Golgi,Lyso,PM,Sec        
    for c1 in range(5):
        for id in uniqueIds[c1]:
            argMax = id2pred[id]
            if argMax[0]==0 and argMax[1]==c1: corr[c1] = corr[c1] + 1
        nPred[c1] = len(dict([(id,1) for id,argMax in id2pred.items() \
                              if argMax[0]==0 and argMax[1]==c1]))
    # Cyto,Mito,Pero
    for c1 in range(5, 8):
        for id in uniqueIds[c1]:
            argMax = id2pred[id]
            if argMax[0]==1 and argMax[2]==0 and argMax[3]==c1-5:
                corr[c1] = corr[c1] + 1
        nPred[c1] = len(dict([(id,1) for id,argMax in id2pred.items() \
                    if argMax[0]==1 and argMax[2]==0 and argMax[3]==c1-5]))
    # Nuclear
    for id in uniqueIds[8]:
        argMax = id2pred[id]
        if argMax[0]==1 and argMax[2]==1: corr[8] = corr[8] + 1
    nPred[8] = len(dict([(id,1) for id,argMax in id2pred.items() \
                          if argMax[0]==1 and argMax[2]==1]))
    nTestSeq = sum(nPred)
    print 'corr:'; print corr
    print 'nPred:'; print nPred
    print 'nTrue:'; print [len(uid) for uid in uniqueIds]
    for c1 in range(9):
        tp = float(corr[c1])
        if nPred[c1] > 0: precision[c1] = tp / nPred[c1]
        else:             precision[c1] = 0.0
        if len(uniqueIds[c1]) > 0: recall[c1] = tp / len(uniqueIds[c1])
        else:                      recall[c1] = 0.0
    accuracy = float(sum(corr)) / nTestSeq
    assert nUniqueTest == nTestSeq, \
           'Incorrect nTestSeq (repeated prediction mismatch) nTestSeq=%d '\
           '(should be %d)' % (nTestSeq, nUniqueTest)
    print '%d unique sequences processed (testing)' % nUniqueTest
    return (accuracy, precision, recall, id2pred)

def writeCompPred(id2pred, seqId, predFile):
    'From feat2acc_test.writeCompPred()'
    for comp, ids in zip(dfsCompList, seqId):
        fp = open(predFile + '.' + comp, 'w')
        for id in ids: fp.write('%s\t%s\n' % (id, decisionTree(id2pred[id])))
        fp.close()

def decisionTree(argMax):
    if argMax[0] == 0: pred = argMax[1]
    elif argMax[2]==0: pred = argMax[3] + 5
    else: pred = 8
    return dfsCompList[pred]

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 5: print USAGE; sys.exit(1)
    resultFile      = sys.argv[1]
    testFeatFile    = sys.argv[2]
    predPattern     = sys.argv[3]
    modelPattern    = sys.argv[4]
    seqPattern      = sys.argv[5]
    svmOpt = ''
    if len(sys.argv) > 6: svmOpt = sys.argv[6]
    if len(sys.argv) > 7 and sys.argv[7].lower() == '-b': binary1Model = False
    else: binary1Model = True
    testSeqFiles = [seqPattern % comp for comp in dfsCompList]
    treeTest1(resultFile, testFeatFile, predPattern, modelPattern, testSeqFiles,\
              binary1Model, svmOpt)
