#Rank features (motifs) based on training accuracy of excluding the features.
#For each fold, rank features with training data and then
#evaluate with testing data.
#Algorithm:
#  for one fold
#  read all features
#  for each feature
#    remove feature
#    write features and run svm
#    record training accuracy
#  rank by reduction in training accuracy
#  number of features vs. testing accuracy
#  return featRanks, list of tuple

import sys, time
from memecv import sum, fastaReadAll, readFeatures
from feat2acc_tree import trainSVM, testSVM

USAGE = \
'''Rank features (motifs) based on training accuracy of excluding the features.
Usage: featflat2rank.py TR_FEAT TR_SEQ RES_DIR [SVM_OPT]
  TR_FEAT     training feature file (ex. feattree/Yeast1000tr3.dat)
  TR_SEQ      training sequence file pattern
              (ex. ~/subloc/data/hallett2005/%s1000tr3.fa)
  RES_PREFIX  output result prefix (ex. rank_flat_c.001/Yeast1000_3)
  SVM_OPT     options for SVMlight
'''

COMPS = ['YeastCyto', 'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastMito',
         'YeastNuclear', 'YeastPero', 'YeastPM', 'YeastSecreted']

def rankTestAcc(trSeqPattern, trainFeatFile0, rankFile, trainFeatFile, \
                modelFilePattern, predFile, svmOpt):
    'Rank features and write to a file'
    featRanks = rankFeat(trSeqPattern, trainFeatFile0, trainFeatFile, \
                modelFilePattern, predFile, svmOpt)
##    testAcc(featRanks)
    writeRank(featRanks, rankFile)

def rankFeat(trSeqPattern, trainFeatFile0, trainFeatFile, modelFilePattern, \
             predFile, svmOpt):
    'Rank features by training accuracy after removing a feature'
    trSeqFiles = [trSeqPattern % comp for comp in COMPS]
    modelFiles = [modelFilePattern % comp for comp in COMPS]
    trainSeqId = []; trainNSeq = []#; testSeqId = []; testNSeq = []
    for seqFilename in trSeqFiles:
        idSeqs = fastaReadAll(seqFilename)
        trainSeqId.append([idSeq[0].split()[0] for idSeq in idSeqs])
        trainNSeq.append(len(idSeqs))
    trainFeat = readFeatures(trainNSeq, trainFeatFile0)
    featRanks = []
    for feat in range(len(trainFeat[0][0])):
        reducedFeat = reduceFeat(trainFeat, feat)
        trainSVM(reducedFeat, trainSeqId, trainFeatFile, modelFiles, svmOpt)
        predArgMax = testSVM(trainFeatFile, modelFiles, predFile)
        trainAcc = evalSVM(trainSeqId, predArgMax)
        featRanks.append((trainAcc, feat, 0))
    featRanks.sort() #by trainAcc and then feat
    return featRanks

def reduceFeat(trainFeat, feat):
    '''Reduce feature set by removing a feature.
    trainFeat[comp][seq][feat] is a 3D array; feat is the index starting from 0.'''
    reducedFeat = []
    for comp1Feat in trainFeat:
        reducedFeat.append([featVec[:feat] + featVec[feat+1:] \
                            for featVec in comp1Feat])
        for featVec, featVec1 in zip(comp1Feat, reducedFeat[-1]):
            assert len(featVec) - 1 == len(featVec1)
    return reducedFeat

def evalSVM(testSeqId, predArgMax):
    nComp = len(testSeqId)
    uniqueIds = dict([(id, 1) for ids in testSeqId for id in ids])
    corr = 0; cumu = 0
    for comp in range(nComp):
        cumu1 = cumu + len(testSeqId[comp])
        corr = corr + sum([(argMax==comp) for argMax in predArgMax[cumu:cumu1]])
        cumu = cumu1
    print '#correct = %d, #seq = %d (unique)' % (corr, len(uniqueIds))
    return float(corr) / len(uniqueIds)

##def testAcc():
##    for nFeat in len(featRanks):
##        topFeat = featRanks[:nFeat]
##        (corr, nEvalSeq, precision, recall) = trainTest(trainFeat, trainSeqId, \
##                trainNSeq, trainSeqId, trainFeatFilename, trainFeatFilename, \
##                modelFilename, predFilename, svmOpt)
##        testAcc = float(corr) / nEvalSeq
##        featRanks[nFeat][2] = testAcc
    
def writeRank(featRanks, rankFile):
    'Write ranking result to a file'
    f = open(rankFile, 'w')
    for trainAcc, feat, testAcc in featRanks:
        f.write('%d\t%f\n' % (feat, trainAcc))
    f.close()

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 4: print USAGE; sys.exit(1)
    if len(sys.argv) > 4: svmOpt = sys.argv[4]
    trainFeatFile0      = sys.argv[1]
    trSeqPattern        = sys.argv[2]
    resPrefix           = sys.argv[3]
    
    rankFile            = resPrefix + '.result'
    trainFeatFile       = resPrefix + 'mc.dat'
    modelFilePattern    = resPrefix + '%s.svm'
    predFile            = resPrefix + '.pred'
    t0 = time.time()
    rankTestAcc(trSeqPattern, trainFeatFile0, rankFile, trainFeatFile, \
                modelFilePattern, predFile, svmOpt)
    print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
