#!/usr/bin/env python
# Evaluate MEME+SVM by cross validation

# Input: YeastER.txt, YeastPM.txt, ...
# For each fold,
#     Training sequences: YeastER1000tr3.fa, YeastPM1000tr3.fa, ...
#     Testing sequences: Yeast1000te3.fa (with labels)
#     MEME output: YeastER1000tr3.meme, YeastPM1000tr3.meme, ...
#     Testing vectors: Yeast1000test.dat
#     For each compartment,
#         Training vectors: Yeast1000tr.dat (only labels changes)
#         SVM Light: Yeast1000.svm, Yeast1000pred.dat
# Accuracy: Yeast1000.result

import sys, os, glob, re, math, random, types, time

USAGE = \
'''Usage: memecv.py <split|feat|eval|eval1> [SEED] [SEQ_PREFIX] [FOLD|NUMFOLD]
Default: SEQDIR = ~/subloc/data/hallett2004/Yeast, NUMFOLD = 10
Note: assumes reading Yeast*.txt

Commands:
  memecv.py feat FOLD [EXT] [SEQDIR] [MOTIFDIR] [FEATDIR] [SEED]
    e.g. memecv.py feat 0 meme3
    e.g. memecv.py feat 0 submotif ~/subloc/data/hallett2004cv motif features

memecv.py split 1000 ../../data/hallett2004/Yeast 10
memecv.py eval1 1000 ../../data/hallett2004/Yeast 0
memecv.py eval1 1000 ../../data/hallett2004/Yeast 0  multiclass
memecv.py eval1 1000 ../../data/hallett2004/Yeast 0  multiclass meme3
memecv.py svmlight2multi 1000 ../../data/hallett2004/Yeast 0
memecv.py svmlight 1000 ../../data/hallett2004/Yeast 0 "-c 1..."'''
#memecv.py eval  1000 ../../data/hallett2004/Yeast 10
#memecv.py eval2 1000 ../Hera Yeast 0

SVM_LEARN = '~/svm_light/svm_learn'
SVM_CLASSIFY = '~/svm_light/svm_classify'
SVM_MULTICLASS_LEARN = '~/svm_multiclass/svm_multiclass_learn'
SVM_MULTICLASS_CLASSIFY = '~/svm_multiclass/svm_multiclass_classify'

def sum(sequence):
    'This function should be removed under Python 2.3+'
    if len(sequence) == 0: return 0
    total = sequence[0]
    for x in sequence[1:]: total = total + x
    return total

# Amino acid & PWM routines

AMINOACIDS = 'ACDEFGHIKLMNPQRSTVWY'

def aminoAcidCode(letter):
    global AMINOACIDS
    if not letter in AMINOACIDS:
        raise Exception('site not in FASTA characters: %s' % letter)
    return AMINOACIDS.index(letter)

def pwmCountSite(pwm, site):
    site = site.upper()
    for pos in range(len(site)):
        aa = aminoAcidCode(site[pos])
        pwm[pos][aa] = pwm[pos][aa] + 1

def pwmNormalizeLog(pwm):
    n = 0
    for aaCount in pwm[0]: n = n + aaCount
    n = float(n)
    for v in pwm:
        for aa in range(len(v)): v[aa] = v[aa] / n
    for v in pwm:
        sumV = 0.0;
        for aaCount in v: sumV = sumV + aaCount
        if abs(1 - sumV) > .001: raise Exception('normalization')
    for v in pwm:
        for aa in range(len(v)): v[aa] = math.log(v[aa])

def pwmLogLik(pwm, site):
    loglik = 0.0
    for logFreq, letter in zip(pwm, site):
        loglik = loglik + logFreq[aminoAcidCode(letter)]
    return loglik

def pwmFeatures(allPwms, seq):
    featVec = []
    for pwm in allPwms:
        maxLogLik = -1e100
        for p in range(len(seq) - len(pwm) + 1):
            loglik = pwmLogLik(pwm, seq[p : p+len(pwm)])
            if 0==p or loglik > maxLogLik: maxLogLik = loglik
        featVec.append(maxLogLik)
        #featVec.append(max([pwmLogLik(pwm, seq[p : p+len(pwm)]) \
        #                    for p in range(len(seq) - len(pwm) + 1)]))
    return featVec

def readMeme(memeFilename):
    print 'Reading %s...' % memeFilename
    pwms = []
    f = open(memeFilename)
    while True:
        line = f.readline()
        if line == '': break
        m = re.search(r'Motif (\d+) sites sorted by position', line)
        if m != None:
            #print 'motif %s' % m.group(1)
            f.readline(); f.readline(); f.readline()
            pwm = None
            while True:
                lineTokens = f.readline().split()
                if len(lineTokens) < 5: break
                site = lineTokens[4]
                if None == pwm: pwm = [[1] * 20 for pos in range(len(site))]
                if len(site) != len(pwm):
                    print 'PWM length is %d but site is %s, lineTokens:' \
                          % (len(pwm), site)
                    print lineTokens
                pwmCountSite(pwm, site)
            pwmNormalizeLog(pwm)
            pwms.append(pwm)
    f.close()
    return pwms

# FASTA related IO routines

fastaCurrSeqName = ''

def fastaInit(seqFilename):
    global fastaCurrSeqName
    fSeq = open(seqFilename)
    line = fSeq.readline()
    if line.startswith('>'): fastaCurrSeqName = line[1:].strip()
    else: raise Exception('first line not start with >')
    return fSeq

def fastaReadNext(fSeq):
    global fastaCurrSeqName
    seq = ''
    while True:
        line = fSeq.readline()
        if len(line) == 0:
            if len(seq) > 0: name = fastaCurrSeqName; break
            else: return (None, None)
        line = line.strip()
        if len(line.strip()) == 0 or line.startswith('#'): continue
        elif line.startswith('>'):
            name = fastaCurrSeqName; fastaCurrSeqName = line[1:]; break
        elif line[-1] == '*': seq = seq + line[:-1]
        else: seq = seq + line
    seq = re.sub('[^'+AMINOACIDS+']', '', seq)
    return (name, seq)

def fastaReadAll(seqFilename):
    fSeq = fastaInit(seqFilename)
    nameSeqs = []
    while True:
        nameSeq = fastaReadNext(fSeq)
        if None == nameSeq[0]: break
        nameSeqs.append(nameSeq)
    fSeq.close()
    return nameSeqs

def fastaWrite(nameSeqs, seqFilename):
    fSeq = open(seqFilename, 'w')
    for name, seq in nameSeqs: fSeq.write('>%s\n%s\n' % (name, seq))
    fSeq.close()

def fastaNSeq(seqFilename):
    fSeq = fastaInit(seqFilename)
    nSeq = 0
    while True:
        (name, seq) = fastaReadNext(fSeq)
        if None == seq: break
        nSeq = nSeq + 1
    fSeq.close()
    return nSeq

# SVM-light related routines

def writeFeatures(compSeqFeat, trainComp, featFilename):
    'Write features in SVM light or multiclass format'
    fFeat = open(featFilename, 'w')
    for compart in range(len(compSeqFeat)):
        for featVec in compSeqFeat[compart]:
            if   trainComp == -1: fFeat.write('0')
            elif trainComp == -2: fFeat.write('%d' % (compart + 1))
            else: fFeat.write('%+d' % (2 * (compart==trainComp) - 1))
            for iFeat in range(len(featVec)):
                fFeat.write('\t%d:%f' % (iFeat + 1, featVec[iFeat]))
            fFeat.write('\n')
    fFeat.close()

def readFeatures(compNSeq, featFilename):
    fFeat = open(featFilename)
    compFeat = []
    for nSeq in compNSeq:
        featVecs = []
        for i in range(nSeq):
            line = fFeat.readline()
            if line == '':
                raise Exception('Feature file ends before calculated number: %s'\
                                % featFilename)
            featVec = []
            for token in line.split()[1:]:
                if int(token[ : token.index(':')]) != len(featVec) + 1:
                    raise Exception('Wrong feature number')
                featVec.append(float(token[token.index(':')+1 : ]))
            featVecs.append(featVec)
        compFeat.append(featVecs)
    fFeat.close()
    return compFeat

def svmLearn(trainFilename, modelFilename, opt = ''):
    cmd = '%s %s %s %s' % (SVM_LEARN, opt, trainFilename, modelFilename)
    print cmd; os.system(cmd)

def svmClassify(testFilename, modelFilename, predFilename):
    cmd = '%s %s %s %s' % (SVM_CLASSIFY, testFilename, modelFilename, \
                           predFilename)
    print cmd; os.system(cmd)

def svmLightReadOutput(filename):
    f = open(filename)
    pred = [float(line) for line in f]
    f.close()
    return pred

def svmMulticlassLearn(trainFilename, modelFilename, opt = ''):
    cmd = '%s %s %s %s' % (SVM_MULTICLASS_LEARN,opt,trainFilename,modelFilename)
    print cmd; os.system(cmd)

def svmMulticlassClassify(testFilename, modelFilename, predFilename):
    cmd = '%s %s %s %s' % (SVM_MULTICLASS_CLASSIFY, testFilename, modelFilename, \
                           predFilename)
    print cmd; os.system(cmd)

def svmMulticlassReadOutput(filename):
    f = open(filename)
    pred = [int(line)-1 for line in f]
    f.close()
    return pred

# PWM to features routines

def seqFeatures(allPwms, seqFilename):
    print 'Matching motifs in %s...' % seqFilename
    fSeq = fastaInit(seqFilename)
    seqFeat = []
    while True:
        (name, seq) = fastaReadNext(fSeq)
        if None == seq: break
        seqFeat.append(pwmFeatures(allPwms, seq))
    fSeq.close()
    return seqFeat

# MEME-SVM cross validation routines

def meme2Feat(fold, randSeed, memeExt, seqDir, memeDir, featDir):
    '''Generate features from motif PWM.
    See memecvtree3.meme2AllFeat for hierarchical version.
    Unlike other functions, caller can specify the directory of
    training/testing sequences and motif files.'''
    comps = []
    compPattern = '%s([a-zA-Z]+)%dtr0.%s' % (memeDir, randSeed, memeExt) 
    for compFile in glob.glob('%s*%dtr0.%s' % (memeDir, randSeed, memeExt)):
        m = re.search(compPattern, compFile); comps.append(m.group(1))
    (trainFiles0, testFiles0, memeFiles0) = foldFilenames(\
        comps, randSeed, fold, memeExt)
    trainFiles = [seqDir + filename for filename in trainFiles0]
    testFiles  = [seqDir + filename for filename in testFiles0 ]
    memeFiles  = [memeDir+ filename for filename in memeFiles0 ]
    (trainFeat, testFeat) = memeFeatures(memeFiles, trainFiles, testFiles)
    #featNormalize(trainFeat, testFeat)
    writeFeatures(testFeat , -1, '%s/Yeast%dtest0.dat' % (featDir, randSeed))
    writeFeatures(trainFeat, -1, '%s/Yeast%dtr0.dat'   % (featDir, randSeed))

def cvEvaluate(comps, nFold, randSeed, isMultiClass):
    'Evaluate previous MEME PWM. Must split files and train MEME first.'
    resultFilename      = 'Yeast%d.result'  % randSeed
    testFeatFilename    = 'Yeast%dtest.dat' % randSeed
    trainFeatFilename   = 'Yeast%dtr.dat'   % randSeed
    modelFilename       = 'Yeast%d.svm'     % randSeed
    predFilename        = 'Yeast%dpred.dat' % randSeed

    corr = 0; nSeq = 0; accFold = []
    precision = [0.0] * len(compSeqFiles); recall = [0.0] * len(compSeqFiles)
    for fold in range(nFold):
        (corr1, nTestSeq1, prec1, reca1) = evaluateFold(comps, \
                nFold, randSeed, fold, trainFeatFilename, testFeatFilename, \
                modelFilename, predFilename, isMultiClass)
        accFold.append(float(corr1) / nTestSeq1)
        corr = corr + corr1; nSeq = nSeq + nTestSeq1
        for comp in range(len(prec1)):
            precision[comp] = precision[comp] + prec1[comp]
            recall[comp] = recall[comp] + reca1[comp]
    writeResult(resultFilename, float(corr) / nSeq, accFold, precision, \
                recall, comps)

def cvEvaluate1Fold(comps, fold, randSeed, isMultiClass, memeExt):
    result1Filename     = 'Yeast%d_%d.result' % (randSeed, fold)
    testFeatFilename    = 'Yeast%dte%d.dat'   % (randSeed, fold)
    trainFeatFilename   = 'Yeast%dtr%d.dat'   % (randSeed, fold)
    modelFilename       = 'Yeast%d_%d.svm'    % (randSeed, fold)
    predFilename        = 'Yeast%dpr%d.dat'   % (randSeed, fold)
    (corr1, nTestSeq1, precision, recall) = evaluateFold(comps, randSeed, fold,\
            trainFeatFilename, testFeatFilename, modelFilename, predFilename,\
            isMultiClass, memeExt)
    writeResult1Fold(result1Filename, float(corr1) / nTestSeq1, precision, \
                     recall, fold, comps, '')

def evaluateFold(comps, randSeed, fold, trainFeatFilename, \
                 testFeatFilename, modelFilename, predFilename, isMultiClass, \
                 memeExt = 'meme'):
    (trainFiles, testFiles, memeFiles) = foldFilenames(\
        comps, randSeed, fold, memeExt)
    (trainFeat, testFeat) = memeFeatures(memeFiles, trainFiles, testFiles)
    #writeFeatures(testFeat , -1, 'Yeast%dtest0.dat' % randSeed)
    #writeFeatures(trainFeat, -1, 'Yeast%dtr0.dat'   % randSeed)
    featNormalize(trainFeat, testFeat)
    if isMultiClass:
        return trainTestMulticlass(trainFeat, testFeat, trainFeatFilename,\
                         testFeatFilename, modelFilename, predFilename)
    else:
        return trainTest(trainFeat, testFeat, trainFeatFilename, \
                         testFeatFilename, modelFilename, predFilename)

def splitSeqFiles(compSeqFiles, nFold, randSeed):
    'Split FASTA sequence files into training and testing data for CV.'
    random.seed(randSeed)
    compSeqs = [fastaReadAll(seqFile) for seqFile in compSeqFiles]
    for comp in range(len(compSeqs)): random.shuffle(compSeqs[comp])
    for fold in range(nFold):
        (trainFiles, testFiles, memeFiles) = foldFilenames(\
            seqFilenames2Comps(compSeqFiles), randSeed, fold)
        for nameSeq, trainFile, testFile in zip(compSeqs, trainFiles, testFiles):
            test1 = len(nameSeq) * fold / nFold
            test2 = len(nameSeq) * (fold + 1) / nFold
            fastaWrite(nameSeq[:test1] + nameSeq[test2:], trainFile)
            fastaWrite(nameSeq[test1 : test2], testFile)
            print 'Wrote to %s and %s' % (trainFile, testFile)

def seqFilenamesOrComps(str):
    if ',' in str: return str.split(',')
    else: return seqFilenames2Comps(glob.glob('%s*.txt' % str))

def seqFilenames2Comps(compSeqFiles):
    comps = []
    for compSeqFile in compSeqFiles:
        m = re.search(r'(\w+).txt', compSeqFile); compName = m.group(1)
        comps.append(compName)
    return comps

def foldFilenames(comps, randSeed, fold, memeExt = 'meme'):
    trainFiles = []; testFiles = []; memeFiles = []
    for compName in comps:
        (trainFile, testFile, memeFile) = foldFilename(compName, randSeed, \
                                                       fold, memeExt)
        trainFiles.append(trainFile); testFiles.append(testFile)
        memeFiles.append(memeFile)
    return (trainFiles, testFiles, memeFiles)

def foldFilename(compName, randSeed, fold, memeExt = 'meme'):
    trainFile = '%s%dtr%d.fa' % (compName, randSeed, fold)
    testFile  = '%s%dte%d.fa' % (compName, randSeed, fold)
    memeFile  = '%s%dtr%d.%s' % (compName, randSeed, fold, memeExt)
    return (trainFile, testFile, memeFile)

def runMeme(trainFiles, memeFiles):
    raise Exception('Should use qarray')
    #os.system('meme %s -nmotifs 10 -minw 4 -mod oops -maxsize 300000 -text > %s' % (trainFiles, memeFiles))

def memeFeatures(memeFiles, trainFiles, testFiles):
    allPwms = []
    for memeFilename in memeFiles:
        allPwms = allPwms + readMeme(memeFilename)
    trainFeat = []; testFeat = []
    for seqFilename in trainFiles:
        trainFeat.append(seqFeatures(allPwms, seqFilename))
    for seqFilename in testFiles:
        testFeat.append(seqFeatures(allPwms, seqFilename))
    #featNormalize(trainFeat, testFeat)
    return (trainFeat, testFeat)

def featNormalize(trainFeat, testFeat):
    nFeat = len(trainFeat[0][0]); nTrainSeq = 0
    x1 = [0] * nFeat; x2 = [0] * nFeat; m = [0] * nFeat; v = [0] * nFeat
    print 'nFeat = %d' % nFeat
    for comp in range(len(trainFeat)):
        for i in range(len(trainFeat[comp])):
            if len(trainFeat[comp][i]) != nFeat:
                print 'trainFeat[%d][%d]: %d' % (comp,i,len(trainFeat[comp][i]))
    for compTrainFeat in trainFeat:
        nTrainSeq = nTrainSeq + len(compTrainFeat)
        for featVec in compTrainFeat:
            for j in range(nFeat):
                x1[j] = x1[j] + featVec[j]
                x2[j] = x2[j] + featVec[j] ** 2
    for j in range(nFeat):
        m[j] = x1[j] / nTrainSeq
        v[j] = (x2[j] - m[j] ** 2) / nTrainSeq
    for compTrainFeat in trainFeat:
        for featVec in compTrainFeat:
            for j in range(nFeat): featVec[j] = (featVec[j] - m[j]) / v[j]
    for compTestFeat in testFeat:
        for featVec in compTestFeat:
            for j in range(nFeat): featVec[j] = (featVec[j] - m[j]) / v[j]

def trainTest(trainFeat, testFeat, trainFeatFilename, testFeatFilename, \
              modelFilename, predFilename, svmOpt = ''):
    '''Train and test the given (multiclass) feature set.
    The training feature file include all features of all comp-seq,
    with +1/-1 labels depending on the compartment (1-vs-all).
    The testing feature file include all features all seqs and null labels.
    Support a general mode in which features are written to files,
    and a reuse mode in which previous SVM-light feature files
    are re-used. In the later mode, training feature files are first read,
    and then re-written with the labels changed; testing feature files are
    re-used, including the labels, so no need to read features first.
    However for evaluation purpose, the number of seq of each comptment is
    still required, so caller must pass argument testFeat as testNSeq,
    a list of int.'''
    testNSeq = None
    if types.IntType == type(testFeat[0]): testNSeq = testFeat
    if testNSeq == None: writeFeatures(testFeat, -1, testFeatFilename)
    for trainComp in range(len(trainFeat)):
        writeFeatures(trainFeat, trainComp, trainFeatFilename)
        svmLearn(trainFeatFilename, modelFilename, svmOpt)
        svmClassify(testFeatFilename, modelFilename, predFilename)
        pred = svmLightReadOutput(predFilename)
        if trainComp == 0:
            predMax = pred; predArgMax = [0] * len(pred)
        else:
            for i in range(len(pred)):
                if pred[i] > predMax[i]:
                    predMax[i] = pred[i]; predArgMax[i] = trainComp
    cumu = 0; corr = 0; precision = []; recall = []
    if testNSeq == None: nComp = len(testFeat)
    else: nComp = len(testNSeq)
    for testCompart in range(nComp):
        if testNSeq == None: nSeq = len(testFeat[testCompart])
        else: nSeq = testNSeq[testCompart]
        nPred = 0; tp = 0
        for argMax in predArgMax:
            if argMax == testCompart: nPred = nPred + 1
        for argMax in predArgMax[cumu : cumu+nSeq]:
            if argMax == testCompart: tp = tp + 1
        if nPred > 0: precision.append(float(tp) / nPred)
        else: precision.append(0.0)
        recall.append(float(tp) / nSeq)
        corr = corr + tp; cumu = cumu + nSeq
    return (corr, cumu, precision, recall)

def trainTestMulticlass(trainFeat, testFeat, trainFeatFilename, \
                        testFeatFilename, modelFilename, predFilename, svmOpt=''):
    '''Train and test using SVM-multiclass.
    See trainTest() for detail descriptions.'''
    testNSeq = None
    if types.IntType == type(testFeat[0]): testNSeq = testFeat
    if testNSeq == None: writeFeatures(testFeat, -1, testFeatFilename)
    writeFeatures(trainFeat, -2, trainFeatFilename)
    svmMulticlassLearn(trainFeatFilename, modelFilename, svmOpt)
    svmMulticlassClassify(testFeatFilename, modelFilename, predFilename)
    predArgMax = svmMulticlassReadOutput(predFilename)

    cumu = 0; corr = 0; precision = []; recall = []
    if testNSeq == None: nComp = len(testFeat)
    else: nComp = len(testNSeq)
    for testCompart in range(nComp):
        if testNSeq == None: nSeq = len(testFeat[testCompart])
        else: nSeq = testNSeq[testCompart]
        nPred = 0; tp = 0
        for argMax in predArgMax:
            if argMax == testCompart: nPred = nPred + 1
        for argMax in predArgMax[cumu : cumu+nSeq]:
            if argMax == testCompart: tp = tp + 1
        if nPred > 0: precision.append(float(tp) / nPred)
        else: precision.append(0.0)
        recall.append(float(tp) / nSeq)
        corr = corr + tp; cumu = cumu + nSeq
    return (corr, cumu, precision, recall)

def writeResult(resultFilename, avgAcc, accFold, precision, recall, comps):
    fRes = open(resultFilename, 'a')
    for i in range(len(precision)):
        fRes.write('%s: Precision = %.2f, Recall = %.2f\n' \
                   % (comps[i], 100.0 * precision[i], 100.0 * recall[i]))
    fRes.write('Accuracy: %.2f (' % (100.0 * avgAcc))
    for acc1 in accFold: fRes.write('%.2f ' % (100.0 * acc1))
    fRes.write(')\n')
    fRes.close()

def writeResult1Fold(resultFilename, acc, precision, recall, fold, \
                     comps, title):
    fRes = open(resultFilename, 'a')
    fRes.write(title)
    for i in range(len(precision)):
        fRes.write('%s: Precision = %.2f, Recall = %.2f\n' \
                   % (comps[i], 100.0 * precision[i], 100.0 * recall[i]))
    fRes.write('Accuracy of fold %d: %.2f\n' % (fold, 100.0 * acc))
    fRes.close()

def svmLightReuse(comps, fold, randSeed, svmOpt = ''):
    result1Filename     = 'Yeast%d_%d.result' % (randSeed, fold)
    testFeatFilename    = 'Yeast%dte%d.dat'   % (randSeed, fold)
    trainFeatFile0      = 'Yeast%dtr%d.dat'   % (randSeed, fold)
    trainFeatFilename   = 'Yeast%dmc%d.dat'   % (randSeed, fold)
    modelFilename       = 'Yeast%d_%d.svm'    % (randSeed, fold)
    predFilename        = 'Yeast%dpr%d.dat'   % (randSeed, fold)

    (trainFiles,testFiles,memeFiles) = foldFilenames(comps, randSeed, fold)

    trainNSeq = []; testNSeq = []
    for seqFilename in trainFiles: trainNSeq.append(fastaNSeq(seqFilename))
    for seqFilename in testFiles: testNSeq.append(fastaNSeq(seqFilename))

    trainFeat = readFeatures(trainNSeq, trainFeatFile0)
    (corr1, nTestSeq1, precision, recall) = trainTest(trainFeat, \
            testNSeq, trainFeatFilename, testFeatFilename, modelFilename, \
            predFilename, svmOpt)
    writeResult1Fold(result1Filename, float(corr1) / nTestSeq1, precision, \
                     recall, fold, comps, 'svmLightReuse (%s)\n' % svmOpt)

def svmLight2Multiclass(comps, fold, randSeed, svmOpt = ''):
    result1Filename     = 'Yeast%d_%d.result' % (randSeed, fold)
    testFeatFilename    = 'Yeast%dte%d.dat'   % (randSeed, fold)
    trainFeatFile0      = 'Yeast%dtr%d.dat'   % (randSeed, fold)
    trainFeatFilename   = 'Yeast%dmc%d.dat'   % (randSeed, fold)
    modelFilename       = 'Yeast%d_%d.svm'    % (randSeed, fold)
    predFilename        = 'Yeast%dpr%d.dat'   % (randSeed, fold)

    (trainFiles,testFiles,memeFiles) = foldFilenames(comps,randSeed,fold)

    trainNSeq = []; testNSeq = []
    for seqFilename in trainFiles: trainNSeq.append(fastaNSeq(seqFilename))
    for seqFilename in testFiles: testNSeq.append(fastaNSeq(seqFilename))

    trainFeat = readFeatures(trainNSeq, trainFeatFile0)
    (corr1, nTestSeq1, precision, recall) = trainTestMulticlass(trainFeat, \
            testNSeq, trainFeatFilename, testFeatFilename, modelFilename, \
            predFilename, svmOpt)
    writeResult1Fold(result1Filename, float(corr1) / nTestSeq1, precision, \
                     recall, fold, comps, \
                     'svmLight2Multiclass (%s)\n' % svmOpt)

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2: print USAGE; sys.exit(1)
    seqPrefix = '~/subloc/data/hallett2004/Yeast'
    nFold = 10
    isMultiClass = False
    svmOpt = ''
    memeExt = 'meme'
    randSeed = 1000
    t0 = time.time()
    if   'test'  == sys.argv[1].lower():
        pwms = readMeme('YeastSecreted.meme')
        for i in range(len(pwms)):
            print 'Motif %d' % i
            for v in pwms[i]: print v
    elif 'split' == sys.argv[1].lower():
        if len(sys.argv) < 3: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        if len(sys.argv) > 3: seqPredix = sys.argv[3]
        if len(sys.argv) > 4: nFold = int(sys.argv[4])
        compSeqFiles = glob.glob('%s*' % seqPredix)
        splitSeqFiles(compSeqFiles, nFold, randSeed)
    elif 'feat'  == sys.argv[1].lower():
        if len(sys.argv) < 3: print USAGE; sys.exit(1)
        fold = int(sys.argv[2])
        seqDir = ''; memeDir = ''; featDir = ''
        if len(sys.argv) > 3: memeExt = sys.argv[3]
        if len(sys.argv) > 4: seqDir = sys.argv[4]
        if len(sys.argv) > 5: memeDir = sys.argv[5]
        if len(sys.argv) > 6: featDir = sys.argv[6]
        if len(sys.argv) > 7: randSeed = int(sys.argv[7])
        meme2Feat(fold, randSeed, memeExt, seqDir, memeDir, featDir)
    elif 'eval'  == sys.argv[1].lower():
        # Not throughly tested; use eval1 with a computer cluster instead
        # CONSIDER REMOVE THIS OPTION
        if len(sys.argv) < 3: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        if len(sys.argv) > 3: seqPrefix = sys.argv[3]
        if len(sys.argv) > 4: nFold = int(sys.argv[4])
        comps = seqFilenamesOrComps(seqPrefix)
        cvEvaluate(comps, nFold, randSeed, isMultiClass)
    elif 'eval1'  == sys.argv[1].lower():
        if len(sys.argv) < 5: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        comps = seqFilenamesOrComps(sys.argv[3]) #'Secreted,Pero,Lyso'
        fold = int(sys.argv[4])
        if len(sys.argv) > 5:
            if 'multiclass' == sys.argv[5]: isMultiClass = True
            elif '1-vs-all' == sys.argv[5]: isMultiClass = False
            else: print 'Fatal error: invalid classifier\n'+USAGE; sys.exit(-1)
        if len(sys.argv) > 6: memeExt = sys.argv[6]
        cvEvaluate1Fold(comps, fold, randSeed, isMultiClass, memeExt)
    elif 'svmlight'  == sys.argv[1].lower():
        if len(sys.argv) < 5: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        comps = seqFilenamesOrComps(sys.argv[3])
        fold = int(sys.argv[4])
        if len(sys.argv) > 5: svmOpt = sys.argv[5]
        svmLightReuse(comps, fold, randSeed, svmOpt)
    elif 'svmlight2multi'  == sys.argv[1].lower():
        if len(sys.argv) < 5: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        comps = seqFilenamesOrComps(sys.argv[3])
        fold = int(sys.argv[4])
        if len(sys.argv) > 5: svmOpt = sys.argv[5]
        svmLight2Multiclass(comps, fold, randSeed, svmOpt)
    elif 'test_light2multi'  == sys.argv[1].lower():
        trainFeatFile0 = 'Yeast1000tr0.dat'
        trainFeatFilename = 'Yeast1000mc0.dat'
        trainFiles = glob.glob('*tr0.fa')
        trainNSeq = []
        for seqFilename in trainFiles: trainNSeq.append(fastaNSeq(seqFilename))
        print 'trainNSeq:'; print trainNSeq
        trainFeat = readFeatures(trainNSeq, trainFeatFile0)
        writeFeatures(trainFeat, -2, trainFeatFilename)
    else: print USAGE; sys.exit(1)
    print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
