#!/usr/bin/python
import time, sys
from memecv import *

USAGE = \
'''Usage: memecvtree.py <feat|svm> SEED FOLD [MEME_EXT]
Examples:
memecvtree.py feat 1000 0 meme3
memecvtree.py svm  1000 0 meme3 "-t 2 -c 1..."
'''

def trainSVM(trainFeat, trainFeatFile, modelFiles, svmOpt = ''):
    if len(modelFiles) == 1:
        writeFeatures(trainFeat, 0, trainFeatFile)
        svmLearn(trainFeatFile, modelFiles[0], svmOpt)
        return
    for trainComp in range(len(trainFeat)):
        writeFeatures(trainFeat, trainComp, trainFeatFile)
        svmLearn(trainFeatFile, modelFiles[trainComp], svmOpt)

def testSVM(testFeatFile, modelFiles, predFile):
    '''testFeatFile remains the same for all target compartments;
    predFile is reused'''
    if len(modelFiles) == 1:
        svmClassify(testFeatFile, modelFiles[0], predFile)
        pred = svmLightReadOutput(predFile) #predFile reused
        predArgMax = [int(pred1 < 0) for pred1 in pred] #+/- into 0/1
        return predArgMax
    for comp in range(len(modelFiles)):
        svmClassify(testFeatFile, modelFiles[comp], predFile)
        pred = svmLightReadOutput(predFile) #predFile reused
        if comp == 0: # first, init as max
            predMax = pred; predArgMax = [0] * len(pred)
        else:
            for i in range(len(pred)):
                if pred[i] > predMax[i]:
                    predMax[i] = pred[i]; predArgMax[i] = comp
    return predArgMax

class CompTreeNode:
    def __init__(self, name, children = []):
        self.name = name
        self.children = children
        if len(children) == 0:
            self.descendants = [name]
        else:
            self.descendants = []
            for child in children:
                self.descendants = self.descendants + child.descendants

    def setFoldSeed(self, fold, randSeed, memeExt = 'meme'):
        self.fold = fold; self.randSeed = randSeed
        self.memeFile = '%s%dtr%d.%s' % (self.name, randSeed, fold, memeExt)
        if len(self.children) > 0:
            self.resultFile    = '%s%d_%d.result' % (self.name, randSeed, fold)
            self.trainFeatFile = '%s%dtr%d.dat'   % (self.name, randSeed, fold)
            self.testFeatFile  = '%s%dte%d.dat'   % (self.name, randSeed, fold)
            self.trainFeatMcFile='%s%dmc%d.dat'   % (self.name, randSeed, fold)
            self.predFile      = '%s%dpr%d.dat'   % (self.name, randSeed, fold)
            assert len(self.children) > 1, '%s only 1 child' % self.name
            if len(self.children) == 2: #binary not multiclass
                self.modelFiles = ['%s%d_%d.svm' \
                                   % (self.children[0].name, randSeed, fold)]
            else:
                self.modelFiles = ['%s%d_%d.svm' % (child.name, randSeed, fold)\
                                   for child in self.children]
        self.trainSeqFiles = []; self.testSeqFiles = []
        for comp in self.descendants: 
            self.trainSeqFiles.append('%s%dtr%d.fa' % (comp, randSeed, fold))
            self.testSeqFiles.append( '%s%dte%d.fa' % (comp, randSeed, fold))
        for child in self.children: child.setFoldSeed(fold, randSeed, memeExt)
    
    def meme2Feat(self, allTestSeqFiles = None):
        '''Scan training and testing seq for MEME motifs under this node,
        then write features to files.
        Note that the required feature vectors for training and testing
        is different.'''
        if len(self.children) == 0: return
        # Store test seq files if root
        if allTestSeqFiles == None: allTestSeqFiles = self.testSeqFiles
        # Organize filenames
        memeFiles = []; flatTrainFiles = []
        for child in self.children:
            memeFiles.append(child.memeFile)
            flatTrainFiles = flatTrainFiles + child.trainSeqFiles
        # Scan all seq files under this branch, using all meme files
        (flatTrainFeat, flatTestFeat) = memeFeatures(memeFiles, \
                                        flatTrainFiles, allTestSeqFiles)
        # Group flattened feature files into subtrees...
##        trainFeat = [[]]*len(self.children); testFeat = [[]]*len(self.children)
##        iChild = 0; nComp = 0
##        for trainFeat1, testFeat1 in zip(flatTrainFeat, flatTestFeat):
##            trainFeat[iChild] = trainFeat[iChild] + trainFeat1
##            testFeat[iChild] = testFeat[iChild] + testFeat1
##            nComp = nComp + 1
##            if nComp == len(self.children[iChild].descendants):
##                iChild = iChild + 1; nComp = 0
##        assert iChild == len(self.children) and nComp != 0, \
##            'iChild=%d != len(children)=%d or nComp=%d != 0' \
##            % (iChild, len(self.children)-1, nComp)
        # Then write subtrees as feature files, -1 means no training target yet
        writeFeatures(flatTrainFeat, -1, self.trainFeatFile)
        writeFeatures(flatTestFeat,  -1, self.testFeatFile)
        # Depth-first search is used (easier recursive implementation)
        for child in self.children: child.meme2Feat(allTestSeqFiles)

    def trainTest(self, svmOpt = ''):
        # count number of seq in each compartment
        self.countNSeq()
        # train
        self.train(svmOpt)
        # test
        nComp = len(self.descendants); nTestSeq = self.teSeqEnd - self.teSeqSt
        correct = [0] * nComp; precision = [.0] * nComp; recall = [.0] * nComp
        entering = [True] * nTestSeq
        self.test(entering, correct, precision, recall)
        accuracy = float(sum(correct)) / nTestSeq
        # write results
        writeResult1Fold(self.resultFile, accuracy, precision, recall, \
                         self.fold, self.descendants, 'svm %s\n' % svmOpt)

    def train(self, svmOpt):
        'Train SVM classifier'
        if len(self.children) == 0: return
        # read and group training feature files
        trainFeat = readFeatures(self.trNSeq, self.trainFeatFile)
        assert len(trainFeat) == len(self.children), \
               'len(trainFeat)=%d != len(children)=%d' \
               % (len(trainFeat), len(self.children))
        # train current node
        trainSVM(trainFeat, self.trainFeatMcFile, self.modelFiles, svmOpt)
        # train children
        for child in self.children: child.train(svmOpt)

    def test(self, entering, correct, precision, recall):
        'Evaluates predictions under this node and fills in given result-array.'
        if len(self.children) == 0:
            p = sum(entering)
            correct[self.iComp] = sum(entering[self.teSeqSt : self.teSeqEnd])
            tp = float(correct[self.iComp])
            if p > 0: precision[self.iComp] = tp / p
            else: precision[self.iComp] = 0.0
            recall[self.iComp] = tp / (self.teSeqEnd - self.teSeqSt)
            return
        predArgMax = testSVM(self.testFeatFile, self.modelFiles, self.predFile)
        for iChild in range(len(self.children)):
            chEnter = [enter1 and argMax == iChild for enter1, argMax \
                       in zip(entering, predArgMax)]
            self.children[iChild].test(chEnter, correct, precision, recall)
    
    def countNSeq(self, trainCumu = 0, testCumu = 0, compCumu = 0):
        '''Count #seq in each compartment, since feature files have no labels.
        Also calculate cumulative starting positions in the depth-first-search
        testing feature vectors. This is a recursive call, but outside caller
        can invoke using the default parameters (trainCumu = testCumu = 0).'''
        self.trSeqSt = trainCumu; self.teSeqSt = testCumu
        if len(self.children) == 0:
            self.trNSeq = [fastaNSeq(self.trainSeqFiles[0])]
            self.teNSeq = [fastaNSeq(self.testSeqFiles[0])]
            trainCumu = trainCumu + self.trNSeq[0]
            testCumu  = testCumu  + self.teNSeq[0]
            self.iComp = compCumu; compCumu = compCumu + 1
        else:
            for child in self.children:
                (trainCumu, testCumu, compCumu) = child.countNSeq(trainCumu, \
                                                  testCumu, compCumu)
            self.trNSeq = [ch.trSeqEnd - ch.trSeqSt for ch in self.children]
            self.teNSeq = [ch.teSeqEnd - ch.teSeqSt for ch in self.children]
        self.trSeqEnd = trainCumu; self.teSeqEnd = testCumu
        return (trainCumu, testCumu, compCumu)

def buildTree(fold, randSeed, memeExt):
    cyto = CompTreeNode('YeastCyto')
    er = CompTreeNode('YeastER')
    golgi = CompTreeNode('YeastGolgi')
    lyso = CompTreeNode('YeastLyso')
    mito = CompTreeNode('YeastMito')
    nuc = CompTreeNode('YeastNuclear')
    pero = CompTreeNode('YeastPero')
    pm = CompTreeNode('YeastPM')
    sec = CompTreeNode('YeastSecreted')
    secPw = CompTreeNode('YeastSecPw', [er, golgi, lyso, pm, sec])
    orga = CompTreeNode('YeastOrga', [cyto, pero, mito])
    intra = CompTreeNode('YeastIntra', [orga, nuc])
    root = CompTreeNode('YeastRoot', [secPw, intra])
    root.setFoldSeed(fold, randSeed, memeExt)
    return root

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2:
        #root = buildTree(0, 1000, 'meme'); root.countNSeq()
        print USAGE; sys.exit(1)
    
    #nFold = 10
    svmOpt = ''
    memeExt = 'meme'
    if   'feat' == sys.argv[1].lower():
        if len(sys.argv) < 4: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        fold = int(sys.argv[3])
        if len(sys.argv) > 4: memeExt = sys.argv[4]
        t0 = time.time()
        root = buildTree(fold, randSeed, memeExt)
        root.meme2Feat()
        print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
    elif 'svm' == sys.argv[1].lower():
        if len(sys.argv) < 4: print USAGE; sys.exit(1)
        randSeed = int(sys.argv[2])
        fold = int(sys.argv[3])
        if len(sys.argv) > 4: memeExt = sys.argv[4]
        if len(sys.argv) > 5: svmOpt = sys.argv[5]
        t0 = time.time()
        root = buildTree(fold, randSeed, memeExt)
        root.trainTest(svmOpt)
        print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
    else: print USAGE; sys.exit(1)
