#!/usr/bin/python
# Classification under a tree topology, trained by the same all (leaf) features.
# This provides a fair comparison with classification under flat hierarchy.

import sys, time
from memecv import sum, fastaNSeq, readFeatures, writeFeatures
from memecvtree import trainSVM, testSVM, writeResult1Fold

USAGE = \
'''Usage: memecvtree2.py FOLD [SVM_OPTIONS] [SEED=1000]
Example: memecvtree2.py 0 "-t 2 -c 1..."
'''

def trainTestLeaf(randSeed, fold, svmOpt):
    # filenames
    dfsCompList = ['YeastER', 'YeastGolgi', 'YeastLyso', 'YeastPM', 'YeastSecreted',
                   'YeastCyto', 'YeastPero', 'YeastMito', 'YeastNuclear']
    resultFile = 'YeastLeaf%d_%d.result' % (randSeed, fold)
    testFeatFile = 'YeastLeaf%dte%d.dat' % (randSeed, fold)
    rootTrFeatMcFile  = 'YeastRoot%dmc%d.dat'  % (randSeed, fold)
    secPwTrFeatMcFile = 'YeastSecPw%dmc%d.dat' % (randSeed, fold)
    intraTrFeatMcFile = 'YeastIntra%dmc%d.dat' % (randSeed, fold)
    orgaTrFeatMcFile  = 'YeastOrga%dmc%d.dat'  % (randSeed, fold)
    rootPredFile  = 'YeastRoot%dpr%d.dat'  % (randSeed, fold)
    secPwPredFile = 'YeastSecPw%dpr%d.dat' % (randSeed, fold)
    intraPredFile = 'YeastIntra%dpr%d.dat' % (randSeed, fold)
    orgaPredFile  = 'YeastOrga%dpr%d.dat'  % (randSeed, fold)
    rootModelFiles = ['YeastSecPw%d_%d.svm' % (randSeed, fold)]
    intraModelFiles = ['YeastOrga%d_%d.svm' % (randSeed, fold)]
    secPwModelFiles = ['%s%d_%d.svm' % (comp, randSeed, fold) \
                       for comp in dfsCompList[:5]]
    orgaModelFiles = ['%s%d_%d.svm' % (comp, randSeed, fold) \
                      for comp in dfsCompList[5:8]]

    # initialization    
    (flatTrFeat, flatTeFeat, testNSeq) = flatFeat(randSeed, fold)
    (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat) = trainFeat(flatTrFeat)
    # dfsTeFeat: Depth-First-Search Test Features
    dfsTeFeat = testFeat(flatTeFeat, testFeatFile, testNSeq)

    # train
    trainSVM(rootTrFeat,  rootTrFeatMcFile,  rootModelFiles,  svmOpt)
    trainSVM(secPwTrFeat, secPwTrFeatMcFile, secPwModelFiles, svmOpt)
    trainSVM(intraTrFeat, intraTrFeatMcFile, intraModelFiles, svmOpt)
    trainSVM(orgaTrFeat,  orgaTrFeatMcFile,  orgaModelFiles,  svmOpt)

    # test (classify)
    rootArgMax  = testSVM(testFeatFile, rootModelFiles,  rootPredFile )
    secPwArgMax = testSVM(testFeatFile, secPwModelFiles, secPwPredFile)
    intraArgMax = testSVM(testFeatFile, intraModelFiles, intraPredFile)
    orgaArgMax  = testSVM(testFeatFile, orgaModelFiles,  orgaPredFile )

    # evaluate and write result
    (accuracy, precision, recall) = evaluate(rootArgMax, secPwArgMax, \
                                             intraArgMax, orgaArgMax, dfsTeFeat)
    writeResult1Fold(resultFile, accuracy, precision, recall, fold, \
                     dfsCompList, 'svm %s\n' % svmOpt)

def flatFeat(randSeed, fold):
    compList = ['YeastCyto', 'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastMito',
                'YeastNuclear', 'YeastPero', 'YeastPM', 'YeastSecreted']
    trainSeqFiles = ['%s%dtr%d.fa' % (comp,randSeed,fold) for comp in compList]
    trainNSeq = [fastaNSeq(seqFile) for seqFile in trainSeqFiles]
    flatTrFeatFile = 'Yeast%dtr%d.dat' % (randSeed, fold)
    flatTrFeat = readFeatures(trainNSeq, flatTrFeatFile)
    testSeqFiles = ['%s%dte%d.fa' % (comp,randSeed,fold) for comp in compList]
    testNSeq = [fastaNSeq(seqFile) for seqFile in testSeqFiles]
    flatTeFeatFile = 'Yeast%dte%d.dat' % (randSeed, fold)
    flatTeFeat = readFeatures(testNSeq, flatTeFeatFile)
    return (flatTrFeat, flatTeFeat, testNSeq)

def trainFeat(flatTrFeat):
    secPwTrFeat = flatTrFeat[1:4] + flatTrFeat[7:9]
    orgaTrFeat = [flatTrFeat[0], flatTrFeat[6], flatTrFeat[4]]
    intraTrFeat = [sum(orgaTrFeat), flatTrFeat[5]]
    rootTrFeat = [sum(secPwTrFeat), sum(intraTrFeat)]
    return (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat)

def testFeat(flatTeFeat, testFeatFile, testNSeq):
    dfsTeFeat = flatTeFeat[1:4] + flatTeFeat[7:9] \
                + [flatTeFeat[0], flatTeFeat[6]] + flatTeFeat[4:6]
    writeFeatures(dfsTeFeat, -1, testFeatFile)
    # These length will be used in evaluation
    assert len(dfsTeFeat[0])==testNSeq[1], '%d!=%d'%(len(dfsTeFeat[0]),testNSeq[1])
    assert len(dfsTeFeat[1])==testNSeq[2], '%d!=%d'%(len(dfsTeFeat[1]),testNSeq[2])
    assert len(dfsTeFeat[2])==testNSeq[3], '%d!=%d'%(len(dfsTeFeat[2]),testNSeq[3])
    assert len(dfsTeFeat[3])==testNSeq[7], '%d!=%d'%(len(dfsTeFeat[3]),testNSeq[7])
    assert len(dfsTeFeat[4])==testNSeq[8], '%d!=%d'%(len(dfsTeFeat[4]),testNSeq[8])
    assert len(dfsTeFeat[5])==testNSeq[0], '%d!=%d'%(len(dfsTeFeat[5]),testNSeq[0])
    assert len(dfsTeFeat[6])==testNSeq[6], '%d!=%d'%(len(dfsTeFeat[6]),testNSeq[6])
    assert len(dfsTeFeat[7])==testNSeq[4], '%d!=%d'%(len(dfsTeFeat[7]),testNSeq[4])
    assert len(dfsTeFeat[8])==testNSeq[5], '%d!=%d'%(len(dfsTeFeat[8]),testNSeq[5])
    return dfsTeFeat

def evaluate(rootArgMax, secPwArgMax, intraArgMax, orgaArgMax, dfsTeFeat):
    # TP & P & T of all compartment
    # er,golgi,lyso,pm,sec
    correct = [0] * len(dfsTeFeat); nPred = [0] * len(dfsTeFeat)
    precision = [.0] * len(dfsTeFeat); recall = [.0] * len(dfsTeFeat)
    testCumu = 0
    for iComp in range(5):
        chEnter = [root1==0 and secPw1==iComp \
                   for root1, secPw1 in zip(rootArgMax, secPwArgMax)]
        teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
        nPred[iComp] = sum(chEnter)
        correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    # cyto,pero,mito
    for iComp in range(5, 8):
        chEnter = [root1==1 and intra1==0 and orga1==iComp-5 \
                   for root1, intra1, orga1 \
                   in zip(rootArgMax, intraArgMax, orgaArgMax)]
        teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
        nPred[iComp] = sum(chEnter)
        correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    # nuclear
    iComp = 8
    chEnter = [root1==1 and intra1==1 \
               for root1, intra1 in zip(rootArgMax, intraArgMax)]
    teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
    nPred[iComp] = sum(chEnter)
    correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    nTestSeq = testCumu

    # calculate precision / recall
    for iComp in range(len(dfsTeFeat)):
        tp = float(correct[iComp])
        if nPred[iComp] > 0: precision[iComp] = tp / nPred[iComp]
        else: precision[iComp] = 0.0
        recall[iComp] = tp / len(dfsTeFeat[iComp])
    accuracy = float(sum(correct)) / nTestSeq
    return (accuracy, precision, recall)
    
# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 2: print USAGE; sys.exit(1)
    fold = int(sys.argv[1])
    svmOpt = ''
    if len(sys.argv) > 2: svmOpt = sys.argv[2]
    randSeed = 1000
    if len(sys.argv) > 3: randSeed = int(sys.argv[3])
    
    t0 = time.time()
    trainTestLeaf(randSeed, fold, svmOpt)
    print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
