#!/usr/bin/python
# Classification under a tree topology, trained by the same all motifs,
# including internal and leaf motifs.

import sys, time
from memecv import sum, fastaNSeq, readFeatures, writeFeatures, memeFeatures
from memecvtree import trainSVM, testSVM, writeResult1Fold

USAGE = \
'''Usage: memecvtree3.py FOLD [MEMEEXT] [SVM_OPTIONS] [SEED]
Default: MEMEEXT = meme, SEED = 1000
Example:
    memecvtree3.py feat 0 meme3
    memecvtree3.py svm 0 meme  "-t 2 -c 1..."
    memecvtree3.py svm 0 meme3 "-t 2 -c 1..."
    memecvtree3.py svm 0
'''

nodeList = ['YeastSecPw', 'YeastIntra', 'YeastOrga',
            'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastPM', 'YeastSecreted',
            'YeastCyto', 'YeastPero', 'YeastMito', 'YeastNuclear']
dfsCompList = nodeList[3:]

def foldFilenames(randSeed, fold, memeExt):
    memeFiles = ['%s%dtr%d.%s' % (name, randSeed, fold, memeExt) \
                 for name in nodeList]
    trainSeqFiles = ['%s%dtr%d.fa' % (name, randSeed, fold) \
                     for name in dfsCompList]
    testSeqFiles  = ['%s%dte%d.fa' % (name, randSeed, fold) \
                     for name in dfsCompList]
    trainFeatFile = 'YeastAllMot%s%dtr%d.dat' % (memeExt, randSeed, fold)
    testFeatFile  = 'YeastAllMot%s%dte%d.dat' % (memeExt, randSeed, fold)
    return (memeFiles, trainSeqFiles, testSeqFiles, trainFeatFile, testFeatFile)

def meme2AllFeat(randSeed, fold, memeExt):
    (memeFiles, trainSeqFiles, testSeqFiles, trainFeatFile, testFeatFile) \
                = foldFilenames(randSeed, fold, memeExt)
    (trainFeat, testFeat) = memeFeatures(memeFiles, trainSeqFiles, testSeqFiles)
    writeFeatures(trainFeat, -1, trainFeatFile)
    writeFeatures(testFeat,  -1, testFeatFile)

def trainTestAllFeat(randSeed, fold, memeExt, svmOpt):
    # filenames
    resultFile = 'YeastAllMot%s%d_%d.result' % (memeExt, randSeed, fold)
    rootTrFeatMcFile  = 'YeastRoot%dmc%d.dat'  % (randSeed, fold)
    secPwTrFeatMcFile = 'YeastSecPw%dmc%d.dat' % (randSeed, fold)
    intraTrFeatMcFile = 'YeastIntra%dmc%d.dat' % (randSeed, fold)
    orgaTrFeatMcFile  = 'YeastOrga%dmc%d.dat'  % (randSeed, fold)
    rootPredFile  = 'YeastRoot%dpr%d.dat'  % (randSeed, fold)
    secPwPredFile = 'YeastSecPw%dpr%d.dat' % (randSeed, fold)
    intraPredFile = 'YeastIntra%dpr%d.dat' % (randSeed, fold)
    orgaPredFile  = 'YeastOrga%dpr%d.dat'  % (randSeed, fold)
    rootModelFiles = ['YeastSecPw%d_%d.svm' % (randSeed, fold)]
    intraModelFiles = ['YeastOrga%d_%d.svm' % (randSeed, fold)]
    secPwModelFiles = ['%s%d_%d.svm' % (comp, randSeed, fold) \
                       for comp in dfsCompList[:5]]
    orgaModelFiles = ['%s%d_%d.svm' % (comp, randSeed, fold) \
                      for comp in dfsCompList[5:8]]

    # initialization    
    # dfsTrFeat/dfsTeFeat: Depth-First-Search Train/Test Features
    (memeFiles, trainSeqFiles, testSeqFiles, trainFeatFile, testFeatFile) \
                = foldFilenames(randSeed, fold, memeExt)
    trainNSeq = [fastaNSeq(seqFile) for seqFile in trainSeqFiles]
    testNSeq  = [fastaNSeq(seqFile) for seqFile in testSeqFiles ]
    dfsTrFeat = readFeatures(trainNSeq, trainFeatFile)
    dfsTeFeat = readFeatures(testNSeq , testFeatFile )
    (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat) = trainFeat(dfsTrFeat)

    # train
    trainSVM(rootTrFeat,  rootTrFeatMcFile,  rootModelFiles,  svmOpt)
    trainSVM(secPwTrFeat, secPwTrFeatMcFile, secPwModelFiles, svmOpt)
    trainSVM(intraTrFeat, intraTrFeatMcFile, intraModelFiles, svmOpt)
    trainSVM(orgaTrFeat,  orgaTrFeatMcFile,  orgaModelFiles,  svmOpt)

    # test (classify)
    rootArgMax  = testSVM(testFeatFile, rootModelFiles,  rootPredFile )
    secPwArgMax = testSVM(testFeatFile, secPwModelFiles, secPwPredFile)
    intraArgMax = testSVM(testFeatFile, intraModelFiles, intraPredFile)
    orgaArgMax  = testSVM(testFeatFile, orgaModelFiles,  orgaPredFile )

    # evaluate and write result
    (accuracy, precision, recall) = evaluate(rootArgMax, secPwArgMax, \
                                             intraArgMax, orgaArgMax, dfsTeFeat)
    writeResult1Fold(resultFile, accuracy, precision, recall, fold, \
                     dfsCompList, 'svm %s\n' % svmOpt)

def trainFeat(dfsTrFeat):
    secPwTrFeat = dfsTrFeat[:5]
    orgaTrFeat = dfsTrFeat[5:8]
    intraTrFeat = [sum(orgaTrFeat), dfsTrFeat[8]]
    rootTrFeat = [sum(secPwTrFeat), sum(intraTrFeat)]
    return (rootTrFeat, secPwTrFeat, intraTrFeat, orgaTrFeat)

def evaluate(rootArgMax, secPwArgMax, intraArgMax, orgaArgMax, dfsTeFeat):
    # TP & P & T of all compartment
    # er,golgi,lyso,pm,sec
    correct = [0] * len(dfsTeFeat); nPred = [0] * len(dfsTeFeat)
    precision = [.0] * len(dfsTeFeat); recall = [.0] * len(dfsTeFeat)
    testCumu = 0
    for iComp in range(5):
        chEnter = [root1==0 and secPw1==iComp \
                   for root1, secPw1 in zip(rootArgMax, secPwArgMax)]
        teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
        nPred[iComp] = sum(chEnter)
        correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    # cyto,pero,mito
    for iComp in range(5, 8):
        chEnter = [root1==1 and intra1==0 and orga1==iComp-5 \
                   for root1, intra1, orga1 \
                   in zip(rootArgMax, intraArgMax, orgaArgMax)]
        teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
        nPred[iComp] = sum(chEnter)
        correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    # nuclear
    iComp = 8
    chEnter = [root1==1 and intra1==1 \
               for root1, intra1 in zip(rootArgMax, intraArgMax)]
    teSeqSt = testCumu; testCumu = testCumu + len(dfsTeFeat[iComp])
    nPred[iComp] = sum(chEnter)
    correct[iComp] = sum(chEnter[teSeqSt : testCumu])
    nTestSeq = testCumu

    # calculate precision / recall
    for iComp in range(len(dfsTeFeat)):
        tp = float(correct[iComp])
        if nPred[iComp] > 0: precision[iComp] = tp / nPred[iComp]
        else: precision[iComp] = 0.0
        recall[iComp] = tp / len(dfsTeFeat[iComp])
    accuracy = float(sum(correct)) / nTestSeq
    return (accuracy, precision, recall)
    
# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 3: print USAGE; sys.exit(1)
    svmOpt = ''
    memeExt = 'meme'
    randSeed = 1000
    fold = int(sys.argv[2])
    if len(sys.argv) > 3: memeExt = sys.argv[3]
    if len(sys.argv) > 4: svmOpt = sys.argv[4]
    if len(sys.argv) > 5: randSeed = int(sys.argv[5])
    
    if   'feat' == sys.argv[1].lower():
        t0 = time.time()
        meme2AllFeat(randSeed, fold, memeExt)
        print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
    elif 'svm' == sys.argv[1].lower():
        t0 = time.time()
        trainTestAllFeat(randSeed, fold, memeExt, svmOpt)
        print 'Elapsed time: %.0f min' % ((time.time() - t0) / 60)
    else: print USAGE; sys.exit(1)
