import sys, re

from ipr2motif import COUNT_TH, LEN_TH

# Evaluation paramters
paramOverlapRatio       = .50
paramRecallThreshold    = .50
paramPrecThreshold      = .50

USAGE = '''motif2hit.py RESULT_FILE MOTIF_PAT MOTPOS_PAT
Evaluate predicted motif instances.

MOTIF_PAT and MOTPOS_PAT use the %s symbol,
e.g. ~/subloc/result/hallett2005/%s.iprpos.'''

COMPS = ['YeastCyto', 'YeastER', 'YeastGolgi', 'YeastLyso', 'YeastMito',
         'YeastNuclear', 'YeastPero', 'YeastPM', 'YeastSecreted']

def scanAllMemeWithRank(motPosFiles, memeFiles, resultFile, comps):
    '''Evaluate MEME-format motif-instances files, by found known motifs
    and their ranking. Process all compartments and write to a result file.'''
    compResults = []
    for motPosFile, memeFile in zip(motPosFiles, memeFiles):
        compResults.append(matchMemeWithRank(motPosFile, memeFile))
    title = 'OverlapRatio = %g, RecallThreshold = %g' \
            % (paramOverlapRatio, paramRecallThreshold)
    writeResultWithRank(title, compResults, comps, resultFile)

def matchMemeWithRank(motPosFile, memeFile):
    'Process one compartment'
    motNameInstances = readMotPos(motPosFile)
    if len(motNameInstances) == 0: return []
##    print motNameInstances
    memeMotifs = readMemeSitePos(memeFile)
    return matchMotifsRank(motNameInstances, memeMotifs)

def matchMotifsRank(motNameInstances, memeMotifs):
    'Match found motifs to true positions. Store the ranking.'
    foundResults = []; foundList = []
    for motName, motIns in motNameInstances:
        motInsIdDict = dict([(trueIns[0], 1) for trueIns in motIns])
        if len(motInsIdDict) < COUNT_TH: continue
        for i in range(len(memeMotifs)):
            if i in foundList: continue
            if matchMotif(foundResults, motName, motIns, memeMotifs[i]):
                foundResults[-1] = foundResults[-1] + (i + 1,)
                foundList.append(i)
                break
    return foundResults

def matchMotif(foundResults, motName, motInstances, memeMotif):
    '''Match intances of one motif type; return true and append to foundResults
    if recall above threshold.'''
    activeMotif = memeMotif[:]
    tp1 = 0
    #print 'overlap=%f' % ((motInstances[0][3]-motInstances[0][2]) * paramOverlapRatio)
    seqIdList = []
    for trueIns in motInstances:
        #print '[%s]' % trueIns[0]
        for iIns in range(len(activeMotif)):
            if not trueIns[0] in seqIdList: seqIdList.append(trueIns[0])
            if matchMotifInstance(trueIns,activeMotif[iIns]):
                tp1 = tp1 + 1; del activeMotif[iIns]; break
    #if tp1 > 0: print '%s found TP = %d' % (motName, tp1)
    if len(seqIdList) == 0 or len(memeMotif) == 0: return False
    if float(tp1) / len(seqIdList) >= paramRecallThreshold or \
       float(tp1) / len(memeMotif) >= paramPrecThreshold :
        foundResults.append((motName, tp1, len(memeMotif), len(seqIdList)))
        return True

def matchMotifInstance(trueIns, ins, overlapRatio = paramOverlapRatio):
    'Match one motif instance; return true if overlap is larger than given bases.'
    if ins[0] != trueIns[0]: return False
    return min(ins[2], trueIns[2]) - max(ins[1], trueIns[1]) \
           > min(trueIns[2] - trueIns[1], ins[2] - ins[1]) * overlapRatio

def writeResultWithRank(title, compResults, comps, resultFile):
    fRes = open(resultFile, 'a')
    fRes.write('%s\n' % title)
    fRes.write('COMPARTMENT\tMOTIFNAME\tTP\tPRED\tTRUENUM\tPREC\tRECALL\tRANK\n')
    print 'COMPARTMENT\tMOTIFNAME\tTP\tPRED\tTRUENUM\tPREC\tRECALL\tRANK'
    nFound = 0; sumTP = 0; sumP = 0; sumT = 0;
    for comp, foundResults in zip(comps, compResults):
        nFound = nFound + len(foundResults)
        for motName, tp1, p1, t1, rank in foundResults:
            line = '%-13s\t%-9s\t%d\t%d\t%d\t%.2f\t%.2f\t%d' \
                   % (comp, motName, tp1, p1, t1, 100.*tp1/p1, 100.*tp1/t1, rank)
            print line; fRes.write(line + '\n')
            sumTP = sumTP + tp1; sumP = sumP + p1; sumT = sumT + t1
    totalPrec = 0.
    if sumP > 0: totalPrec = 100.*sumTP/sumP
    totalRec = 0.
    if sumT > 0: totalRec = 100.*sumTP/sumT
    summary = '%d KNOWN-MOTIFS(%s): \t%d\t%d\t%d\t%.2f\t%.2f' \
              % (nFound, title, sumTP, sumP, sumT, totalPrec, totalRec)
    print summary
    fRes.write(summary + '\n')
    fRes.close()

def readMotPos(posFile):
    '''Read motif position file.
    Return format: [(name1, [(seqId,st,end),(), ...]), (name2,trueInstances)...]
    scanregex.readMotPos() is depreciated.'''
    f = open(posFile)
    motNameInstances = []
    f.readline() # Column descriptions
    currName = ''
    while True:
        tokens = f.readline().split()
        if len(tokens) == 0 or tokens[0] != currName:
            if currName!='' and len(dict([(id,1) for id,s,e in currIns]))>=COUNT_TH:
                motNameInstances.append((currName, currIns))
            if len(tokens) == 0: break
            currName = tokens[0]; currIns=[] #; currValid = True
        seqId = tokens[2]; st = int(tokens[4]); end = int(tokens[5])
        if end - st <= LEN_TH: #currValid = False
            currIns.append((seqId, st, end))
    f.close()
    return motNameInstances

RE_MEME_SITES = re.compile(r'Motif (\w+) sites')

def readMemeSitePos(memeFilename):
    '''Read motif instances in MEME output.
    Modified from meme_motif.readMemeMotIns().
    Ex. motifs = readMemeSitePos()
    (seqId, st, end, site) = motifs[iMotif][iInstance]
    Note that "end" is given in python style: site = seq[st:end]'''
    f = open(memeFilename)
    motifs = []
    while True:
        line = f.readline()
        if line == '': break
        if RE_MEME_SITES.search(line):
            #section "Motif X sites"
            f.readline(); f.readline(); f.readline()
            motif1Pos = []
            while True: #each sequence or instance
                line = f.readline()
                if line.startswith('--'): break
                tokens = line.split()
                assert len(tokens) >= 4
                seqId = tokens[0]
                st = int(tokens[1]) - 1
                sites = line[53:].split()[0]
                motif1Pos.append((seqId, st, st + len(sites), sites))
            motifs.append(motif1Pos)
    f.close()
    return motifs

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) != 4: print USAGE; sys.exit(1);   
    resultFile = sys.argv[1]
    memeFilePattern = sys.argv[2]
    motPosFilePattern = sys.argv[3]

    motPosFiles = [motPosFilePattern % comp for comp in COMPS]
    memeFiles = [memeFilePattern % comp for comp in COMPS]
    scanAllMemeWithRank(motPosFiles, memeFiles, resultFile, COMPS)
