'''motifrank2alignhit.py - Select motifs and conserved sites (proteins)
Related to motif2hit.py
'''

import sys, re
from memecv import readFeatures
from motif2hit import COMPS, matchMemeWithRank, readMemeSitePos

COMPNAMES = ['Cyto', 'ER', 'Golgi', 'Lyso', 'Mito', 'Nuclear', 'Pero', 'PM', \
             'Secreted']

ALIGNMENT_HEAD = \
'''--------------------------------------------------------------------------------
        %d Motif %s sites %s
--------------------------------------------------------------------------------
Sequence name             Start LogLikeli             Site 
-------------             ----- ---------            ------
'''

ALIGNMENT_BOTTOM = \
'--------------------------------------------------------------------------------\n\n'

USAGE = \
'''motifrank2hit FOLDS SUMFILE ANNOPAT ALIGNPAT RANKPAT FEATPAT MOTIFPAT [SEQPAT] [MOTPOSPAT]
Annotate ranked motif.

FOLDS: the number of folds, or 1 if only one file
SUMFILE: the summary file for matlab to read
ANNOPAT: the output annotated ranking pattern (folds)
ALIGNPAT: the output conserved alignment pattern (folds)
RANKPAT: the ranking pattern (folds)
FEATPAT: the feature pattern (folds)
MOTIF_PAT: the motif file pattern (compartments and folds)
SEQ_PAT: the sequence file pattern (compartments and folds)
MOTPOS_PAT: the known motif position file pattern (compartments)

Default directories are current directory.
All file patterns use the %s symbol.

Example:
motifrank2hit.py $1/Yeast1000_%d.annoipr $1/Yeast1000_%d.sites
$1/Yeast1000_%d.result features/Yeast1000_%d.dat motif/%s1000tr%d.submotif
~/subloc/data/hallett2005.single2/%s1000tr%d.fa
~/subloc/result/hallett2005/%s.iprpos
'''

def motifrank2alignhit(sumFile, rankFiles, annoRankFiles, alignFiles, \
                       featFiles, motifFiles, motPosFiles, comps):
    fSummary = open(sumFile, 'w')
    for rankFile, annoRankFile, alignFile, featFile, motifFiles1 \
        in zip(rankFiles,annoRankFiles,alignFiles,featFiles,motifFiles):
        motifRank = readRank(rankFile, comps)
        compResults = []
        for motPosFile, motifFile in zip(motPosFiles, motifFiles1):
            compResults.append(matchMemeWithRank(motPosFile, motifFile))
        annotateMotifRanks(motifRank, compResults, comps)
        compMotifSites = [readMemeSitePos(motifFile) for motifFile in motifFiles1]
        for motifSites in compMotifSites:
            #each motifs should have the same number of proteins,
            #which is the compartment size
            for site in motifSites[1:]: assert len(site) == len(motifSites[0])
        compNSeq = [len(motifSites[0]) for motifSites in compMotifSites]
        compFeat = readFeatures(compNSeq, featFile)

        print '%s:' % annoRankFile
        writeAnnoRankAlign(motifRank, compMotifSites, compFeat, annoRankFile, \
                           alignFile)
        writeKnownVsRank(motifRank, fSummary)
        print
    fSummary.close()

def readRank(rankFile, comps):
    motifRank = []
    f = open(rankFile)
    i = 0
    for line in f:
        i = i + 1
        motifCode = int(line.split()[0])
        iComp = motifCode / 10
        order = motifCode - 10 * iComp
        motifRank.append((i, iComp, comps[iComp], order + 1))
    f.close()
    return motifRank

def annotateMotifRanks(motifRank, compResults, comps):
    for i in range(len(motifRank)): motifRank[i] = motifRank[i] + (None,) * 4
    for comp, foundResults in zip(comps, compResults):
        for foundResult in foundResults:
            order = foundResult[-1]
            isMapped = False
            for i in range(len(motifRank)):
                if motifRank[i][2] == comp and motifRank[i][3] == order:
                    motifRank[i] = motifRank[i][:4] + foundResult[:-1]
                    isMapped = True; break
            if not isMapped: print motifRank; print foundResults; print foundResult
            assert isMapped

def writeAnnoRankAlign(motifRank, compMotifSites, compFeat, annoRankFile, \
                       alignFile):
    'Write annotated ranking file and the conserved alignment file.'
    annoFp = open(annoRankFile, 'w'); alignFp = open(alignFile, 'w')
    line = 'No %-16s%-11s%5s%5s%5s%5s%5s' % ('MOTIF', 'KNOWN', 'PREC', 'REC',\
                                             'TP', 'PRED', 'TRUE')
    print line; annoFp.write(line + '\n')
    for i, iComp, comp, order, motName, tp1, p1, t1 in motifRank:
        (nSite, nCompSeq) = writeAlignment(alignFp, compMotifSites, compFeat, \
                                           iComp, comp, order, i, motName)
        annoLine = '%02d %-16s' % (i, comp + str(order))
        if motName:
            annoLine = annoLine + '%-11s%5.1f%5.1f%5d%5d%5d' \
                       % (motName, 100.*tp1/p1, 100.*tp1/t1, tp1, p1, t1)
        annoFp.write(annoLine + '\n')
        print '%-58spresence%4d/%4d' % (annoLine, nSite, nCompSeq)
    alignFp.close(); annoFp.close()

def writeAlignment(alignFp, compMotifSites, compFeat, iComp, comp, order, rank,\
                   motName):
    '''Filter conserved instances and write to the file.
    Note that order is the order of initial finding of a motif;
    rank is the ranking of discriminative power.'''
    motifSite = compMotifSites[iComp][order - 1]
    if motName: motName = '(' + motName + ')'
    else: motName = ''
    alignFp.write(ALIGNMENT_HEAD % (rank, comp + str(order), motName))
    nSite = 0
    iFeat = iComp * 10 + order - 1
    maxLike = max([featVec[iFeat] for featVec in compFeat[iComp]])
    for featVec, ins in zip(compFeat[iComp], motifSite):
        if featVec[iFeat] >= min(maxLike - 2., 2):
            alignFp.write('%-25s %5d %9.2e            %s\n' \
                          % (ins[0], ins[1] + 1, featVec[iFeat], ins[3]))
            nSite = nSite + 1
    alignFp.write(ALIGNMENT_BOTTOM)
    return (nSite, len(motifSite))

def writeKnownVsRank(motifRank, fSummary):
    strs = []
    for i, iComp, comp, order, motName, tp1, p1, t1 in motifRank:
        if motName: strs.append('1')
        else: strs.append('0')
    fSummary.write('\t'.join(strs) + '\n')

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) < 5: print USAGE; sys.exit(1)
    motPosPrefix = '%s.motpos'
    nFold = int(sys.argv[1])
    sumFile = sys.argv[2]
    annoRankPattern = sys.argv[3]
    alignPattern = sys.argv[4]
    rankPattern = sys.argv[5]
    featPattern = sys.argv[6]
    motifPattern = sys.argv[7]
    #if len(sys.argv) > 8: seqPattern = sys.argv[8]
    if len(sys.argv) > 9: motPosPattern = sys.argv[9]

    comps = COMPS # tree would be different
    if nFold <= 1:
        rankFiles = [rankPattern]
        annoRankFiles = [annoRankPattern]
        alignFiles = [alignPattern]
        featFiles = [featPattern]
        motifFiles = [[motifPattern % c for c in comps]]
        motPosFiles = [motPosPattern % c for c in comps]
        motifrank2alignhit(sumFile, rankFiles, annoRankFiles, alignFiles, \
                           featFiles, motifFiles, motPosFiles, comps)
        sys.exit(0)
    rankFiles = [rankPattern % f for f in range(nFold)]
    annoRankFiles = [annoRankPattern % f for f in range(nFold)]
    alignFiles = [alignPattern % f for f in range(nFold)]
    featFiles = [featPattern % f for f in range(nFold)]
    motifFiles = [[motifPattern % (c,f) for c in comps] for f in range(nFold)]
    motPosFiles = [motPosPattern % c for c in comps]
    motifrank2alignhit(sumFile, rankFiles, annoRankFiles, alignFiles, featFiles,\
                       motifFiles, motPosFiles, comps)
