import sys, getopt, time

from motif2hit import COMPS, COUNT_TH, matchMotifInstance, readMotPos, writeResultWithRank

# Evaluation paramters, default values
paramPredToTrue         = 4
paramOverlapRatio       = .5
paramRecallThreshold    = .33

USAGE = '''posterior2hit.py [-options] RESULT_FILE POSTERIOR_PAT MOTPOS_PAT
Evaluate predicted motif instances.

POSTERIOR_PAT and MOTPOS_PAT use the %s symbol,
e.g. ~/subloc/result/hallett2005/%s.iprpos.

Options:
   -p <Pred-to-true>    : prediction-to-true ratio
   -o <Overlap>         : overlap threshold
   -r <Recall>          : recall threshold        '''

def allPostToHit(motPosFiles, postFiles, resultFile, comps):
    compResults = []
    for motPosFile, postFile in zip(motPosFiles, postFiles):
        print 'Reading known ... ',
        motNameInstances = readMotPos(motPosFile)
        if len(motNameInstances) == 0:
            print 'None'
            compResults.append([]); continue
        print 'Reading %s... ' % postFile,
        sortedMotPos = readPosterior(postFile)
        print 'Matching...'
        compResults.append(postToHit(motNameInstances, sortedMotPos))
    title = 'Pred/true=%d,Recall>=%g,Overlap>=%g' \
            % (paramPredToTrue, paramRecallThreshold, paramOverlapRatio)
    writeResultWithRank(title, compResults, comps, resultFile)

def postToHit(motNameInstances, sortedMotPos):
    'Match found motifs to true positions. Store the ranking.'
    foundResults = []; foundList = []
    for motName, motIns in motNameInstances:
        motInsIdDict = dict([(trueIns[0], 1) for trueIns in motIns])
        if len(motInsIdDict) < COUNT_TH: continue
        for i in range(len(sortedMotPos)):
            if i in foundList: continue
            if matchMotifPost(foundResults, motName, motIns, sortedMotPos[i]):
                foundResults[-1] = foundResults[-1] + (i + 1,)
                foundList.append(i)
                break
    return foundResults

def matchMotifPost(foundResults, motName, motInstances, postMotif):
    '''Match intances of one motif type to the posterior of a candidate motif.
    Return true and append to foundResults if prec-rec above threshold.'''
    postMotif = postMotif[: len(motInstances) * paramPredToTrue]
    activeMotif = postMotif[:] #make a copy
    tp1 = 0
    for trueIns in motInstances:
        for iIns in range(len(activeMotif)):
            if matchMotifInstance(trueIns,activeMotif[iIns], paramOverlapRatio):
                tp1 = tp1 + 1; del activeMotif[iIns]; break
    if len(motInstances) == 0 or len(postMotif) == 0: return False
    if float(tp1) / len(motInstances) >= paramRecallThreshold:
        foundResults.append((motName, tp1, len(postMotif), len(motInstances)))
        return True

def readPosterior(postFile):
    #Read file and store posterior in a list
    f = open(postFile)
    motSeqPost = []; seqPost = None
    for line in f:
        if line.startswith('Motif'):
            if seqPost: motSeqPost.append(seqPost)
            assert int(line.split()[1]) == len(motSeqPost) + 1
            seqPost = []; continue
        tokens = line.split()
        seqId = tokens[1]
        begin = [float(c) for c in tokens[2]]
        end   = [float(c) for c in tokens[3]]
        seqPost.append((seqId, begin, end))
        assert len(tokens[2]) == len(tokens[3])
    motSeqPost.append(seqPost)
    f.close()

    #From posterior cruve to sorted motif instances
    print 'Sorting... ',
    sortedMotPos = []
    for seqPost in motSeqPost:
        posList = []
        for seqId, begin, end in seqPost:
            for i in range(len(begin)):
                if begin[i] == 0: continue
                for j in range(i+1, len(begin)):
                    if end[j] > 0: break
                    #elif begin[j] > 0: break #?
                if end[j] > 0:
                    #Set sites to None; caller can fill in later
                    posList.append(((begin[i]+.5)*(end[j]+.5)/100., seqId, \
                                    i+1, j+1, None))
                #elif begin[i]>1: print 'Motif %d: %s[%d : ?] = %.0f'%(len(sortedMotPos)+1,seqId,i+1,begin[i])
        posList.sort(); #using the first element, posterior prob
        posList.reverse()
        sortedMotPos.append([x[1:]+x[:1] for x in posList])
    return sortedMotPos

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) == 2: #TEST!
        t0 = time.time()
        sortedMotPos = readPosterior(sys.argv[1])
        for iMotif in range(len(sortedMotPos)):
            print 'Motif %d' % iMotif
            for pos in sortedMotPos[iMotif][:10]:
                if pos[-1] > .02: print pos[2]-pos[1], pos
            print
        print 'Elapsed time: %.0f sec' % (time.time() - t0)
        sys.exit();

    try:
        (opts, args) = getopt.getopt(sys.argv[1:], 'p:o:r:')
    except getopt.GetoptError:
        print USAGE; sys.exit(2)
    for o, a in opts:
        if   o == '-p': paramPredToTrue      = int(a)
        elif o == '-o': paramOverlapRatio    = float(a)
        elif o == '-r': paramRecallThreshold = float(a)
    if len(args) != 3: print USAGE; sys.exit(2)
    (resultFile, postFilePattern, motPosFilePattern) = args

    t0 = time.time()
    postFiles = [postFilePattern % comp for comp in COMPS]
    motPosFiles = [motPosFilePattern % comp for comp in COMPS]
    allPostToHit(motPosFiles, postFiles, resultFile, COMPS)
    print 'Elapsed time: %.0f sec' % (time.time() - t0)
