# pslt2motif.py
# This script attemps to find discriminative motifs base on the PSLT2 predictor.

import sys, xml.dom.minidom

USAGE = '''pslt2motif.py [TOPNUM] [PSLT2_XML]
Arguments:
  TOPNUM    number of top motifs selected (default: 10)
  PSLT2_XML pslt2 motif sets model file
            (default: ~/subloc/PSLT2/LocalizationLikelihoods.xml)
'''

NCOMP = 9

COMPNAMES = ['YeastER','YeastGolgi','YeastCyto','YeastNuclear','YeastPero',\
             'YeastPM','YeastLyso','YeastMito','YeastSecreted']

def pslt2motif(psltFile, interproFile, motifFile, Top):
    'The main process. List top motifs based on the PSLT2 model.'
    psltMotifSets = readPsltModel(psltFile)
    interproMap = readInterProMap(interproFile)
    for comp in range(NCOMP):
        motifScores = countMotif(psltMotifSets[comp])
        topMotifNames = selectTopMotif(motifScores, interproMap, nTop)
        listTopMotifs(topMotifNames, COMPNAMES[comp], motifFile)

def readInterProMap(interproFile):
    'Read motif number to interpro ID map.'
    interproMap = {}
    f = open(interproFile)
    for line in f:
        tokens = line.split()
        if len(tokens) > 1:
            assert len(tokens) == 2
            interproMap[int(tokens[0])] = tokens[1]
    f.close()
    return interproMap

def readPsltModel(psltFile):
    'Read the PSLT2 model XML and return the cond prob arrays.'
    doc = xml.dom.minidom.parse(psltFile)
    rootNode = doc.getElementsByTagName('pslt2')[0]
    compNodes = rootNode.getElementsByTagName('compartment')
    assert len(compNodes) == NCOMP
    psltMotifSets = []
    for compNode in compNodes:
        compMotifSets = []
        for motifSetNode in compNode.getElementsByTagName('motif_set'):
            motifsStr = motifSetNode.getAttribute('ids')
            motifs = [int(motif) for motif in motifsStr.split(',')[:-1]]
            condProb = []; i = 1
            for tableNode in motifSetNode.getElementsByTagName('table'):
                assert tableNode.getAttribute('index') == str(i), \
                       'index=%s, i=%d' % (tableNode.getAttribute('index'), i)
                i = i + 1
                condProb.append(float(tableNode.getAttribute(\
                                      'conditional_probability')))
            compMotifSets.append((motifs, condProb))
        psltMotifSets.append(compMotifSets)
    return psltMotifSets

def countMotif(compMotifSets):
    'Score the motifs of one compartment based on the PSLT2 model.'
    motifScores = {}
    for motifs, condProb in compMotifSets:
        for motif in motifs:
            if not motif in motifScores: motifScores[motif] = 0.0
        for i in range(1, 1 + len(condProb)):
            powJ = 1
            for motif in motifs:
                if powJ & i:
                    motifScores[motif] = motifScores[motif] + condProb[i - 1]
                powJ = powJ << 1
    return motifScores.items()

def selectTopMotif(motifScores, interproMap, nTop):
    '''Select and/or sort motif scores, and map to InterPro ID.
    Only motifs with an interpro mapping are counted.'''
    motifScores.sort(lambda x,y: cmp(y[1], x[1])) #high to low
    topMotifNames = []
    for motifScore in motifScores:
        if motifScore[0] in interproMap:
            topMotifNames.append(interproMap[motifScore[0]])
            if len(topMotifNames) >= nTop: break
    return topMotifNames

def writeTopMotifs(topMotifNames, compName, motifFile):
    'List top motifs.'
    #for motifScore in topMotifScores: print motifScore,
    f = open(motifFile, 'w')
    f.write(compName)
    for motifName in topMotifNames: f.write('\t' + motifName)
    f.write('\n')
    f.close()

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) > 1 and sys.argv[1] == '--help': print USAGE; sys.exit(1)
    nTop = 10
    psltFile = '../PSLT2/LocalizationLikelihoods.xml'
    interproFile = '../PSLT2/MapIDInterPro.txt'
    motifFile = '../PSLT2/pslt90.motif'
    if len(sys.argv) > 1: nTop = int(sys.argv[1])
    if len(sys.argv) > 2: psltFile = sys.argv[2]
    if len(sys.argv) > 3: motifFile = sys.argv[3]
    pslt2motif(psltFile, interproFile, motifFile, nTop)
