#!/usr/bin/env python
# Convert Murphy's labeled file format (separated by //) into FASTA format
# Split each labels (compartments) into different files
# Split into K-fold cross validation training-testing files
# Note: when the label line include commas, only the first label is used

import sys, random

USAGE = '''Usage:
seq2cv.py LABELED_FILE SEED FOLDS LABELS
    Convert LABELED_FILE into 2*FOLDS*LABELNUM FASTA files in the same directory,
    using the alphabets in LABELS as labels.
    ex. seq2cv.py yeast_protein ABCDEF

seq2cv LABELED_FILE [SEED] [FOLDS]
    Convert LABELED_FILE into 2*FOLDS*9 FASTA files in the same directory,
    using predefined labels for hallett 2005 (yeast).
    Default values: SEED=1000, FOLDS=10.
    ex. seq2cv.py ~/data/hallett2005/Hallett.out 

Note: when the label line include commas, only the first label is used.'''

HALLETT05LABELS = {
    'Cytosol'           : 'Cyto',
    'Spindle'           : 'Cyto',
    'ER'                : 'ER',
    'Golgi'             : 'Golgi',
    'Lysosome'          : 'Lyso',
    'Endosome'          : 'Lyso',
    'Vacuole'           : 'Lyso',
    'Mitochondrion'     : 'Mito',
    'Nucleus'           : 'Nuclear',
    'Nuclear envelop'   : 'Nuclear',
    'Peroxisome'        : 'Pero',
    'Plasma membrane'   : 'PM',
    'Bud neck'          : 'PM',
    'Cell wall'         : 'PM',
    'Secreted'          : 'Secreted',
    }

def generateCvFasta(labeledFileName, labelAbbr, seed, nFold, prefix):
    abbrDict = dict([(abbr, 1) for abbr in labelAbbr.values()])
    abbrs = abbrDict.keys()
    print 'Loading sequence file...'
    (id2seqAbbr, skippedLabels, nSeq) = loadLabeledFile(labeledFileName, labelAbbr)
    print 'Saving FASTA files of each fold...'
    saveCvFiles(id2seqAbbr, abbrs, seed, nFold, prefix)

    print '%d files written (%d folds, %d compartments)' \
          % (2 * nFold * len(abbrs), nFold, len(abbrs))
    print '%d unique sequences' % nSeq
    print 'Skipped lables:'; print skippedLabels
    print ''
    print 'Execute maketree.sh and then check_seq2cv.py to check!'

def loadLabeledFile(labeledFileName, labelAbbr):
    nSeq = 0
    skippedLabels = {}
    id2seqAbbr = {}
    labeledFile = open(labeledFileName)
    while True: 
        name = labeledFile.readline()
        while name!='' and len(name.strip())==0: name = labeledFile.readline()
        if name == '': break
        nSeq = nSeq + 1
        name = name.strip()
        labels = labeledFile.readline().strip().split(',')
        seq = ''; seqLine = '\n'
        while seqLine != '':
            seqLine = labeledFile.readline()
            if seqLine.startswith('//') or seqLine.startswith('\\\\'): break
            elif len(seqLine.strip()) == 0: continue
            seq = seq + seqLine.strip()
        assert name not in id2seqAbbr, 'Sequence ID duplicated'
        abbrDict = dict([(labelAbbr[label],1) for label in labels \
                         if label in labelAbbr])
        id2seqAbbr[name] = (seq, abbrDict.keys())
        for label in labels:
            if not label in labelAbbr: skippedLabels[label] = 1
    labeledFile.close()
    assert len(id2seqAbbr) == nSeq
    return (id2seqAbbr, skippedLabels.keys(), nSeq)

def saveCvFiles(id2seqAbbr, abbrs, seed, nFold, prefix):
    (sortedAbbrs, abbrUniqIds) = generateAbbrUniqSeq(id2seqAbbr, abbrs)
    print 'Sorted compartments: '; print sortedAbbrs
    random.seed(seed)
    for ids in abbrUniqIds.values(): random.shuffle(ids)
    for fold in range(nFold):
        id2seqAbbrTr = {}; id2seqAbbrTe = {}
        for ids in abbrUniqIds.values():
            start = len(ids)*fold/nFold
            end = len(ids)*(fold+1)/nFold
            for id in ids[:start]+ids[end:]: id2seqAbbrTr[id] = id2seqAbbr[id]
            for id in ids[start : end]     : id2seqAbbrTe[id] = id2seqAbbr[id]
        saveId2SeqAbbr(id2seqAbbrTr,abbrs,'%s%%s%dtr%d.fa' % (prefix,seed,fold))
        saveId2SeqAbbr(id2seqAbbrTe,abbrs,'%s%%s%dte%d.fa' % (prefix,seed,fold))

def generateAbbrUniqSeq(id2seqAbbr, abbrs):
    # Sort compartments (abbr) by number of sequences
    abbrNSeq = dict([(abbr, 0) for abbr in abbrs])
    for id, seqAbbrs in id2seqAbbr.items():
        for abbr in seqAbbrs[1]: abbrNSeq[abbr] = abbrNSeq[abbr] + 1
    abbrNSeqList = abbrNSeq.items()
    abbrNSeqList.sort(lambda x,y: cmp(x[1], y[1]))
    sortedAbbrs = [abbr for abbr, nseq in abbrNSeqList]

    # Assign id to (uniquely) smallest compartment (abbr)
    abbrUniqIds = dict([(abbr, []) for abbr in sortedAbbrs])
    for id, seqAbbrs in id2seqAbbr.items():
        (seq, abbrs1) = seqAbbrs
        for abbr in sortedAbbrs:
            if abbr in abbrs1: abbrUniqIds[abbr].append(id); break
    return (sortedAbbrs, abbrUniqIds)

def saveId2SeqAbbr(id2seqAbbr, abbrs, namePat):
    abbr2fp = dict([(abbr, open(namePat % abbr, 'w')) for abbr in abbrs])
    for id, seqAbbrs in id2seqAbbr.items():
        (seq, abbrs1) = seqAbbrs
        for abbr in abbrs1: abbr2fp[abbr].write('>%s\n%s\n' % (id, seq))
    for fp in abbr2fp.values(): fp.close()

# main
if __name__ == '__main__': #allows import from interpreter
    if len(sys.argv) <= 1: print USAGE; sys.exit(-1)
    labeledFileName = sys.argv[1]
    seed = 1000
    if len(sys.argv) > 2: seed = int(sys.argv[2])
    nFold = 10
    if len(sys.argv) > 3: nFold = int(sys.argv[3])
    if len(sys.argv) > 3:
        labelStr = sys.argv[3]
        labelAbbr = dict([(ch,ch) for ch in labelStr]) #[ch for ch in labelStr]
        outPrefix = labeledFileName
    else:
        labelAbbr = HALLETT05LABELS
        outPrefix = 'Yeast'

    generateCvFasta(labeledFileName, labelAbbr, seed, nFold, outPrefix)
