'''sgda_analize.py - calculate statistics about motif conservation
'''

import re, sys, os, glob
from memecv import sum
from seq2cv import HALLETT05LABELS, loadLabeledFile
from motif2hit import readMemeSitePos

USAGE = \
'''sgda_analyze.py TOP_MOTIF_FILE ALN_PATTERN CONS_FILE [MOTIF_NUM]
Calculate statistics about conservation of the ranked motif.

TOP_MOTIF_FILE: top motif file (meme format)
ALN_PATTERN: alignment file pattern
CONS_FILE: output conservation-statistics file
MOTIF_NUM: number of top motifs to analyze

File patterns use the %s symbol.

Example:
sgda_analize.py rank_flat_c.01/Yeast.rankmotif
~/subloc/result/sgd_align/aln/%s.aln rank_flat_c.01/Yeast.cons 20
'''

SIMILAR_SYMBOLS = ' .:*'

RE_MOTIF_START = re.compile(r'\d+ Motif Yeast\w+\d+ sites')

def saveMotifsSymbolRatio(topMotifFile, alnPattern, motifStatFile, nTop):
    '''Calculate conservation statistics and store in a file (main process).
    topMotifFile is the top motif file in MEME output format;
    alnPattern is the SGD alignment file pattern in printf format;
    motifStatFile is the output conservation statistics file;
    nTop is the number of top motifs to analyze (0 for all)'''
    motifs = readMemeSitePos(topMotifFile)
    if nTop > 0: motifs = motifs[:nTop]
    alignments = readAlignments(alnPattern, motifs)
    f = open(motifStatFile, 'w')
    #symbolRatio = countBackground(alignments)
    #f.write('\t'.join(['%.4f' % ratio for ratio in symbolRatio]) + '\n')
    nProtein = 0; totalWidth = 0.0
    for motif in motifs:
        (symbolRatio, nProtein1, avgWidth1) = count1Motif(motif, alignments)
        f.write('\t'.join(['%.4f' % ratio for ratio in symbolRatio]) + '\n')
        nProtein = nProtein + nProtein1; totalWidth = totalWidth + avgWidth1
    f.close()
    print 'Totally %d instances in %d top motifs' % (nProtein, len(motifs))
    print 'Avg instance width is %.3f peptides' % (totalWidth / len(motifs))
    print 'Product: %.3f' % (nProtein * (totalWidth / len(motifs)))

def saveBgSymbolRatio(alnPattern, bgStatFile):
    nFiles = 0
    count = [0] * len(SIMILAR_SYMBOLS)
    #reFile = re.compile((alnPattern % r'([\w\-]+)').replace('.', '\.'))
    for alnFile in glob.glob(alnPattern % '*'):
        aln = readAlignment(alnFile)
        if aln:
            countSymbolRatio(count, aln, 0, 0)
            nFiles = nFiles + 1
        else:
            print '%s: alignment not complete' % alnFile
    sumCount = float(sum(count))
    #for i in range(len(count)): count[i] = count[i] / sumCount
    symbolRatio = [c / sumCount for c in count]
    f = open(bgStatFile, 'w')
    f.write('\t'.join(['%.4f' % ratio for ratio in symbolRatio]) + '\n')
    f.close()
    print '%d non-empty alignment files processed' % nFiles

def count1Motif(motif, alignments):
    count = [0] * len(SIMILAR_SYMBOLS)
    totalWidth = 0
    for protein, st, end, sites in motif:
        if protein not in alignments: continue
        countSymbolRatio(count, alignments[protein], st, end)
        assert len(sites) == end - st
        totalWidth = totalWidth + (end - st)
    sumCount = float(sum(count))
    for iSymbol in range(len(count)):
        count[iSymbol] = count[iSymbol] / sumCount
    return (count, sumCount, totalWidth / sumCount)

def countSymbolRatio(count, aln, st, end):
    domainSymbols = getDomainSymbols(aln, st, end)
    if not domainSymbols: return
    d = 1.0 / len(domainSymbols)
    for c in domainSymbols:
        iSymbol = SIMILAR_SYMBOLS.index(c)
        count[iSymbol] = count[iSymbol] + d

def getDomainSymbols(aln, st, end):
    alignedSeq = aln[0][1]
    seq = alignedSeq.replace('-', '')
    if end == 0 and st == 0: end = len(seq)
    assert st < end
    if len(seq) < end: return None
    alignedSymbols = aln[-1][1]
    i = 0
    alignedSt = -1; alignedEnd = -1
    for j in range(len(alignedSeq)):
        if alignedSeq[j] != '-':
            i = i + 1
            if i == st + 1: alignedSt = j
            if i == end: alignedEnd = j + 1; break
    assert alignedSt >= 0 and alignedEnd >= 0, \
           '%s: %d~%d (len=%d), len(alignedSeq)=%d' % (aln[0][0], st, end, \
           len(seq), len(alignedSeq))
    assert seq[st:end] == alignedSeq[alignedSt:alignedEnd].replace('-', ''),\
           '%s: %d~%d %s => %d~%d %s' % (aln[0][0], st, end, seq[st:end],\
           alignedSt, alignedEnd, alignedSeq[alignedSt:alignedEnd])
    return alignedSymbols[alignedSt : alignedEnd]

def locateMotifs(motifs, seqs):
    for motif in motifs:
        for protein, st, end in motif:
            pass

def readTopMotifs(topMotifFile):
    motifs = []
    f = open(topMotifFile)
    while True:
        line = f.readline()
        if not line: break
        if not RE_MOTIF_START.search(line): continue
        f.readline()
        motif = []
        while True:
            line = f.readline().strip()
            if not line or line.startswith('-'): break
            tokens = line.split()
            assert len(tokens) == 2
            motif.append(tokens)
        motifs.append(motif)
    f.close()
    return motifs

def readAlignments(alnPattern, motifs):
    alignments = {}
    #collect unique protein names in motifs
    proteins = dict([(p,1) for m in motifs for p,s,e,ins in m]).keys()
    for protein in proteins:
        filename = alnPattern % protein
        if not os.path.exists(filename): continue
        alignment = readAlignment(filename)
        if alignment: alignments[protein] = alignment
        else: print '%s: alignment not complete' % protein
    return alignments

def readAlignment(filename):
    alignment = []
    f = open(filename)
    for line in f:
        tokens = line.strip().split('\t')
        if len(tokens) != 2: alignment = None; break
        alignment.append(tuple(tokens))
    f.close()
    return alignment

def testMismatch(topMotifFile, alnPattern, labeledFileName):
    motifs = readMemeSitePos(topMotifFile)
    alignments = readAlignments(alnPattern, motifs)
    (id2seqAbbr, skippedLabels, nSeq) = loadLabeledFile(labeledFileName, \
                                                        HALLETT05LABELS)
    sumMatchRatio = 0.0; nAlignment = 0
    for motif in motifs:
        for protein, st, end, instance in motif:
            if protein not in alignments: continue
            (species, alignedSeq) = alignments[protein][0]
            if species.find('Scer') < 0:
                print '%s: no Scer in species %s' % (protein, species)
            seq = alignedSeq.replace('-', '')
            sumMatchRatio = sumMatchRatio + diff(instance, seq[st:end])
            nAlignment = nAlignment + 1

            (seq1, abbrs) = id2seqAbbr[protein]
            seq1 = seq1.replace(' ', '')
            if seq1 != seq:
                if len(seq) < end:
                    assert end <= len(seq1)
                    print '%s: different from FASTA and domain is outside' \
                          % protein
                else:
                    print '%s: different from FASTA' % protein
    print 'Average match ratio: %f' % (sumMatchRatio / nAlignment)

def diff(s1, s2):
    m = 0
    for c1, c2 in zip(s1, s2):
        if c1 == c2: m = m + 1
    return float(m) / len(s1)

def testDomainSymbol(alnFile, st):
    end = st + 10
    aln = readAlignment(alnFile)
    print getDomainSymbols(aln, st, end)
    for species, alignedSeq in aln:
        print '%-15s %s' % (species, alignedSeq)

# main
if __name__ == '__main__': #allows import from interpreter    
    if len(sys.argv) <= 2: print USAGE; sys.exit(1)
    if len(sys.argv) <= 3:
        alnPattern = sys.argv[1]
        bgStatFile = sys.argv[2]
        saveBgSymbolRatio(alnPattern, bgStatFile)
        sys.exit(0)        
    nTop = 0
    topMotifFile = sys.argv[1]
    alnPattern = sys.argv[2]
    motifStatFile = sys.argv[3]
    if len(sys.argv) > 4: nTop = int(sys.argv[4])
    saveMotifsSymbolRatio(topMotifFile, alnPattern, motifStatFile, nTop)
