'''sgda_extract.py: Extract alignment from SGD HTML pages
Protein names (file names) are given from stdin, e.g. call by
python sgda_job.py Hallett.out 19 | python sgda_extract.py
'''

import re, sys, os

ALN_SEPARATOR     = 'weak similarity'
ROW_SEPARATOR     = 'SGD_Scer_'
LINE_SEPARATOR    = '<a target="infowin2" href="http://db.yeastgenome.org/'
SYMBOLS_SEPARATOR = '>Symbols</a>'

RE_SPECIES = re.compile(r'>?([\w\./]+)</a>')
RE_PEPTIDE = re.compile(r'<tt>([-A-Za-z])</tt>')
RE_SYMBOLS = re.compile(r'<tt>([\s\*:\.])</tt>')

SPECIAL_NO_ORTHO = 'Fungal orthologs have not yet been identified for S. cerevisiae ORF'
SPECIAL_NO_PAGE = 'Object not found!'

def extractByStream(stream, htmlPattern, alnPattern):
    '''Extract alignments from HTML pages with protein names given by
    the input stream (stdin or file).'''
    for line in stream:
        protein = line[:-1].strip() #remove \n
        if len(protein) == 0: continue
        loadExtractSave(htmlPattern % protein, alnPattern % protein)

def loadExtractSave(inFile, outFile):
    'Extract alignment from a HTML file and save to a ClustalW ALN file.'
    if os.path.exists(outFile): return
    inFp = open(inFile); html = inFp.read(); inFp.close()
    alignment = extract(html)
    if not alignment: explain(html, inFile); return
    outFp = open(outFile, 'w'); outFp.write(alignment); outFp.close()

def extract(html):
    'Extract alignment from one HTML page.'
    #extract the whole alignment section based on the phrase "weak similarity"
    htmlSections = html.split(ALN_SEPARATOR)
    if len(htmlSections) < 2: return None
    alignSection = htmlSections[1]
    #separate rows (the alignment of a fragment of all species)
    rowSubsections = alignSection.split(ROW_SEPARATOR)
    assert len(rowSubsections) > 1, 'Invalid structure ("SGD_Scer_")'
    species = []; alignedSeq = []
    for rowSubsection in rowSubsections[1:]:
        #separate lines within each row (species 47 RGVHQTNSPPS... 96)
        lines = rowSubsection.split(LINE_SEPARATOR)
        assert len(lines) > 1, 'Invalid structure ("<a target=infowin2...")'
        if rowSubsection is not rowSubsections[-1]: lines = lines[:-1]
        lines[0] = ROW_SEPARATOR + lines[0]
        assert len(species) == 0 or len(lines) + 1 == len(species), \
               'Invalid species number (%s vs %d %s)' % (species, len(lines), lines)
        linelen = 0
        for iSpecies in range(len(lines)):
            #extract species and aligned seq
            match = RE_SPECIES.search(lines[iSpecies])
            assert match, 'Invalid structure (">SPECIES</a>" vs %d %s\n%s)' \
                          % (iSpecies, lines[iSpecies], lines)
            species1 = match.group(1)
            alignedSeq1 = ''.join(RE_PEPTIDE.findall(lines[iSpecies]))
            #add species and aligned seq to array
            if len(species) == len(lines) + 1:
                assert species1 == species[iSpecies], \
                       'Invalid species (%s)' % species1
                alignedSeq[iSpecies] = alignedSeq[iSpecies] + alignedSeq1
            else:
                species.append(species1)
                alignedSeq.append(alignedSeq1)
            if linelen == 0:
                linelen = len(alignedSeq1)
            else:
                assert linelen == len(alignedSeq1), \
                       'Invalid seq len (%s)' % alignedSeq1
        assert rowSubsection.find(SYMBOLS_SEPARATOR) >= 0, \
               'Invalid structure (">Symbols</a>")'
        symbolPart = rowSubsection.split(SYMBOLS_SEPARATOR)[1]
        alignedSym1 = ''.join(RE_SYMBOLS.findall(symbolPart))
        assert len(alignedSym1) == linelen
        if len(species) == len(lines) + 1:
            alignedSeq[-1] = alignedSeq[-1] + alignedSym1
        else:
            species.append('Symbols')
            alignedSeq.append(alignedSym1)
    return ''.join(['%s\t%s\n' % (species1, seq1) for species1, seq1 \
                    in zip(species, alignedSeq)])

def explain(html, inFile):
    if html.find(SPECIAL_NO_ORTHO) >=0:
        print '%s: ortholog not yet identified' % inFile
    elif html.find(SPECIAL_NO_PAGE) >=0:
        print '%s: object not found' % inFile
    else:
        print '%s: no alignment for unknown reason' % inFile

# main
if __name__ == '__main__': #allows import from interpreter
    htmlPattern = '%s.html'; alnPattern = '%s.aln'
    if len(sys.argv) > 1: htmlPattern = sys.argv[1]
    if len(sys.argv) > 2: alnPattern = sys.argv[2]
    extractByStream(sys.stdin, htmlPattern, alnPattern)
