'''sgda_analize.py - calculate statistics about motif conservation
'''

import sys, glob, time, getopt
from motif2hit import readMemeSitePos
from sgda_analyze import readAlignments, readAlignment, getDomainSymbols

USAGE = \
'''sgda_ident.py TOP_MOTIF_FILE ALN_PATTERN CONS_FILE
Calculate statistics about conservation of the ranked motif.

TOP_MOTIF_FILE: top motif file (meme format)
ALN_PATTERN: alignment file pattern
CONS_FILE: output conservation-statistics file
MOTIF_NUM: number of top motifs to analyze (none for background)

File patterns use the %s symbol.

Example:
sgda_ident.py rank_flat_c.01/Yeast.post20sites
~/subloc/result/sgd_align/aln/%s.aln rank_flat_c.01/Yeast20sites.strong

For background, run sgda_ident.py ALN_PATTERN CONS_FILE
'''

STRONG = ':*'
IDENTICAL = '*'
conservedSymbols = IDENTICAL

def analyzeIdentical(topMotifFile, alnPattern, motifStatFile):
    motifs = readMemeSitePos(topMotifFile)
    alignments = readAlignments(alnPattern, motifs)
    f = open(motifStatFile, 'w')
    for motif in motifs:
        (nIdentical, nDomain) = count1Motif(motif, alignments)
        f.write('%d\t%d\n' % (nIdentical, nDomain))
    f.close()

def count1Motif(motif, alignments):
    nIdentical = 0; nDomain = 0
    for protein, st, end, sites in motif:
        if protein not in alignments: continue
        assert len(sites) == end - st
        domainSymbols = getDomainSymbols(alignments[protein], st, end)
        if not domainSymbols: continue
        nIdentical = nIdentical + isIdentical(domainSymbols)
        nDomain = nDomain + 1
    return (nIdentical, nDomain)

def isIdentical(domainSymbols):
    if len(domainSymbols) == 0: return 0
    for c in domainSymbols:
        if c not in conservedSymbols: return 0
    return 1

def analyzeBgIdentical(alnPattern, bgStatFile):
    nIdentical = 0; nDomain = 0
    for alnFile in glob.glob(alnPattern % '*'):
        aln = readAlignment(alnFile)
        if not aln: print '%s: alignment not complete' % alnFile; continue
        (nIdent1, nDomain1) = slidingWindowIdenticalRt(aln, 4)
        nIdentical = nIdentical + nIdent1
        nDomain = nDomain + nDomain1
    identRatio = nIdentical / float(nDomain)
    f = open(bgStatFile, 'w')
    f.write('%d\t%d\n' % (nIdentical, nDomain))
    f.close()

def slidingWindowIdenticalRt(aln, width):
    alignedSeq = aln[0][1]
    alignedSymbols = aln[-1][1]
    nIdentical = 0; nDomain = 0
    for alignedSt in range(len(alignedSeq) - width + 1):
        if alignedSeq[alignedSt] == '-': continue
        alignedEnd = 0; aaCount = 1
        for j in range(alignedSt + 1, len(alignedSeq)):
            if alignedSeq[j] == '-': continue
            aaCount = aaCount + 1
            if aaCount == width: alignedEnd = j + 1; break
        if alignedEnd == 0: break
        domainSymbols = alignedSymbols[alignedSt : alignedEnd]
        assert alignedSymbols
        nIdentical = nIdentical + isIdentical(domainSymbols)
        nDomain = nDomain + 1
    assert nDomain == len(alignedSeq.replace('-', '')) - width + 1, \
           '%d != %d' % (nDomain, len(alignedSeq.replace('-', '')) - width + 1)
    return (nIdentical, nDomain)


# main
if __name__ == '__main__': #allows import from interpreter    
    try:
        (opts, args) = getopt.getopt(sys.argv[1:], 's')
    except getopt.GetoptError:
        print USAGE; sys.exit(2)
    for o, a in opts:
        if o == '-s': conservedSymbols = STRONG
    if len(args) != 2 and len(args) != 3: print USAGE; sys.exit(2)
    if len(args) == 2: #background
        (alnPattern, bgStatFile) = args
        t0 = time.time()
        analyzeBgIdentical(alnPattern, bgStatFile)
        print 'Elapsed time: %.0f sec' % (time.time() - t0)
        sys.exit(0)
    (topMotifFile, alnPattern, motifStatFile) = args
    analyzeIdentical(topMotifFile, alnPattern, motifStatFile)
