# Read motif-position files, and meme output file, and match the two
# 1. For each compartment,
# 2.   Read true motif position
# 3.   Read sequence file (to map seqId to seqNo)
# 4.   Read MEME output
# 5.   Match and calculate precision and recall
# 6. Calculate total precision / recall
# 7. Save as result file
#
# Example result:
# TITLE
#   COMPARTMENT     MOTIF   TP  PRED    TRUENUM PREC    RECALL
#   YeastER         hdel    3   10      5       30.00   50.00
#   YeastER         kk      3   10      5       30.00   50.00
#   YeastER         rr      3   10      5       30.00   50.00
#   YeastNuclear    nuc4    3   10      5       30.00   50.00
#   4 KNOWN MOTIFS FOUND    6   40      20      30.00   50.00

import re, time, sys, glob
from memecv import *
from scanregex import * #readMotPos

# Evaluation paramters
paramOverlapRatio       = .75
paramRecallThreshold    = .66
paramPrecThreshold      = .66

def scanAllMeme(motPosFiles, memeFiles, seqFiles, resultFile, comps):
    'Process all compartments'
    compResults = []
    for motPosFile, memeFile, seqFile in zip(motPosFiles, memeFiles, seqFiles):
        compResults.append(matchMeme(motPosFile, memeFile, seqFile))
    title = 'OverlapRatio = %g, RecallThreshold = %g' \
            % (paramOverlapRatio, paramRecallThreshold)
    writeResult(title, compResults, comps, resultFile)

def matchMeme(motPosFile, memeFile, seqFile):
    'Process one compartment'
    motNameInstances = readMotPos(motPosFile)
    if len(motNameInstances) == 0: return []
    nameSeqs = fastaReadAll(seqFile)
    memeMotifs = readMemeMotIns(memeFile, nameSeqs)
    return matchMotifs(motNameInstances, memeMotifs)

def readMemeMotIns(memeFilename, nameSeqs):
    '''Read motif instances in MEME output.
    MEME output: (seqId, seqNo, st, end) = memePos[iMotif][iInstance]'''
##    print 'Reading %s...' % memeFilename
    f = open(memeFilename)
    motifPos = []
    while True:
        line = f.readline()
        if line == '': break
        m = re.search(r'Motif (\d+) block diagrams', line)
        if m != None:
            assert len(motifPos) + 1 == int(m.group(1)), \
                   'Motif %s in MEME output should be %d' \
                   % (m.group(1), len(motifPos) + 1)
            #print 'motif %s' % m.group(1)
            f.readline(); f.readline(); f.readline()
            motif1Pos = []
            while True:
                line = f.readline()
                m = re.match(r'>?(\w+)\s.+(\d+_)?\[(\d+)\](_\d+)?', line)
                if m == None: break
                assert len(motifPos) + 1 == int(m.group(3)), \
                       'MEME output "%s...[%s]" should be [%d]' \
                       % (m.group(1), m.group(3), len(motifPos) + 1)
                seqId = m.group(1)
                seqNo = fastaNumberByName(nameSeqs, seqId)
                if m.group(2) != None: st = int(m.group(2)[:-1]) #starting at 0
                else: st = 0
                if m.group(4) != None: end = int(m.group(4)[1:])
                else: end = 0
                end = len(nameSeqs[seqNo][1]) - end # motif==seq[st:end]
                motif1Pos.append((seqId, seqNo, st, end))
            motifPos.append(motif1Pos)
    f.close()
    return motifPos

def fastaNumberByName(nameSeqs, name):
    'Return sequence number (indexed from zero) of given sequence name.'
    for i in range(len(nameSeqs)):
        if name == nameSeqs[i][0]: return i
    return -1

def matchMotifs(motNameInstances, memeMotifs):
    'Match found motifs to true positions'
    foundResults = []
    for motName, motIns in motNameInstances:
        for i in range(len(memeMotifs)):
            if matchMotif(foundResults, motName, motIns, memeMotifs[i]):
                del memeMotifs[i]; break
    return foundResults

def matchMotif(foundResults, motName, motInstances, memeMotif):
    '''Match intances of one motif type; return true and append to foundResults
    if recall above threshold.'''
    activeMotif = memeMotif[:]
    tp1 = 0
    #print 'overlap=%f' % ((motInstances[0][3]-motInstances[0][2]) * paramOverlapRatio)
    seqIdList = []
    for trueIns in motInstances:
        #print '[%s]' % trueIns[0]
        for iIns in range(len(activeMotif)):
            if not trueIns[0].startswith('NP'):
                print '%s not starting with NP' % trueIns[0]
            if not trueIns[0] in seqIdList: seqIdList.append(trueIns[0])
            if matchMotifInstance(trueIns,activeMotif[iIns]):
                tp1 = tp1 + 1; del activeMotif[iIns]; break
    #if tp1 > 0: print '%s found TP = %d' % (motName, tp1)
    if float(tp1) / len(seqIdList) >= paramRecallThreshold or \
       float(tp1) / len(memeMotif) >= paramPrecThreshold :
        foundResults.append((motName, tp1, len(memeMotif), len(seqIdList)))
        return True

def matchMotifInstance(trueIns, ins):
    'Match one motif instance; return true if overlap is larger than given bases.'
    overlap = min(trueIns[3] - trueIns[2], ins[3] - ins[2]) * paramOverlapRatio
    return ins[0] == trueIns[0] and ins[2] <= trueIns[3] - overlap \
           and trueIns[2] + overlap <= ins[3]

def writeResult(title, compResults, comps, resultFile):
    fRes = open(resultFile, 'a')
    fRes.write('%s\n' % title)
    fRes.write('COMPARTMENT\tMOTIF\tTP\tPRED\tTRUENUM\tPREC\tRECALL\n')
    print 'COMPARTMENT\tMOTIF\tTP\tPRED\tTRUENUM\tPREC\tRECALL'
    nFound = 0; sumTP = 0; sumP = 0; sumT = 0;
    for comp, foundResults in zip(comps, compResults):
        nFound = nFound + len(foundResults)
        for motName, tp1, p1, t1 in foundResults:
            line = '%s\t%s\t%d\t%d\t%d\t%.2f\t%.2f' \
                   % (comp, motName, tp1, p1, t1, 100.*tp1/p1, 100.*tp1/t1)
            print line; fRes.write(line + '\n')
            sumTP = sumTP + tp1; sumP = sumP + p1; sumT = sumT + t1
    totalPrec = 0.
    if sumP > 0: totalPrec = 100.*sumTP/sumP
    totalRec = 0.
    if sumT > 0: totalRec = 100.*sumTP/sumT
    summary = '%d KNOWN MOTIFS FOUND: \t%d\t%d\t%d\t%.2f\t%.2f' \
              % (nFound, sumTP, sumP, sumT, totalPrec, totalRec)
    print summary
    fRes.write(summary + '\n')
    fRes.close()

# main
if __name__ == '__main__': #allows import from interpreter
    comp = sys.argv[1]
    memeFile = sys.argv[2]
    if len(sys.argv) > 3: paramRecallThreshold = float(sys.argv[3])
    if len(sys.argv) > 4: paramOverlapRatio    = float(sys.argv[4])
    if len(sys.argv) > 5: paramPrecThreshold   = float(sys.argv[5])
    motPosFile = '../../result/hallett_yeast/Yeast%s.motpos' % comp
    seqFile = 'Yeast%s.txt' % comp
    resultFile = memeFile + '.motres'
    
    compResults = [matchMeme(motPosFile, memeFile, seqFile)]
    title = 'OverlapRatio = %g, RecallThreshold = %g' \
            % (paramOverlapRatio, paramRecallThreshold)
    writeResult(title, compResults, [comp], resultFile)
