/* check.c
 *
 * Unit test cases, and main() for test runner.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"
#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "msa.h"                /* squid's multiple alignment i/o       */

#include "CuTest.h"             /* C unit test framework                */
#include "subloc.h"             /* SubLoc declarations                  */

/***********************************************
 * IO
 ***********************************************/

static int ncate = 2;
static char *seqfiles[] = {"CuTest/yeast_seq_O.fa", "CuTest/yeast_seq_U.fa"};
static char *seqcompfile = "CuTest/yeast_seq.fa";
static char *complabels[] = {"O", "U"};

static char *seq_o1 = "MQDLEIFLSIFAFIFVFYFGAHRTVMNRNKSDVPYLQ";
static char *seq_o_last =
"MAVQTPRQRLANAKFNKNNEKYRKYGKKKEGKTEKTAPVISKTWLGILLFLLVGGGVLQLISYIL";
static char *seq_u2 =
"MAEKRTLIAVIADEDTTTGLLLAGIGQITPETQEKNFFVYQEGKTTKEEITDKFNHFTEE"
"RDDIAILLINQHIAENIRARVDSFTNAFPAILEIPSKDHPYDPEKDSVLKRVRKLFGE";

static void assertDsqEquals(CuTest *tc, unsigned char *ex, int exlen,
                            unsigned char *ac, int aclen)
{
  int t;
  CuAssertIntEquals(tc, exlen, aclen);
  for (t = 0; t < exlen; t++) CuAssertIntEquals(tc, ex[t], ac[t]);
}

static void test_LoadDsqs(CuTest *tc)
{
  //printf("test_LoadSeq\n");
  int nseqs = 0, idx;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = LoadDsqs(seqfiles[0], &nseqs, &sqinfo);
  unsigned char *dsq_o1 = DigitizeSequence(seq_o1, strlen(seq_o1));
  unsigned char *dsq_o_last = DigitizeSequence(seq_o_last, strlen(seq_o_last));
  //for (idx = 0; idx < 5; idx++) printf("%d [%s]\n", idx, seqs[idx]);
  assertDsqEquals(tc, dsq_o1, strlen(seq_o1), dsqs[1], sqinfo[1].len);
  assertDsqEquals(tc, dsq_o_last, strlen(seq_o_last), dsqs[nseqs-1], sqinfo[nseqs-1].len);
  /*CuAssertStrEquals(tc, seq_o1, seqs[1]);
  CuAssertStrEquals(tc, seq_o_last, seqs[nseqs-1]);*/
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(sqinfo); free(dsqs);
}

static void test_LoadCateDsqs(CuTest *tc)
{
  //printf("test_LoadCateDsqs\n");
  int k, idx;
  int *catenseq = NULL;
  SQINFO **catesqinfo = NULL;
  unsigned char ***catedsqs = LoadCateDsqs(ncate, seqfiles, &catenseq, &catesqinfo);
  unsigned char *dsq_o1 = DigitizeSequence(seq_o1, strlen(seq_o1));
  unsigned char *dsq_o_last = DigitizeSequence(seq_o_last, strlen(seq_o_last));
  unsigned char *dsq_u2 = DigitizeSequence(seq_u2, strlen(seq_u2));
  assertDsqEquals(tc, dsq_o1, strlen(seq_o1), catedsqs[0][1], catesqinfo[0][1].len);
  assertDsqEquals(tc, dsq_o_last, strlen(seq_o_last), catedsqs[0][catenseq[0]-1],
                  catesqinfo[0][catenseq[0]-1].len);
  assertDsqEquals(tc, dsq_u2, strlen(seq_u2), catedsqs[1][2], catesqinfo[1][2].len);
  CuAssertIntEquals(tc, 343, catenseq[0] + catenseq[1]);
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
}

static void test_LoadLabeledDsqs(CuTest *tc)
{
  //printf("test_LoadLabeledDsqs\n");
  int k, idx;
  int *catenseq = NULL;
  SQINFO **catesqinfo = NULL;
  unsigned char ***catedsqs = LoadLabeledDsqs(seqcompfile, ncate, complabels,
                                              &catenseq, &catesqinfo);
  unsigned char *dsq_o1 = DigitizeSequence(seq_o1, strlen(seq_o1));
  unsigned char *dsq_o_last = DigitizeSequence(seq_o_last, strlen(seq_o_last));
  unsigned char *dsq_u2 = DigitizeSequence(seq_u2, strlen(seq_u2));
  assertDsqEquals(tc, dsq_o1, strlen(seq_o1), catedsqs[0][1], catesqinfo[0][1].len);
  assertDsqEquals(tc, dsq_o_last, strlen(seq_o_last), catedsqs[0][catenseq[0]-1],
                  catesqinfo[0][catenseq[0]-1].len);
  assertDsqEquals(tc, dsq_u2, strlen(seq_u2), catedsqs[1][2], catesqinfo[1][2].len);
  CuAssertIntEquals(tc, 343, catenseq[0] + catenseq[1]);
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
}

static int nhmmfiles = 2;
static char *hmmfiles[] = {"CuTest/O.hmm", "CuTest/U.hmm"};

static void test_LoadHMMs(CuTest *tc)
{
  //printf("test_LoadHMMs\n");
  int i;
  struct plan7_s **hmms = LoadHMMs(nhmmfiles, hmmfiles);
  CuAssertIntEquals(tc, 2, hmms[0]->M);
  CuAssertIntEquals(tc, 12, hmms[1]->M);
  for (i = 0; i < nhmmfiles; i++) (void) FreePlan7(hmms[i]);
  free(hmms);
}

static void test_SaveFeatFile(CuTest *tc)
{
  double llr0[][3] = {{123.45, -54.321, -1000.0},
                      {-54.321, 10.0, 1.23},
                      {-10.0, 20.0, 30.0},
                      {-54.321, -123.45, 100.0}};
  char *featfile = "CuTest/features.dat";

  int idx;
  double *(llr[4]);
  for (idx = 0; idx < 4; idx++) llr[idx] = llr0[idx];
  SaveFeatFile(4, 3, llr, featfile);
  /* Read saved feature file and check content later. */
}

static void test_ListFilesWithExt(CuTest *tc)
{
  int i, nfiles;
  char **files = ListFilesWithExt("CuTest/", ".ext", &nfiles);
  CuAssertIntEquals(tc, 3, nfiles);
  for (i = 0; i < nfiles; i++)
    {
      CuAssertIntEquals(tc, strlen("CuTest/exttest1.ext"), strlen(files[i]));
      CuAssertStrEquals(tc, ".ext", files[i] + strlen(files[i]) - 4);
    }
  for (i = 0; i < nfiles; i++) free(files[i]);
  free(files);
}

static CuSuite *IoGetSuite(void)
{
  CuSuite *suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, test_LoadDsqs);
  SUITE_ADD_TEST(suite, test_LoadCateDsqs);
  SUITE_ADD_TEST(suite, test_LoadLabeledDsqs);
  SUITE_ADD_TEST(suite, test_LoadHMMs);
  SUITE_ADD_TEST(suite, test_SaveFeatFile);
  SUITE_ADD_TEST(suite, test_ListFilesWithExt);
  return suite;
}

/***********************************************
 * Classify
 ***********************************************/

const int PHMM_LEN = 8;
const int MAXITER = 100;

/* Swissprot_CYB5_YEAST O */
static char*seq1="MPKVYSYQEVAEHNGPENFWIIIDDKVYDVSQFKDEHPGGDEIIMDLGGQDATESFVDIG"
                 "HSDEALRLLKGLYIGDVDKTSERVSVEKVSTSENQSKGSGTLVVILAILMLGVAYYLLNE";
/* >Swissprot_VATF_YEAST U */
static char*seq2="MAEKRTLIAVIADEDTTTGLLLAGIGQITPETQEKNFFVYQEGKTTKEEITDKFNHFTEE"
                 "RDDIAILLINQHIAENIRARVDSFTNAFPAILEIPSKDHPYDPEKDSVLKRVRKLFGE";

static void test_Classify(CuTest *tc)
{
  //printf("test_Classify\n");
  int i;
  struct plan7_s **hmms = LoadHMMs(nhmmfiles, hmmfiles);
  unsigned char *dsq1 = DigitizeSequence(seq1, strlen(seq1));
  unsigned char *dsq2 = DigitizeSequence(seq2, strlen(seq2));
  float *scores1 = Classify(nhmmfiles, hmms, dsq1, strlen(seq1));
  float *scores2 = Classify(nhmmfiles, hmms, dsq2, strlen(seq2));
  printf("Seq1 loglik");
  for (i = 0; i < nhmmfiles; i++) printf(" %.3f", scores1[i]);
  printf("\nSeq2 loglik");
  for (i = 0; i < nhmmfiles; i++) printf(" %.3f", scores2[i]);
  printf("\n");
  /*int prediction1 = argmaxf(nhmmfiles, scores1);
  int prediction2 = argmaxf(nhmmfiles, scores1);
  CuAssertIntEquals(tc, 0, prediction1);
  CuAssertIntEquals(tc, 1, prediction2);*/
  CuAssertDblEquals(tc, -3.030, scores1[0], .005);
  CuAssertDblEquals(tc, -4.290, scores1[1], .005);
  CuAssertDblEquals(tc, -3.614, scores2[0], .005);
  CuAssertDblEquals(tc, -3.855, scores2[1], .005);
  for (i = 0; i < nhmmfiles; i++) (void) FreePlan7(hmms[i]);
  free(hmms);
  free(scores1);
  free(scores2);
}

static void test_EvaluateSeqs(CuTest *tc)
{
  //printf("test_EvaluateSeqs\n");
  int k;
  int catenseq[] = {1, 1};
  unsigned char *dsq1 = DigitizeSequence(seq1, strlen(seq1));
  unsigned char *dsq2 = DigitizeSequence(seq2, strlen(seq2));
  unsigned char *cate0dsqs[] = {dsq1}, *cate1dsqs[] = {dsq2};
  unsigned char **catedsqs[] = {cate0dsqs, cate1dsqs};
  SQINFO sqinfo1[1], sqinfo2[1];
  SQINFO *catesqinfo[] = {sqinfo1, sqinfo2};
  struct plan7_s **hmms = LoadHMMs(nhmmfiles, hmmfiles);
  float accuracy;
  catesqinfo[0][0].len = strlen(seq1);
  catesqinfo[1][0].len = strlen(seq2);
  accuracy = EvaluateSeqs(nhmmfiles, hmms, catenseq, catedsqs, catesqinfo);
  CuAssertDblEquals(tc, .5, accuracy, .001);
  for (k = 0; k < nhmmfiles; k++) (void) FreePlan7(hmms[k]);
  free(hmms);
}

static void test_CrossValidate(CuTest *tc)
{
  //printf("test_CrossValidate\n");
  char *labels[] = {"S", "T"};

  int *catenseq = NULL;
  SQINFO **catesqinfo = NULL;
  unsigned char ***catedsqs = LoadLabeledDsqs(seqcompfile, ncate, labels,
                                              &catenseq, &catesqinfo);
  RunCrossValidate(ncate, catenseq, catedsqs, catesqinfo, 4, 0, PHMM_LEN, 0.01,
                   2, 10, 10, 20., 530., 0, 0, "", NULL);
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
}

static CuSuite *ClassifyGetSuite(void)
{
  CuSuite *suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, test_Classify);
  SUITE_ADD_TEST(suite, test_EvaluateSeqs);
  /*SUITE_ADD_TEST(suite, test_CrossValidate);*/
  return suite;
}

/***********************************************
 * Generative training
 ***********************************************/

static void test_GenTrain(CuTest *tc)
{
  //printf("test_Train\n");
  RunGenTrain("CuTest/yeast_seq_B.fa", "CuTest/trn_B8.hmm", PHMM_LEN, 10, 2, 20., 530., 1);
}

static CuSuite *TrainGetSuite(void)
{
  CuSuite *suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, test_GenTrain);
  return suite;
}

/***********************************************
 * Discriminative training
 ***********************************************/

static void test_DiscTrain(CuTest *tc)
{
  //printf("test_DiscTrain\n");
  int           ncate        = 2;
  char         *seqfiles[]   ={"CuTest/yeast_seq_S.fa","CuTest/yeast_seq_T.fa"};
  char         *hmmd_files[] = {"CuTest/s_disc8i1.hmm", "CuTest/t_disc8i1.hmm"};
  char        **hmmg_files   = NULL; /* not save generative HMM for now*/
  int           setlength    = PHMM_LEN;
  float         learnrt      = 0.01;
  int           ninitg       = 1;
  int           ninitd       = 1;
  int           maxiterg     = 3;
  int           maxiterd     = 10;
  float         inslimit     = 20.;
  float         avglen       = 530.;
  int           verbose      = 1;
  struct plan7_s      **hmmg       = NULL;
  struct plan7_s      **hmmd       = NULL;
  int                  *catenseq   = NULL;
  unsigned char      ***catedsqs   = NULL;
  SQINFO              **catesqinfo = NULL;

  SetAlphabet(hmmAMINO);
  catedsqs = LoadCateDsqs(ncate, seqfiles, &catenseq, &catesqinfo);
  if (maxiterg > 0 && NULL != hmmg_files)
    hmmg = (struct plan7_s **) MallocOrDie(ncate*sizeof(struct plan7_s *));
  hmmd = DiscTrain(ncate, catenseq, catedsqs, catesqinfo, setlength, learnrt,
                   ninitg, ninitd, maxiterg, maxiterd, inslimit, avglen,
                   verbose, hmmg);
  if (maxiterg > 0 && NULL != hmmg_files) /* otherwise NULL == hmmg */
    {
      SaveHMMs(ncate, hmmg, hmmg_files);
      FreeHMMs(ncate, hmmg); free(hmmg);
    }
  if (maxiterd > 0) /* otherwise NULL == hmmd */
    {
      if (NULL != hmmd_files) SaveHMMs(ncate, hmmd, hmmd_files);
      FreeHMMs(ncate, hmmd);
      free(hmmd);
    }
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
}

static CuSuite *DiscTrainGetSuite(void)
{
  CuSuite *suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, test_DiscTrain);
  return suite;
}

/***********************************************
 * Main
 ***********************************************/

/* Test runner for all test suite */
static void RunAllTests(void)
{
  int i;
  CuString *output = CuStringNew();
  CuSuite *suite = CuSuiteNew();

  /* Test suites added here */
  CuSuiteAddSuite(suite, IoGetSuite());
  CuSuiteAddSuite(suite, ClassifyGetSuite());
  CuSuiteAddSuite(suite, TrainGetSuite());
  CuSuiteAddSuite(suite, DiscTrainGetSuite());

  CuSuiteRun(suite);
  CuSuiteSummary(suite, output);
  CuSuiteDetails(suite, output);
  printf("%s\n", output->buffer);

  free(output->buffer); free(output);
  for (i = 0; i < suite->count; i++)
  {
    /* No need to free suite->list[i]->name */
    if (NULL != suite->list[i]->message) free((void*) suite->list[i]->message);
    free(suite->list[i]);
  }
  free(suite);
}

/* Main function, just run the test runner. */
int main(int argc, char **argv)
{
  RunAllTests();
  return 0;
}
