/* subdisc.c
 *
 * Discriminative training an HMM from unaligned sequences
 * (using both positive and negative examples).
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"             /* function declarations for SUBLOC     */

#define CHECK_NAN(x) if (!((x)>=0 || (x)<0)) printf("%s is NaN: %f\n", __STRING(x), x)

/* Stupid routine called only in HMMShiftPositive();
 * better use the iterator pattern but this naive approach works so far.
 */
static float
MinRatio(int n, float *up, float *down)
{
  int j;
  float optimum;
  if (n <= 0) (void)Die("Invalid array length (%d) in MinRatio()", n);
  optimum = up[0] / down[0];
  for (j = 1; j < n; j++)
    if (up[j] / down[j] - optimum < -FLT_EPSILON) optimum = up[j] / down[j];
  /*if (!(optimum >= 0 || optimum < 0)) printf("MinRatio = %f\n", optimum);*/
  return optimum;
}

/* Stupid routine called only in HMMShiftPositive();
 * better use the iterator pattern but this naive approach works so far.
 */
static void
AddProduct(int n, float *arr, float c, float *delta)
{
  int i;
  for (i = 0; i < n; i++) arr[i] += c * delta[i];
  for (i = 0; i < n; i++)
    if (!(arr[i] >= -FLT_EPSILON))
      printf("arr[%d] = %f < 0; delta[i] = %f; c = %f\n", i, arr[i], delta[i], c);
  for (i = 0; i < n; i++) assert(arr[i] >= -FLT_EPSILON);
}

static void
DiscForBackNaN(int ncate, unsigned char *dsq1, SQINFO *sqinfo1,
               struct plan7_s **hmms, struct mledpmat_s **fwd,
               struct mledpmat_s **bck, int k, float t10, float sclpr)
{
  char hmmfile[200];
  int i, k1;
  printf("Sequence %s: t10=%.2f, sclprob=%f\nscl: ",sqinfo1->name,t10,sclpr);
  for (i=1; i<sqinfo1->len; i++) printf("%.2f ",fwd[k]->scl[i]);
  printf("\nFWD:\n");PrintMleDpMat(fwd[k],sqinfo1->len,hmms[k]->M);
  printf("BCK:\n");  PrintMleDpMat(bck[k],sqinfo1->len,hmms[k]->M);
  fflush(stdout);
  FILE *seqfp = fopen("assert_disc.fa", "w");
  char *seq = DedigitizeSequence(dsq1, sqinfo1->len);
  WriteSeq(seqfp, SQFILE_FASTA, seq, sqinfo1);
  fclose(seqfp);
  free(seq);
  for (k1=0; k1<ncate; k1++)
    {
      sprintf(hmmfile, "assert%02d.hmm", k1);
      SaveHMM(hmms[k1], hmmfile);
    }
  for (k1=k; k1<ncate; k1++) {FreeMleDpMat(fwd[k1]); FreeMleDpMat(bck[k1]);}
  free(fwd); free(bck);
}

/* Calculate sufficient statistics, or weighted forward-backward statistics,
 * or weighted gamma and xi matrix. E-step of the discriminative training.
 * To calculate log conditional likelihood without updating,
 * call with the argument newhmms == NULL.
 */
static float
DiscCountSuffStat(int ncate, int *catenseq, unsigned char ***catedsqs,
                  SQINFO **catesqinfo, struct plan7_s **hmms,
                  struct plan7_s **newhmms, int *ret_trncorrect)
{
  int    cate, k, idx;
  double wgt;
  float  logcondlik = 0.0;
  int    trncorrect = 0;

  struct mledpmat_s **fwd = (struct mledpmat_s **) MallocOrDie(
                            ncate * sizeof(struct mledpmat_s *));
  struct mledpmat_s **bck = (struct mledpmat_s **) MallocOrDie(
                            ncate * sizeof(struct mledpmat_s *));
  double *sclpr = (double *) MallocOrDie(ncate * sizeof(double));
  double *lr0 = (double *) MallocOrDie(ncate * sizeof(double));
  lr0[0] = 1.0;

  for (cate = 0; cate < ncate; cate++)
    {
      unsigned char **dsq = catedsqs[cate];
      SQINFO *sqinfo = catesqinfo[cate];
      for (idx = 0; idx < catenseq[cate]; idx++)
        {
          double loglik0 = 0.0, loglik;
          double sumlr = 1.0; /* 1 + exp(L2-L1) + exp(L3-L1) + ... */
          for (k = 0; k < ncate; k++)
            {
              sclpr[k] = MleForward(dsq[idx], sqinfo[idx].len, hmms[k], &(fwd[k]));
              MleBackward(dsq[idx], sqinfo[idx].len, hmms[k], fwd[k]->scl, &(bck[k]));
              loglik = MleFwdScore(fwd[k], sqinfo[idx].len, hmms[k]->M, sclpr[k]);
              if (0 == k) loglik0 = loglik;
              else { lr0[k] = sreEXP2(loglik - loglik0); sumlr += lr0[k]; }
            }
          for (k = 0; k < ncate; k++)
            {
              /* REQUIRES double precision! */
              double post = lr0[k] / sumlr;
              if (k == cate) { logcondlik += sreLOG2(post); wgt = 1.0 - post; }
              else wgt = -post;
              if (NULL != newhmms)
                {
                  MleForbackCount(hmms[k], dsq[idx], sqinfo[idx].len, sclpr[k],
                                  wgt, fwd[k], bck[k], newhmms[k]);
                  if (!(newhmms[k]->t[1][0]>=0.||newhmms[k]->t[1][0]<0.))
                    { /* Infinity; terminate */
                      DiscForBackNaN(ncate, dsq[idx], sqinfo+idx, hmms, fwd, bck, k,
                                     newhmms[k]->t[1][0], sclpr[k]);
                      free(sclpr); free(lr0); return 0.0; /* freed fwd, bck */
                    }
                }
              FreeMleDpMat(fwd[k]); FreeMleDpMat(bck[k]);
            }
          if (argmaxd(ncate, lr0) == cate) trncorrect++;
        }
    }
  free(fwd); free(bck); free(sclpr); free(lr0);
  *ret_trncorrect = trncorrect;
  return logcondlik;
}

/* Shift discriminatively trained HMM parameters (transition and emission)
 * so that all simplex are positive. This is part of the MMIE discriminative
 * training algorithm originated in the speech recognition community.
 * Currently we use two smoothing constant shared across all states;
 * future works include different smoothing constant for each state.
 */
static void
DiscShiftPositive(struct plan7_s *hmm0, struct plan7_s *hmm)
{
  float smooth_t = 0.0; /* smoothing constant for transition  */
  float smooth_e = 0.0; /* smoothing constant for emission    */
  int i;                /* counter for states                 */

  for (i = 1; i < hmm->M ; i++)
    {
      if (hmm0->t[i][TMM] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (hmm0->t[i][TMI] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (i < hmm->M-1 && hmm0->t[i][TMD] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (hmm0->t[i][TIM] == 0.0) printf("t[%d][TIM]==0\n", i);
      if (hmm0->t[i][TII] == 0.0) printf("t[%d][TII]==0\n", i);
      if (i > 1 && i < hmm->M-1 && hmm0->t[i][TDM] == 0.0) printf("t[%d][TDM]==0\n", i);
      if (i > 1 && i < hmm->M-1 && hmm0->t[i][TDD] == 0.0) printf("t[%d][TDD]==0\n", i);

      if (hmm->t[i][TMM] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (hmm->t[i][TMI] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (i < hmm->M-1 && hmm->t[i][TMD] == 0.0) printf("t[%d][TMM]==0\n", i);
      if (hmm->t[i][TIM] == 0.0) printf("t[%d][TIM]==0\n", i);
      if (hmm->t[i][TII] == 0.0) printf("t[%d][TII]==0\n", i);
      if (i > 1 && i < hmm->M-1 && hmm->t[i][TDM] == 0.0) printf("t[%d][TDM]==0\n", i);
      if (i > 1 && i < hmm->M-1 && hmm->t[i][TDD] == 0.0) printf("t[%d][TDD]==0\n", i);
    }
  for (i = 1; i < hmm->M ; i++)
    {
      if (i < hmm->M - 1)
        smooth_t = MIN(smooth_t, MinRatio(3, &(hmm->t[i][TMM]), &(hmm0->t[i][TMM])));
      else /* TMM and TMI only */
        smooth_t = MIN(smooth_t, MinRatio(2, &(hmm->t[i][TMM]), &(hmm0->t[i][TMM])));
      smooth_t = MIN(smooth_t, MinRatio(2, &(hmm->t[i][TIM]), &(hmm0->t[i][TIM])));
      if (i > 1 && i < hmm->M-1)
        smooth_t = MIN(smooth_t, MinRatio(2, &(hmm->t[i][TDM]), &(hmm0->t[i][TDM])));
      smooth_e = MIN(smooth_e, MinRatio(Alphabet_size, hmm->mat[i], hmm0->mat[i]));
      smooth_e = MIN(smooth_e, MinRatio(Alphabet_size, hmm->ins[i], hmm0->ins[i]));
    }
  smooth_e = MIN(smooth_e, MinRatio(Alphabet_size, hmm->mat[hmm->M], hmm0->mat[hmm->M]));

  //printf("smooth_t = %e, smooth_e = %e\n", smooth_t, smooth_e);
  smooth_t = -2 * smooth_t; /* standard sign */
  smooth_e = -2 * smooth_e; /* standard sign */
  for (i = 1; i < hmm->M ; i++)
    {
      if (i < hmm->M - 1)
        AddProduct(3, &(hmm->t[i][TMM]), smooth_t, &(hmm0->t[i][TMM]));
      else /* TMM and TMI only */
        AddProduct(2, &(hmm->t[i][TMM]), smooth_t, &(hmm0->t[i][TMM]));
      AddProduct(2, &(hmm->t[i][TIM]), smooth_t, &(hmm0->t[i][TIM]));
      if (i > 1 && i < hmm->M-1)
        AddProduct(2, &(hmm->t[i][TDM]), smooth_t, &(hmm0->t[i][TDM]));
      AddProduct(Alphabet_size, hmm->mat[i], smooth_e, hmm0->mat[i]);
      AddProduct(Alphabet_size, hmm->ins[i], smooth_e, hmm0->ins[i]);
    }
  AddProduct(Alphabet_size, hmm->mat[hmm->M], smooth_e, hmm0->mat[hmm->M]);
}

/* Damp one simplex, like transitions or emissions from one state.
 */
static void
Damp1Simplex(int n, float *new, float learnrt, float *old)
{
  int i;
  for (i = 0; i < n; i++) new[i] = (1 - learnrt) * old[i] + learnrt * new[i];
#ifndef NDEBUG
  float total = 0.0;
  for (i = 0; i < n; i++) total += new[i];
  if (!(.95 <= total && total <= 1.05))
    {
      for (i = 0; i < n; i++) printf("%f ", new[i]);
      printf("\n");
    }
  assert(.95 <= total && total <= 1.05);
#endif
}

/* Damp after normalization, and not before.
 */
static void
DiscDamp(struct plan7_s *hmm0, struct plan7_s *hmm, float learnrt)
{
  int i;                /* counter for states                 */
  for (i = 1; i < hmm->M ; i++)
    {
      if (i < hmm->M - 1)
        Damp1Simplex(3, &(hmm->t[i][TMM]), learnrt, &(hmm0->t[i][TMM]));
      else
        Damp1Simplex(2, &(hmm->t[i][TMM]), learnrt, &(hmm0->t[i][TMM]));
      Damp1Simplex(2, &(hmm->t[i][TIM]), learnrt, &(hmm0->t[i][TIM]));
      if (i > 1 && i < hmm->M-1)
        Damp1Simplex(2, &(hmm->t[i][TDM]), learnrt, &(hmm0->t[i][TDM]));
      Damp1Simplex(Alphabet_size, hmm->mat[i], learnrt, hmm0->mat[i]);
      Damp1Simplex(Alphabet_size, hmm->ins[i], learnrt, hmm0->ins[i]);
    }
  Damp1Simplex(Alphabet_size, hmm->mat[hmm->M], learnrt, hmm0->mat[hmm->M]);
}

/* Purpose:  MMIE discriminative training.
 *
 * Args:     ncate      - number of categories
 *           catenseq   - number of sequences of each category
 *           catedsqs   - array of sequences of each category
 *           catesqinfo - array of sequence information of each category
 *           oldhmms    - array of old HMMs (unchanged), 0..ncate
 *           prior      - profile HMM prior, used in initialization
 *           learnrt    - learning rate
 *           maxiter    - maximum iterations
 *           verbose    - verbose output
 *           ret_logcondlik - return averaged log conditional likelihood
 *
 * Return:   (struct plan7_s **) a new HMM array
 */
struct plan7_s **
DiscTrainOptimize(int ncate, int *catenseq, unsigned char ***catedsqs,
    SQINFO **catesqinfo, struct plan7_s **oldhmms, struct p7prior_s *prior,
    float learnrt, int maxiter, int verbose, float *ret_logcondlik)
{
  int    iter, k, idx;        /* index of iteration, category, and sequence */
  int    nseqs = 0;           /* total number of sequences across cateogries */
  int    trncorrect = 0;      /* correct training samples */
  struct plan7_s **hmms;      /* current HMM estimation */
  struct plan7_s **newhmms;   /* re-estimated HMM */
  float  hmmdistance;         /* measure of parameter change between two models */
  double distance_threshold = 0.001 * ncate;/* convergence criterion */

  hmms = (struct plan7_s **) MallocOrDie(ncate * sizeof(struct plan7_s *));
  newhmms = (struct plan7_s **) MallocOrDie(ncate * sizeof(struct plan7_s *));

  for (k = 0; k < ncate; k++) { nseqs += catenseq[k]; hmms[k] = oldhmms[k]; }

  if (verbose)
    printf("Iteration      length      change      score    accuracy\n---------      ------      ------      -----    --------\n");

  for (iter = 1; iter <= maxiter && !killnow; iter++)
    {
      float logcondlik;
      if (verbose >= 2) for (k = 0; k < ncate; k++) PrintHMM(hmms[k]);

      /* E-step: collect full forward-backward matrices (sufficient stat) */
      for (k = 0; k < ncate; k++)
        {
          /* allocate for a counts-based HMM */
          if (NULL == (newhmms[k] = AllocPlan7(hmms[k]->M)))
            Die("Failed to allocate for a new HMM\n");
          ZeroPlan7(newhmms[k]);
          Plan7SetNullModel(newhmms[k], hmms[k]->null, hmms[k]->p1);
        }
      //printf("DiscCountSuffStat...\n");
      logcondlik = DiscCountSuffStat(ncate, catenseq, catedsqs, catesqinfo,
                                     hmms, newhmms, &trncorrect);
      if (logcondlik == 0.0) /* if overflow, break without freeing hmm */
        {
          FreeHMMs(ncate,newhmms); printf("Terminate due to overflow\n"); break;
        }

      /* M-step: normalize an existing counts-based model (parameters) */
      //printf("M-step...\n");
      hmmdistance = 0.;
      for (k = 0; k < ncate; k++)
        {
          //if (verbose >= 2) PrintHMM(newhmms[k]);
          P7PriorifyHMM(newhmms[k], prior);
          //printf("DiscShiftPositive(hmms[%d])...\n", k);
          DiscShiftPositive(hmms[k], newhmms[k]);
          P7RetractWing(newhmms[k]);
          Plan7Renormalize(newhmms[k]);
          Plan7GlobalConfig(newhmms[k]);
          //printf("DiscDamp(hmms[%d])...\n", k);
          DiscDamp(hmms[k], newhmms[k], learnrt);
          if (newhmms[k]->M != hmms[k]->M) hmmdistance = FLT_MAX;
          else hmmdistance += P7Distance(newhmms[k], hmms[k]);
        }

      /* Verbose output */
      if (verbose)
        {
          printf(" %4d ::    ", iter);            /* what iteration             */
          printf("   %6d   ", newhmms[0]->M);     /* new model length           */
          if (FLT_MAX - hmmdistance > FLT_EPSILON)
            printf("%9.4f   ", hmmdistance);      /* change from old model      */
          else printf("     **     ");            /* (architecture changed)     */
          printf("%8.2f    ", logcondlik / nseqs);/* average score of old model */
          printf("%8.2f    ", (float) trncorrect / nseqs);/* training accuracy  */
          (void)puts("");
        }
      /* Switch new model to current model */
      if (iter > 1) FreeHMMs(ncate, hmms);
      for (k = 0; k < ncate; k++) hmms[k] = newhmms[k];
      /* Check for convergence. */
      if (hmmdistance < distance_threshold) break;
    }
  free(newhmms); /* HMMs in this array are copied to hmms */
  *ret_logcondlik = DiscCountSuffStat(ncate, catenseq, catedsqs, catesqinfo,
                    hmms, NULL, &trncorrect) / nseqs;
  return hmms;
}

/* Purpose:  MMIE corrective training. Start from average and random positions,
 *           run generative training for given iterations, and then run
 *           discriminative training. Return the model with maximum conditional
 *           likelihood.
 *
 * Args:     ncate      - number of categories
 *           catenseq   - number of sequences of each category
 *           catedsqs   - array of sequences of each category
 *           catesqinfo - array of sequence information of each category
 *           setlength  - profile HMM length
 *           learnrt    - learning rate
 *           ninitg     - number of random init for generative training
 *           ninitd     - number of random init for discriminative training
 *           maxiterg   - maximum iterations for generative training
 *           maxiterd   - maximum iterations for discriminative training
 *           verbose    - verbose output
 *           tophmmg    - pre-allocated array to store generative trained HMMs
 *
 * Return:   (struct plan7_s **) discriminative trained HMMs,
 *           NULL if maxiterd == 0 (but caller still has tophmmg)
 */
struct plan7_s **
DiscTrain(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo,
          int setlength, float learnrt, int ninitg, int ninitd,
          int maxiterg, int maxiterd, float inslimit, float avglen, int verbose,
          struct plan7_s **tophmmg)
{
  struct p7prior_s *prior  = NULL; /* prior for HMM                         */
  struct plan7_s **tophmmd = NULL; /* top disc HMM, to be returned          */
  struct plan7_s **hmmg    = NULL; /* generative HMMs (in each init)        */
  struct plan7_s **hmmd    = NULL; /* generative HMMs (in each init)        */
  float *toploglik         = NULL; /* top log-likelihood                    */
  float  loglik            = 0.;   /* log2 likelihood of each HMM           */
  float  max_logcondlik    = 0.;   /* top log conditional likelihood        */
  float  logcondlik        = 0.;   /* log2 conditional likelihood           */
  int    k;                        /* category index                        */
  int    iinit;                    /* init index                            */
  int    topinit;                  /* the top init index (zero-start best?) */

  hmmg = (struct plan7_s **) MallocOrDie(ncate * sizeof(struct plan7_s *));
  if (NULL != tophmmg) toploglik = (float *) MallocOrDie(ncate * sizeof(float));
  prior = P7DefaultPrior();
  for (iinit = 0; iinit < ninitd; iinit++)
    {
      if (verbose) printf("Init set %d/%d for disc training...\n", iinit, ninitd);
      /* FIXME maxiterg == 0 never tested */
      for (k = 0; k < ncate; k++)
        {
          hmmg[k] = GenTrain(catenseq[k], catedsqs[k], catesqinfo[k], setlength,
                             maxiterg, ninitg, inslimit, avglen, verbose, &loglik);
          if (NULL != tophmmg && (0 == iinit || loglik > toploglik[k]))
            {
              if (0 < iinit) FreePlan7(tophmmg[k]);
              toploglik[k] = loglik; tophmmg[k] = hmmg[k];
            }
          /* non-top hmmg may be used in disc training so not freed now. */
        }
      if (maxiterd > 0)
        {
          hmmd = DiscTrainOptimize(ncate, catenseq, catedsqs, catesqinfo, hmmg,
                                   prior, learnrt, maxiterd, verbose, &logcondlik);
          if (0 == iinit || logcondlik - max_logcondlik > FLT_EPSILON)
            {
              max_logcondlik = logcondlik;
              if (0 < iinit) { FreeHMMs(ncate, tophmmd); free(tophmmd); }
              tophmmd = hmmd;
            }
          else {FreeHMMs(ncate, hmmd);}
        }
      if (maxiterd > 0 && hmmg[0] == hmmd[0]) continue; /* don't free hmmg */
      for (k = 0; k < ncate; k++)
        {
          /* if not returning top hmmg or this hmmg is not top, free now. */
          if (NULL == tophmmg || tophmmg[k] != hmmg[k]) FreePlan7(hmmg[k]);
        }
    }
  P7FreePrior(prior);
  if (NULL != tophmmg) free(toploglik);
  free(hmmg);
  return tophmmd;
}
