/* subfeat.c
 *
 * Load many profile HMMs and calculate the log likelihood ratio HMMs versus
 * a null model for each sequence.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: subfeat [-options] <seqfile> <featfile> <hmmprefix1> <hmmprefix2>...\n"
"Options:\n"
"   -n <num-motif>    : number of motifs, using random masking (default 10)\n"
"   -v <level>        : verbose level (default 0, least information)\n";

static struct opt_s OPTIONS[] = {
  { "-n", TRUE, sqdARG_INT    },
  { "-v", TRUE, sqdARG_INT    }
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_NMOTIF      = 10;

static char **HmmFilenames(int nhmmprefix, char **hmmprefix, int nmotif, int *ret_nhmm);
static void RunHMMsLLR(char *seqfile, int nhmm, char **hmmfiles, char *featfile, int verbose);

int
main(int argc, char **argv)
{
  char        *optname;                         /* name of option from Getopt() */
  char        *optarg;                          /* argument from Getopt()       */
  int          optind;                          /* index in argv[]              */
  int          verbose          = 0;
  int          nmotif           = DEFAULT_NMOTIF;
  int          nhmmprefix       = 0;
  char        *seqfile          = NULL;
  char        *featfile         = NULL;
  char       **hmmprefix        = NULL;
  char       **hmmfiles         = NULL;
  int          nhmm             = 0;
  int          ihmm             = 0;
  time_t       starttime        = 0;
  int          minutes          = 0;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-v") == 0) verbose       = atoi(optarg);
    else if (strcmp(optname, "-n") == 0) nmotif        = atoi(optarg);
    /*else if (strcmp(optname, "-s") == 0) seed          = atoi(optarg);*/
    else if (strcmp(optname, "-h") == 0)
      {
        printf("subfeat: generate LLR feature file from a collection of profile HMMs\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 3)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];
  featfile = argv[optind++];
  hmmprefix = argv + optind;
  nhmmprefix = argc - optind;
  hmmfiles = HmmFilenames(nhmmprefix, hmmprefix, nmotif, &nhmm);

  starttime = time((time_t *) NULL);
  RunHMMsLLR(seqfile, nhmm, hmmfiles, featfile, verbose);
  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  printf("Finished in %d:%d\n", minutes / 60, minutes % 60);

  for (ihmm = 0; ihmm < nhmm; ihmm++) free(hmmfiles[ihmm]);
  free(hmmfiles);
  return 0;
}

/***********************************************
 * Filenames
 ***********************************************/

static char **
HmmFilenames(int nhmmprefix, char **hmmprefix, int nmotif, int *ret_nhmm)
{
  int nhmm = nhmmprefix * nmotif;
  char **hmmfiles = (char **) MallocOrDie(sizeof(char *) * nhmm);
  int ihmm = 0, i, j;
  for (i = 0; i < nhmmprefix; i++)
    for (j = 0; j < nmotif; j++)
      {
        hmmfiles[ihmm] = (char *) MallocOrDie(strlen(hmmprefix[i]) + 7);
        sprintf(hmmfiles[ihmm], "%s%02d.hmm", hmmprefix[i], j);
        ihmm++;
      }
  assert(ihmm == nhmm);
  *ret_nhmm = nhmm;
  return hmmfiles;
}

/***********************************************
 * Likelihood ratio
 ***********************************************/

/* Calculate the likelihood of the null model.
 * Should be in a shared module if used by other programs.
 */
static float
CalNullModelScore(float null[MAXABET], float p1, unsigned char *dsq, int L)
{
  int i;
  float score = L * sreLOG2(p1) + sreLOG2(1.0 - p1);
  for (i = 1; i <= L; i++) score += sreLOG2(null[dsq[i]]);
  return score;
}

static double **
CalHMMsLLR(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, int nhmm,
           char **hmmfiles, int verbose)
{
  int idx, ihmm;
  double **llr = (double **) MallocOrDie(nseqs * sizeof(double *));
  if (verbose) printf("Allocating %d x %d LLR array...\n", nseqs, nhmm);
  for (idx = 0; idx < nseqs; idx++)
    llr[idx] = (double *) MallocOrDie(nhmm * sizeof(double));
  for (ihmm = 0; ihmm < nhmm; ihmm++)
    {
      if (verbose) printf("Processing HMM file %s...\n", hmmfiles[ihmm]);
      struct plan7_s *hmm = LoadHMM(hmmfiles[ihmm]);
      if (verbose > 1)
        {
          int x;
          PrintHMM(hmm);
          printf("p1=%.3f, null= ", hmm->p1);
          for (x = 0; x < Alphabet_size; x++) printf("%f ", hmm->null[x]);
          printf("\n");
        }
      for (idx = 0; idx < nseqs; idx++)
        {
          struct mledpmat_s *fwd;
          float scl_score = MleForward(dsqs[idx], sqinfo[idx].len, hmm, &fwd);
          double loglik = MleFwdScore(fwd, sqinfo[idx].len, hmm->M, scl_score);
          double loglik0 = CalNullModelScore(hmm->null, hmm->p1, dsqs[idx],
                                            sqinfo[idx].len);
          if (verbose > 1) printf("%f - %f, ", loglik, loglik0);
          llr[idx][ihmm] = loglik - loglik0;
          FreeMleDpMat(fwd);
        }
      if (verbose > 1) printf("\n");
      FreePlan7(hmm);
    }
  return llr;
}

static void
RunHMMsLLR(char *seqfile, int nhmm, char **hmmfiles, char *featfile, int verbose)
{
  int             idx;
  int             nseqs         = 0;
  unsigned char **dsqs          = NULL;
  SQINFO         *sqinfo        = NULL;
  double        **llr           = NULL;

  SetAlphabet(hmmAMINO);
  if (verbose) printf("Loading sequence %s...\n", seqfile);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo); /* load sequence and digitize */
  llr = CalHMMsLLR(nseqs, dsqs, sqinfo, nhmm, hmmfiles, verbose);
  if (verbose) printf("Saving LLR features %s...\n", featfile);
  SaveFeatFile(nseqs, nhmm, llr, featfile);
  for (idx = 0; idx < nseqs; idx++) free(llr[idx]);
  free(llr);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}
