/* subio.c
 *
 * IO routines for subcellular localization motif finder.
 */

#include "config.h"
#include "squidconf.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>

#include "squid.h"
#include "ssi.h"
#include "structs.h"
#include "funcs.h"
#include "subloc.h"

/* Trim ending characters ( \t\r\n) by filing zeros. */
static void trim(char *str)
{
  char *cp = str + strlen(str) - 1;
  while (cp >= str && strchr(" \t\r\n", *cp) != NULL) *(cp--) = 0;
}

/***********************************************
 * FASTA sequence files
 ***********************************************/

/* Purpose:  Load sequence file and return the digitized sequence and SQINFO
 *           array.
 *
 * Args:     seqfile        - filename
 *           ret_nseq       - OUTPUT reference of number of sequence
 *           ret_sqinfo     - OUTPUT reference of SQINFO array
 *
 * Return:   (unsigned char **) digitized sequence array
 */
unsigned char **
LoadDsqs(char *seqfile, int *ret_nseq, SQINFO **ret_sqinfo)
{
  int idx;
  char **seqs = NULL;
  unsigned char **dsqs = NULL;
  SetAlphabet(hmmAMINO);
  /* read the unaligned seqs from file */
  if (0 == ReadMultipleRseqs(seqfile, SQFILE_FASTA, &seqs, ret_sqinfo, ret_nseq))
    Die("Failed to read sequence file %s", seqfile);
  /*if (0 == DetermineAlphabet(seqs, *ret_nseq))
    Die("Failed to determine alphabet used in seqs in %s", seqfile);*/

  dsqs = MallocOrDie(sizeof(unsigned char *) * (*ret_nseq));
  for (idx = 0; idx < *ret_nseq; idx++)
    {
      dsqs[idx] = DigitizeSequence(seqs[idx], (*ret_sqinfo)[idx].len);
      /*for (t = 0; t < sqinfo[idx].len; t++) printf("%d ", dsq[idx][t]);*/
      free(seqs[idx]);
      trim((*ret_sqinfo)[idx].desc);
    }
  free(seqs);

  return dsqs;
}

/* Purpose:  Load sequence files and return a combined sequence array
 *           and the corresponding categories.
 *
 * Args:     ncate     - number of categories
 *           filenames - array of filename string, 0..ncate
 *           ret_nseq  - OUTPUT reference of number of sequence
 *           ret_cates - OUTPUT reference of array of sequences
 *
 * Return:   (char **) array of loaded sequences
 */
unsigned char ***
LoadCateDsqs(int ncate, char **filenames, int **ret_catenseq,
             SQINFO ***ret_catesqinfo)
{
  int k;
  int *catenseq = (int *) MallocOrDie(ncate * sizeof(int));
  unsigned char ***catedsqs = (unsigned char ***) MallocOrDie(
                              ncate * sizeof(unsigned char**));
  SQINFO **catesqinfo = (SQINFO **) MallocOrDie(ncate * sizeof(SQINFO*));
  for (k = 0; k < ncate; k++)
    catedsqs[k] = LoadDsqs(filenames[k], catenseq + k, catesqinfo + k);
  *ret_catenseq = catenseq; *ret_catesqinfo = catesqinfo; return catedsqs;
}

/* Purpose:  Return whether a FASTA seq description matches a label (compartment).
 *           Called by LoadSeqComp() only.
 */
static int
IsLabel(char *label, char *desc)
{
  return (strstr(label, desc) != NULL);
}

/* Purpose:  Load sequence file and return the categorized array of
 *           sequence / SQINFO array matching the given labels (compartments).
 *           The description field of the name (greater-than-line) MUST
 *           all match a label.
 *
 * Args:     seqfile        - filename
 *           ncate          - number of categories
 *           labels         - array of label string, 0..ncate
 *           ret_catenseq   - OUTPUT reference of categorized number of sequence
 *           ret_catesqinfo - OUTPUT reference of categorized SQINFOs
 *
 * Return:   (char ***) array of categorized digitized sequences
 */
unsigned char ***
LoadLabeledDsqs(char *seqfile, int ncate, char **labels, int **ret_catenseq,
                SQINFO ***ret_catesqinfo)
{
  int idx, k;
  int nseq = 0;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = LoadDsqs(seqfile, &nseq, &sqinfo);
  int *catenseq = (int *) MallocOrDie(ncate * sizeof(int));
  unsigned char ***catedsqs = (unsigned char ***) MallocOrDie(
                              ncate * sizeof(unsigned char**));
  SQINFO **catesqinfo = (SQINFO **) MallocOrDie(ncate * sizeof(SQINFO*));
  for (k = 0; k < ncate; k++) catenseq[k] = 0;
  for (idx = 0; idx < nseq; idx++)
    for (k = 0; k < ncate; k++)
      if (IsLabel(labels[k], sqinfo[idx].desc)) {catenseq[k]++; break;}
  for (k = 0; k < ncate; k++)
    {
      catedsqs[k] = (unsigned char **) MallocOrDie(catenseq[k] *
                                                   sizeof(unsigned char*));
      catesqinfo[k] = (SQINFO *) MallocOrDie(catenseq[k] * sizeof(SQINFO));
      catenseq[k] = 0;/* used as counter later and will regain same value */
    }
  for (idx = 0; idx < nseq; idx++)
    {
      for (k = 0; k < ncate; k++)
        if (IsLabel(labels[k], sqinfo[idx].desc))
          {
            catedsqs[k][catenseq[k]] = dsqs[idx];
            catesqinfo[k][catenseq[k]] = sqinfo[idx];
            catenseq[k]++;
            break;
          }
      if (k == ncate) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));//not in labels
    }
  free(dsqs); free(sqinfo);
  *ret_catenseq = catenseq; *ret_catesqinfo = catesqinfo; return catedsqs;
}

/* Free digitized sequence and sqinfo. Modified from SQUID code */
void
FreeDsqinfo(unsigned char *dsq, SQINFO *sqinfo)
{
  if (dsq != NULL) free(dsq);
  if (sqinfo->flags & SQINFO_SS) free(sqinfo->ss);
  if (sqinfo->flags & SQINFO_SA) free(sqinfo->sa);
}

/* Purpose:  Free categorized sequence array.
 *
 * Args:     ncate          - number of categories
 *           catenseq       - categorized number of sequence
 *           catedsqs       - categorized digitized sequence array
 *           catesqinfo     -  categorized SQINFO array
 */
void
FreeCateDsqs(int ncate, int *catenseq, unsigned char ***catedsqs,
             SQINFO **catesqinfo)
{
  int k, idx;
  for (k = 0; k < ncate; k++)
    {
      for (idx = 0; idx < catenseq[k]; idx++)
        FreeDsqinfo(catedsqs[k][idx], &(catesqinfo[k][idx]));
      free(catedsqs[k]); free(catesqinfo[k]);
    }
  free(catedsqs); free(catesqinfo); free(catenseq);
}

/***********************************************
 * Profile HMM files
 ***********************************************/

/* Load HMM file. */
struct plan7_s *
LoadHMM(char *hmmfile)
{
  /***********************************************
   * Open HMM file (might be in HMMERDB or current directory).
   * Read a single HMM from it. (Config HMM, if necessary).
   * Alphabet globals are set by reading the HMM.
   ***********************************************/
  struct plan7_s *hmm;
  HMMFILE *hmmfp;
  if ((hmmfp = HMMFileOpen(hmmfile, "HMMERDB")) == NULL)
    Die("Failed to open HMM file %s\n", hmmfile);
  if (!HMMFileRead(hmmfp, &hmm))
    Die("Failed to read any HMMs from %s\n", hmmfile);
  if (hmm == NULL)
    Die("HMM file %s corrupt or in incorrect format? Parse failed", hmmfile);
  return hmm;
}

/* Purpose:  Load hmm files.
 *
 * Args:     nhmm      - number of HMMs
 *           filenames - array of filename string, 0..nhmm
 *
 * Return:   (struct plan7_s **) array of loaded HMM
 */
struct plan7_s **
LoadHMMs(int nhmm, char **filenames)
{
  int ihmm;
  struct plan7_s **hmms = (struct plan7_s **) MallocOrDie(
                          nhmm * sizeof(struct plan7_s *));
  for (ihmm = 0; ihmm < nhmm; ihmm++)
    {
      hmms[ihmm] = LoadHMM(filenames[ihmm]);
      P7Logoddsify(hmms[ihmm], 0); /* 0: posterior decoding, not viterbi */
    }
  return hmms;
}

/* Save HMM file. */
void
SaveHMM(struct plan7_s *hmm, char *hmmfile)
{
  FILE *hmmfp;
  /* save as text file (I don't know why binary file cannot be read by HMMER) */
  if ((hmmfp = fopen(hmmfile, "w")) == NULL)
    Die("Failed to open %s for writing hmm result\n", hmmfile);
  WriteAscHMM(hmmfp, hmm);
  fclose(hmmfp);
}

/* Purpose:  Save HMM files.
 *
 * Args:     nhmm     - number of HMMs
 *           hmms     - array of HMMs
 *           hmmfiles - array of filename string
 */
void
SaveHMMs(int nhmm, struct plan7_s **hmms, char **hmmfiles)
{
  int ihmm;
  for (ihmm = 0; ihmm < nhmm; ihmm++) SaveHMM(hmms[ihmm], hmmfiles[ihmm]);
}

/* Free an array of HMM, but do NOT free the array.
 * Caller must execute free(hmms) to free the memory of the array
 */
void
FreeHMMs(int ncate, struct plan7_s **hmms)
{
  int k;
  for (k = 0; k < ncate; k++) FreePlan7(hmms[k]);
}

void
PrintHMM(struct plan7_s *hmm)
{
  int k, ts, x;
  /* printf("p1=%.3f, null= ", hmm->p1);
  for (x = 0; x < Alphabet_size; x++) printf("%f ", hmm->null[x]); */
  printf(" MM MI MD IM II DM DD ");
  /* Match emissions */
  for (x = 0; x < Alphabet_size; x++) printf(" %c", Alphabet[x]);
  printf(" "); /* Insert emissions */
  for (x = 0; x < Alphabet_size; x++) printf(" %c", Alphabet[x]);
  printf("\n");
  for (k = 1; k <= hmm->M; k++)
    {
      if (k < hmm->M) {
        for (ts = 0; ts < 7; ts++) printf("%3.0f", 100. * hmm->t[k][ts]);
      } else printf("                     ");
      printf(",");
      for (x = 0; x < Alphabet_size; x++)
        printf("%2.0f", 10. * hmm->mat[k][x]);
      printf(",");
      if (k < hmm->M)
        for (x = 0; x < Alphabet_size; x++)
          printf("%2.0f", 10. * hmm->ins[k][x]);
      printf("\n");
    }
}

/***********************************************
 * DP matrix
 ***********************************************/

void
PrintMleDpMat(struct mledpmat_s *fwd, int L, int M)
{
  int i, k;
  if (NULL != fwd->scl)
    {
      printf("scl: ");
      for (i = 0; i <= L; i++) printf("%f ", fwd->scl[i]);
      printf("\n");
    }
  printf("mmx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k <= M; k++) printf("%e ", fwd->mmx[i][k]);
      printf("\n");
    }
  printf("imx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k <= M; k++) printf("%e ", fwd->imx[i][k]);
      printf("\n");
    }
  printf("dmx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k <= M; k++) printf("%e ", fwd->dmx[i][k]);
      printf("\n");
    }
}

void
WarnMleDpMat(struct mledpmat_s *fwd, int L, int M)
{
  int i, k, tobreak = 0;
  if (NULL != fwd->scl)
    for (i = 0; i <= L; i++)
      if (!(fwd->scl[i] >= 0.)) {printf("scl[%d]=%f\n", i, fwd->scl[i]); break;}
  for (i = 0; i <= L; i++)
    {
      if (!(fwd->xmx[i][XMN] >= 0.)) {printf("XMN[%d]=%f\n", i, fwd->xmx[i][XMN]); tobreak=1;}
      if (!(fwd->xmx[i][XMB] >= 0.)) {printf("XMB[%d]=%f\n", i, fwd->xmx[i][XMB]); tobreak=1;}
      if (!(fwd->xmx[i][XMJ] >= 0.)) {printf("XMJ[%d]=%f\n", i, fwd->xmx[i][XMJ]); tobreak=1;}
      if (!(fwd->xmx[i][XME] >= 0.)) {printf("XME[%d]=%f\n", i, fwd->xmx[i][XME]); tobreak=1;}
      if (!(fwd->xmx[i][XMC] >= 0.)) {printf("XMC[%d]=%f\n", i, fwd->xmx[i][XMC]); tobreak=1;}
      if (tobreak) break;
    }
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k < M; k++)
        if (!(fwd->mmx[i][k] >= 0.)) {
          printf("mmx[%d][%d]=%f\n", i, k, fwd->mmx[i][k]); i=L+1; break;
        }
    }
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k < M; k++)
        if (!(fwd->imx[i][k] >= 0.)) {
          printf("imx[%d][%d]=%f\n", i, k, fwd->imx[i][k]); i=L+1; break;
        }
    }
  for (i = 0; i <= L; i++)
    {
      for (k = 2; k < M; k++)
        if (!(fwd->dmx[i][k] >= 0.)) {
          printf("dmx[%d][%d]=%f\n", i, k, fwd->dmx[i][k]); i=L+1; break;
        }
    }
}

static void
PrintDpMatrix(struct dpmatrix_s *fwd, int L, int M)
{
  int i, k;
  printf("mmx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k <= M; k++) printf("%d ", fwd->mmx[i][k]);
      printf("\n");
    }
  printf("imx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 1; k <= M; k++) printf("%d ", fwd->imx[i][k]);
      printf("\n");
    }
  printf("dmx\n");
  for (i = 0; i <= L; i++)
    {
      for (k = 2; k <= M; k++) printf("%d ", fwd->dmx[i][k]);
      printf("\n");
    }
}

static void
WarnDpMatrix(struct dpmatrix_s *fwd, int L, int M)
{
  int i, k;
  for (i = 1; i < L; i++)
    {
      for (k = 1; k < M; k++)
        if (!(fwd->mmx[i][k] > -20000000)) printf("mmx[%d][%d]=%d\n", i, k, fwd->mmx[i][k]);
    }
  for (i = 2; i < L; i++)
    {
      for (k = 1; k < M; k++)
        if (!(fwd->imx[i][k] > -20000000)) printf("imx[%d][%d]=%d\n", i, k, fwd->imx[i][k]);
    }
  for (i = 2; i < L; i++)
    {
      for (k = 2; k < M; k++)
        if (!(fwd->dmx[i][k] > -20000000)) printf("dmx[%d][%d]=%d\n", i, k, fwd->dmx[i][k]);
    }
}

void
DedigitizeSeqBuf(unsigned char *dsq, int L, char *buf)
{
  int i;
  for (i = 0; i < L; i++) buf[i] = Alphabet[dsq[i+1]];
  buf[L] = '\0';
}

/***********************************************
 * LLR feature files
 ***********************************************/

void SaveFeatFile(int nseqs, int nhmm, double **llr, char *featfile)
{
  int idx, ihmm;
  FILE *fp = fopen(featfile, "w");
  for (idx = 0; idx < nseqs; idx++)
    {
      for (ihmm = 0; ihmm < nhmm; ihmm++) fprintf(fp, "%f\t", llr[idx][ihmm]);
      fprintf(fp, "\n", llr[idx][ihmm]);
    }
  fclose(fp);
}

/***********************************************
 * Motif instance files
 ***********************************************/

static char locfile_header[] =
"********************************************************************************\n"
"Motif discovery by profile HMM\n"
"\n"
"nmotif = %d, motiflen = %d, maxiterg = %d, maxiterd = %d, ninitg = %d, ninitd = %d, seed = %d\n"
"********************************************************************************\n"
"\n";

static char locfile_bottom[] =
"********************************************************************************\n"
"Finished in %d:%d\n";

void
SaveLocHeader(char *locfile, int nmotif, int setlength, int maxiterg,
              int maxiterd, int ninitg, int ninitd, int seed)
{
  FILE *fp = fopen(locfile, "a");
  if (NULL == fp)
    Die("Failed to write to location file %s: %s", locfile, strerror(errno));
  fprintf(fp, locfile_header, nmotif, setlength, maxiterg, maxiterd, ninitg,
          ninitd, seed);
  fclose(fp);
}

void SaveLocBottom(char *locfile, int minutes)
{
  FILE *fp = fopen(locfile, "a");
  if (NULL == fp)
    Die("Failed to write to location file %s: %s", locfile, strerror(errno));
  fprintf(fp, locfile_bottom, minutes / 60, minutes % 60);
  fclose(fp);
}

static char motif_sites[] =
"--------------------------------------------------------------------------------\n"
"        Motif %d sites sorted by position p-value\n"
"--------------------------------------------------------------------------------\n"
"Sequence name             Start   P-value             Site \n"
"-------------             ----- ---------            ------\n";

static char motif_blockdiag[] =
"--------------------------------------------------------------------------------\n"
"        Motif %d block diagrams\n"
"--------------------------------------------------------------------------------\n"
"SEQUENCE NAME            POSITION P-VALUE  MOTIF DIAGRAM\n"
"-------------            ----------------  -------------\n";

static char motif_separator[] =
"--------------------------------------------------------------------------------\n\n";

static char leftstr[11];
static char rightstr[11];
static char matstr[100];

void
SaveLoc(char *locfile, int nseqs, unsigned char **dsqs, SQINFO *sqinfo,
        struct matlocs_s *matlocs, int imotif)
{
  int i;
  FILE *fp = fopen(locfile, "a");
  fprintf(fp, motif_sites, imotif);
  for (i = 0; i < nseqs; i++)
    {
      int b = matlocs[i].begin[0];
      int e = matlocs[i].end[0];
      int b1 = MAX(1, b - 10);
      int e1 = MIN(sqinfo[i].len, e + 10);
      DedigitizeSeqBuf(dsqs[i] + b - 1, e - b + 1, matstr);
      DedigitizeSeqBuf(dsqs[i] + b1 - 1, b - b1, leftstr);
      DedigitizeSeqBuf(dsqs[i] + e, e1 - e, rightstr);
      fprintf(fp, "%-26s %4d  1.00e-02 %10s %s %-10s\n", sqinfo[i].name, b,
              leftstr, matstr, rightstr);
    }
  fprintf(fp, motif_separator);

  fprintf(fp, motif_blockdiag, imotif);
  for (i = 0; i < nseqs; i++)
    {
      if (1 == matlocs[i].begin[0])
        fprintf(fp, "%-26s        1.0e-02  [%d]_%d\n", sqinfo[i].name,
                imotif, sqinfo[i].len - matlocs[i].end[0]);
      else if (sqinfo[i].len == matlocs[i].end[0])
        fprintf(fp, "%-26s        1.0e-02  %d_[%d]\n", sqinfo[i].name,
                matlocs[i].begin[0] - 1, imotif);
      else
        fprintf(fp, "%-26s        1.0e-02  %d_[%d]_%d\n", sqinfo[i].name,
                matlocs[i].begin[0]-1, imotif, sqinfo[i].len-matlocs[i].end[0]);
    }
  fprintf(fp, motif_separator);

  fclose(fp);
}

/***********************************************
 * IO related utilities
 ***********************************************/

/* Create an array of strings containing filenames matching an extension.
 * Ex. dir="CuTest/", ext=".ext", returns {"CuTest/exttest1.ext"...}
 * Note that dir must either be an empty string, or ends with a splash.
 * Caller must free the allocated array, including the filename strings.
 * Callers: subfeat.c (*.hmm)
 */
char **
ListFilesWithExt(char *dir, char *ext, int *ret_nfiles)
{
  char **files;
  DIR *dirp;
  struct dirent *entry;
  int dirlen = strlen(dir);
  int extlen = strlen(ext);
  int nfiles = 0;
  int i = 0;

  dirp = opendir(dir);
  if (NULL == dirp) Die("Error opening directory: %s\n", dir);
  while (entry = readdir(dirp))
    {
      if (0 == strcmp(ext, entry->d_name + strlen(entry->d_name) - extlen))
        nfiles++;
    }
  files = (char **) MallocOrDie(nfiles * sizeof(char*));
  rewinddir(dirp);
  while (entry = readdir(dirp))
    {
      int namelen = strlen(entry->d_name);
      if (0 == strcmp(ext, entry->d_name + namelen - extlen))
        {
          files[i] = (char*) MallocOrDie(dirlen + namelen + 1);
          strcpy(files[i], dir);
          strcat(files[i], entry->d_name);
          i++;
        }
    }
  assert(i == nfiles);
  (void) closedir(dirp);
  *ret_nfiles = nfiles;
  return files;
}

/***********************************************
 * Interrupt handling, for ctrl-c and kills
 ***********************************************/

int killnow = (int) FALSE; /* flag for SIGINT catching */

/* Function: handle_interrupt()
 *
 * Purpose:  Handle ctrl-C and kills. Set global flag "killnow",
 *           which will cause this to be the final iteration.
 *           Do not reset the handler, though, so a second signal
 *           will really kill the process.
 */
static void
handle_interrupt(int sig)
{
  if (sig == SIGTERM || sig == SIGINT)
    {
      killnow = TRUE;
      Warn("Caught a lethal interrupt signal. Will stop after this iteration.");
    }
}

void
setup_interrupt()
{
  /* Catch signals gracefully, allowing users to kill runs without
   * losing models. SIGINT is an interactive interrupt (e.g. ctrl-C).
   * SIGTERM is a 'kill'.
   */
  if (signal(SIGINT,  SIG_IGN) != SIG_IGN) signal(SIGINT,  handle_interrupt);
  if (signal(SIGTERM, SIG_IGN) != SIG_IGN) signal(SIGTERM, handle_interrupt);
}
