#ifndef SUBLOC_H
#define SUBLOC_H

/* Declaration of dynamic programming matrix structure,
 * for both forward and backward matrix.
 * The only difference between this structure and the dpmatrix_s
 * is that float is used instead of int (and a scale vector),
 * since true scaled probability is used instead of log.
 * Note that, just like HMMER 2.3 dpmatrix_s, N = L+1.
 * This is how forward and backward is done in HMMER 1.8.
 */
struct mledpmat_s {
  float *scl;                   /* scales         [0.1..N]             */

  float **xmx;                  /* special scores [0.1..N][BECJN]      */
  float **mmx;                  /* match   scores [0.1..N][0.1..M]     */
  float **imx;                  /* insert  scores [0.1..N][0.1..M-1.M] */
  float **dmx;                  /* delete  scores [0.1..N][0.1..M-1.M] */

  /* Hidden ptrs where the real memory is kept; this trick was
   * introduced by Erik Lindahl with the Altivec port; it's used to
   * align xmx, etc. on 16-byte boundaries for cache optimization.
   */
  void *xmx_mem, *mmx_mem, *imx_mem, *dmx_mem;

  /* The other trick brought in w/ the Lindahl Altivec port; dp matrix
   * is retained and grown, rather than reallocated for every HMM or sequence.
   * Keep track of current allocated-for size in rows (sequence length N)
   * and columns (HMM length M). Also keep track of pad sizes: how much
   * we should overallocate rows or columns when we reallocate. If pad = 0,
   * then we're not growable in this dimension.
   */
  int maxN;                     /* alloc'ed for seq of length N; N+1 rows */
  int maxM;                     /* alloc'ed for HMM of length M; M+1 cols */

  int padN;                     /* extra pad in sequence length/rows */
  int padM;                     /* extra pad in HMM length/columns   */
};

/* Match location array structure */
struct matlocs_s {
  int n;                        /* The number of matches */
  int *begin;                   /* Beginings, between 1 and L, [0..n-1] */
  int *end;                     /* Endings  , between 1 and L, [0..n-1] */
};

/* Match location structure for posterior decoding or similar methods */
struct matloc_s {
  int seqidx;
  char *seqname;
  int begin;
  int end;
  float score;
};

/* Functions
 */

/* IO */
unsigned char **LoadDsqs(char *seqfile, int *ret_nseq, SQINFO **ret_sqinfo);
unsigned char ***LoadCateDsqs(int ncate, char **filenames, int **ret_catenseq, SQINFO ***ret_catesqinfo);
unsigned char ***LoadLabeledDsqs(char *seqfile, int ncate, char **labels, int **ret_catenseq, SQINFO ***ret_catesqinfo);
void FreeDsqinfo(unsigned char *dsq, SQINFO *sqinfo);
void FreeCateDsqs(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo);
struct plan7_s *LoadHMM(char *hmmfile);
struct plan7_s **LoadHMMs(int nhmm, char **filenames);
void SaveHMM(struct plan7_s *hmm, char *hmmfile);
void SaveHMMs(int nhmm, struct plan7_s **hmms, char **hmmfiles);
void FreeHMMs(int ncate, struct plan7_s **hmms);
void PrintHMM(struct plan7_s *hmm);
void SaveFeatFile(int nseqs, int nhmm, double **llr, char *featfile);
void SaveLocHeader(char *locfile, int nmotif, int setlength, int maxiterg, int maxiterd, int ninitg, int ninitd, int seed);
void SaveLocBottom(char *locfile, int minutes);
void SaveLoc(char *locfile, int nseqs, unsigned char **dsqs, SQINFO *sqinfo, struct matlocs_s *matlocs, int imotif);
char **ListFilesWithExt(char *dir, char *ext, int *ret_nfiles);

/* Basic */
int argmaxf(int n, float  *arr);
int argmaxd(int n, double *arr);
int argmaxi(int n, int    *arr);
int    maxi(int n, int    *arr);
float  maxf(int n, float  *arr);
float  minf(int n, float  *arr);
int    sumi(int n, int    *arr);
float  sumf(int n, float  *arr);
float  avgi(int n, int    *arr);
float  avgf(int n, float  *arr);
void floatsadd(int n, float *f, float x);

/* Classification and cross-validation */
float *Classify(int nhmm, struct plan7_s **hmms, unsigned char *dsq, int L);
float EvaluateSeqs(int ncate, struct plan7_s **hmms, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo);
void CrossValidate(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo, int nfold, int setlength, float learnrt, int ninit, int maxiterg, int maxiterd, float inslimit, float avglen, int verbose, float *accgen, float *accdisc);

/* Forward-backward algorithm */
float MleForward(unsigned char *dsq, int L, struct plan7_s *hmm, struct mledpmat_s **ret_mx);
float MleBackward(unsigned char *dsq, int L, struct plan7_s *hmm, float *scl, struct mledpmat_s **ret_mx);
double MleFwdScore(struct mledpmat_s *fwd, int L, int M, double sclprob);
void PrintMleDpMat(struct mledpmat_s *fwd, int L, int M);
void WarnMleDpMat(struct mledpmat_s *fwd, int L, int M);

/* Generative training */
struct p7prior_s *P7SimplePrior(void);
float P7Distance(struct plan7_s *newhmm, struct plan7_s *oldhmm);
struct plan7_s *P7Init(int setlength, struct p7prior_s *prior, float *randomseq, float p1, int nseqs, unsigned char **dsqs, SQINFO *sqinfo);
void P7RetractWing(struct plan7_s *hmm);
void MleForbackCount(struct plan7_s *hmm, unsigned char *dsq, int L, float score, float weight, struct mledpmat_s *fwd, struct mledpmat_s *bck, struct plan7_s *count);
float TrainCountSuffStat(int nseqs, unsigned char **dsq, SQINFO *sqinfo, float *wt, struct plan7_s *hmm, struct plan7_s *newhmm);
struct plan7_s *TrainOptimize(struct plan7_s *hmm, int nseqs, unsigned char **dsq, SQINFO *sqinfo, int setlength, struct p7prior_s *prior, int maxiter, float inslimit, int verbose, float *ret_score);
struct plan7_s *GenTrain(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, int setlength, int maxiter, int ninit, float inslimit, float avglen, int verbose, float *ret_score);

/* Discriminative training */
struct plan7_s **DiscTrainOptimize(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo, struct plan7_s **oldhmms, struct p7prior_s *prior, float learnrt, int maxiter, int verbose, float *ret_logcondlik);
struct plan7_s **DiscTrain(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo, int setlength, float learnrt, int ninitg, int ninitd, int maxiterg, int maxiterd, float inslimit, float avglen, int verbose, struct plan7_s **tophmmg);

/* Scanning by Viterbi and Posterior */
void FreeMatlocs(struct matlocs_s *matlocs);
void Trace2Loc(struct matlocs_s *matlocs, struct p7trace_s *tr, int M);
void ScanMatches(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, struct plan7_s *hmm, struct matlocs_s *matlocs, struct p7trace_s **tr_arr);
void PosteriorBE(unsigned char *dsq, int L, struct plan7_s *hmm, float *begin, float *end);

/* Wrappers */
void RunGenTrain(char *seqfile, char *hmmfile, int setlength, int maxiter, int ninit, float inslimit, float avglen, int verbose);
void RunCrossValidate(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo, int nfold, int nrepeat, int setlength, float learnrt, int ninit, int maxiterg, int maxiterd, float inslimit, float avglen, int seed, int verbose, char *title, char *resultfile);

extern int killnow;

void setup_interrupt();

#endif /* SUBLOC_H */
