/* submotif.c
 *
 * Perform generative training of HMM, and use it to scan for motifs.
 * Search for multiple motifs by masking found motif instances with
 * random background sequences.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: submotif [-options] <SEQFILES>\n"
"Options:\n"
"   -l <length>       : profile HMM (match-states) length (default 10)\n"
"   -n <num-motif>    : number of motifs, using random masking (default 1)\n"
"   -i <iterations>   : maximum iterations, generative (default 20)\n"
"   -I <iterations>   : maximum iterations, discriminative (default 0)\n"
"   -b <begin>        : random beginnings, generative (default 1)\n"
"   -B <begin>        : random beginnings, discriminative (default 1)\n"
"   -r <learn-rate>   : learning rate (default 0.01)\n"
"   -m <max-insert>   : maximum insertions (default 20.0)\n"
"   -a <avg-length>   : average length for background (default 530.0)\n"
"   -s <seed>         : random seed (default by random)\n"
"   -v <level>        : verbose level (default 0, least information)\n"
"   -o <location-dir> : output location directory (default current directory)\n"
"   -g <gen-hmm-dir>  : output gen-HMM directory (default current directory)\n"
"   -d <disc-hmm-dir> : output disc-HMM directory (default current directory)\n";

static struct opt_s OPTIONS[] = {
  { "-l", TRUE, sqdARG_INT    },
  { "-i", TRUE, sqdARG_INT    },
  { "-b", TRUE, sqdARG_INT    },
  { "-n", TRUE, sqdARG_INT    },
  { "-m", TRUE, sqdARG_FLOAT  },
  { "-a", TRUE, sqdARG_FLOAT  },
  { "-s", TRUE, sqdARG_INT    },
  { "-v", TRUE, sqdARG_INT    },
  { "-o", TRUE, sqdARG_STRING },
  { "-g", TRUE, sqdARG_STRING },
  { "-d", TRUE, sqdARG_STRING },
  { "-r", TRUE, sqdARG_FLOAT  },
  { "-B", TRUE, sqdARG_INT    },
  { "-I", TRUE, sqdARG_INT    },
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_PHMM_LEN    = 10;
const float DEFAULT_LEARN_RATE  = 0.01;
const int   MAXITER_G           = 20;
const int   MAXITER_D           = 0;
const int   NINIT_G             = 1;
const int   NINIT_D             = 1;
const int   DEFAULT_NMOTIF      = 1;
const float DEFAULT_INSLIMIT    = 20.0;
const float DEFAULT_AVGLEN      = 530.0;

static char *LocFilename(char *locdir, char *seqfile);
static void HmmFilenames(char *hmmg_dir, char *hmmd_dir, int ncate, char **seqfiles, int nmotif, int setlength, char ****ret_hmmg_files, char ****ret_hmmd_files);
static void RunSearchMultiMotifs(int ncate, char **seqfiles, char **locfiles, char ***hmmg_files, char ***hmmd_files, int nmotif, int setlength, int ninitg, int ninitd, int maxiterg, int maxiterd, float learnrt, float inslimit, float avglen, int verbose);
static void PrintHeader(int ncate, int setlength, int nmotif, int maxiterg, int maxiterd, int ninitg, int ninitd, float learnrt, int seed, int verbose, char *locdir, char *hmmg_dir, char *hmmd_dir, char **seqfiles);

int
main(int argc, char **argv)
{
  int   setlength       = DEFAULT_PHMM_LEN;
  float learnrt         = DEFAULT_LEARN_RATE;
  int   maxiterg        = MAXITER_G;
  int   maxiterd        = MAXITER_D;
  int   ninitg          = NINIT_G;
  int   ninitd          = NINIT_D;
  int   nmotif          = DEFAULT_NMOTIF;
  float inslimit        = DEFAULT_INSLIMIT;
  float avglen          = DEFAULT_AVGLEN;
  int   seed            = (int) time((time_t *) NULL); /* "random" seed */
  int   verbose         = 0;
  char *optname;                /* name of option found by Getopt()         */
  char *optarg;                 /* argument found by Getopt()               */
  int   optind;                 /* index in argv[]                          */

  char        *hmmg_dir         = "";
  char        *hmmd_dir         = "";
  char        *locdir           = "";
  char       **seqfiles         = NULL;
  char       **locfiles         = NULL;
  char      ***hmmg_files       = NULL;
  char      ***hmmd_files       = NULL;
  int          ncate;
  time_t       starttime;
  int          minutes;
  int          imotif;
  int          k;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-l") == 0) setlength     = atoi(optarg);
    else if (strcmp(optname, "-i") == 0) maxiterg      = atoi(optarg);
    else if (strcmp(optname, "-I") == 0) maxiterd      = atoi(optarg);
    else if (strcmp(optname, "-b") == 0) ninitg        = atoi(optarg);
    else if (strcmp(optname, "-B") == 0) ninitd        = atoi(optarg);
    else if (strcmp(optname, "-n") == 0) nmotif        = atoi(optarg);
    else if (strcmp(optname, "-r") == 0) learnrt       = atof(optarg);
    else if (strcmp(optname, "-m") == 0) inslimit      = atof(optarg);
    else if (strcmp(optname, "-a") == 0) avglen        = atof(optarg);
    else if (strcmp(optname, "-s") == 0) seed          = atoi(optarg);
    else if (strcmp(optname, "-o") == 0) locdir        = optarg;
    else if (strcmp(optname, "-g") == 0) hmmg_dir      = optarg;
    else if (strcmp(optname, "-d") == 0) hmmd_dir      = optarg;
    else if (strcmp(optname, "-v") == 0) verbose       = atoi(optarg);
    else if (strcmp(optname, "-h") == 0)
      {
        printf("sublearn: learn a HMM\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 1)
    Die("Incorrect number of arguments.\n%s\n", usage);

  if (nmotif > 100) Die("Number of motifs is too large: %d\n", nmotif);

  seqfiles = argv + optind;
  ncate = argc - optind;
  locfiles = (char **) MallocOrDie(sizeof(char *) * ncate);
  for (k = 0; k < ncate; k++) locfiles[k] = LocFilename(locdir, seqfiles[k]);
  HmmFilenames(hmmg_dir, hmmd_dir, ncate, seqfiles, nmotif, setlength,
               &hmmg_files, &hmmd_files);

  /* Main routines begins */
  starttime = time((time_t *) NULL);
  PrintHeader(ncate, setlength, nmotif, maxiterg, maxiterd, ninitg, ninitd,
              learnrt, seed, verbose, locdir, hmmg_dir, hmmd_dir, seqfiles);
  for (k = 0; k < ncate; k++)
    SaveLocHeader(locfiles[k], nmotif, setlength, maxiterg, maxiterd, ninitg,
                  ninitd, seed);
  RunSearchMultiMotifs(ncate, seqfiles, locfiles, hmmg_files, hmmd_files,
                       nmotif, setlength, ninitg, ninitd, maxiterg, maxiterd,
                       learnrt, inslimit, avglen, verbose);
  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  for (k = 0; k < ncate; k++) SaveLocBottom(locfiles[k], minutes);
  printf("Finished in %d:%d\n", minutes / 60, minutes % 60);
  /* Main routines ends */

  for (imotif = 0; imotif < nmotif; imotif++)
    {
      for (k = 0; k < ncate; k++)
        {
          free(hmmg_files[imotif][k]); free(hmmd_files[imotif][k]);
        }
      free(hmmg_files[imotif]); free(hmmd_files[imotif]);
    }
  free(hmmg_files); free(hmmd_files);

  for (k = 0; k < ncate; k++) free(locfiles[k]);
  free(locfiles);
  return 0;
}

/***********************************************
 * Filenames
 ***********************************************/

/* Return location filename based on sequence filename, by replacing extension
 * to ".submotif"; e.g. YeastPero.txt to YeastPero.submotif.
 * Caller need not free returned name.
 */
static char *
LocFilename(char *locdir, char *seqfile)
{
  char *locfile;
  int slash, dot;

  for (slash = strlen(seqfile)-1; slash >= 0 && seqfile[slash] != '/'; slash--) ;
  slash++;
  for (dot = strlen(seqfile)-1; dot >= slash && seqfile[dot] != '.'; dot--) ;
  if (dot < slash) dot = strlen(seqfile);

  locfile = (char *) MallocOrDie(strlen(locdir)+dot-slash+12);
  strcpy(locfile, locdir);
  strncat(locfile, seqfile + slash, dot - slash);
  strcat(locfile, ".submotif");
  return locfile;
}

/* Create generative and discriminative HMM filenames like
 * hmm/YeastER4_03.hmm and hmm/YeastER4_8_03.hmm
 */
static void
HmmFilenames(char *hmmg_dir, char *hmmd_dir, int ncate, char **seqfiles,
             int nmotif, int setlength, char ****ret_hmmg_files,
             char ****ret_hmmd_files)
{
  char ***hmmg_files, ***hmmd_files;
  int slash, dot, imotif, k;

  hmmg_files = (char ***) MallocOrDie(sizeof(char**) * nmotif);
  hmmd_files = (char ***) MallocOrDie(sizeof(char**) * nmotif);
  for (imotif = 0; imotif < nmotif; imotif++)
    {
      hmmg_files[imotif] = (char **) MallocOrDie(sizeof(char*) * ncate);
      hmmd_files[imotif] = (char **) MallocOrDie(sizeof(char*) * ncate);
    }
  for (k = 0; k < ncate; k++)
    {
      char *seqfile = seqfiles[k];
      for (slash = strlen(seqfile)-1; slash>=0 && seqfile[slash]!='/'; slash--);
      slash++;
      for (dot = strlen(seqfile)-1; dot>=slash && seqfile[dot]!='.'; dot--);
      if (dot < slash) dot = strlen(seqfile);

      for (imotif = 0; imotif < nmotif; imotif++)
        {
          hmmg_files[imotif][k] = (char*) MallocOrDie(strlen(hmmg_dir)
                                                      + dot - slash + 12);
          strcpy(hmmg_files[imotif][k], hmmg_dir);
          strncat(hmmg_files[imotif][k], seqfile + slash, dot - slash);
          sprintf(hmmg_files[imotif][k] + strlen(hmmg_files[imotif][k]),
                  "%d_%02d.hmm", setlength, imotif);

          hmmd_files[imotif][k] = (char*) MallocOrDie(strlen(hmmd_dir)
                                                      + dot - slash + 12);
          strcpy(hmmd_files[imotif][k], hmmd_dir);
          strncat(hmmd_files[imotif][k], seqfile + slash, dot - slash);
          sprintf(hmmd_files[imotif][k] + strlen(hmmd_files[imotif][k]),
                  "%d_%d_%02d.hmm", setlength, ncate - 1, imotif);
        }
    }
  *ret_hmmg_files = hmmg_files; *ret_hmmd_files = hmmd_files;
}

/***********************************************
 * Motif search
 ***********************************************/

/* Convert a multinomial PDF simplex to a CDF.
 * Note that caller need to free the allocated CDF.
 */
static float *PdfToCdf(int n, float *pdf)
{
  int x;
  float *cdf = (float *) MallocOrDie(sizeof(float) * n);
  assert(n >= 1);
  cdf[0] = pdf[0];
  for (x = 1; x < n; x++) cdf[x] = cdf[x - 1] + pdf[x];
  return cdf;
}

static unsigned char RandomAlphabet(float *cdf)
{
  int x;
  float u = sre_random();
  for (x = 0; x < Alphabet_size && u >= cdf[x]; x++) ;
  //assert(x < Alphabet_size);
  if (x >= Alphabet_size) {
    printf("WARNING: u=%f, x=%d, CDF = [", u, x);
    for (x = 0; x < Alphabet_size; x++) printf("%d ", (int)(1000.*cdf[x]));
    printf("]\n");
  }
  return x;
}

/* Perform generative training of HMM, and use it to scan for motifs.
 * Search for multiple motifs by masking found motif instances with
 * random background sequences.
 */
static void
SearchMultiMotifs(int ncate, int *catenseq, unsigned char ***catedsqs,
                  SQINFO **catesqinfo, char **locfiles, char ***hmmg_files,
                  char ***hmmd_files, int nmotif, int setlength, int ninitg,
                  int ninitd, int maxiterg, int maxiterd, float learnrt,
                  float inslimit, float avglen, int verbose)
{
  struct plan7_s      **hmmg            = NULL;
  struct plan7_s      **hmmd            = NULL;
  float                *null_cdf        = NULL;
  int                   imotif, k;

  hmmg = (struct plan7_s **) MallocOrDie(ncate*sizeof(struct plan7_s *));
  for (imotif = 0; imotif < nmotif; imotif++)
    {
      hmmd = DiscTrain(ncate, catenseq, catedsqs, catesqinfo, setlength, learnrt,
                       ninitg, ninitd, maxiterg, maxiterd, inslimit, avglen,
                       verbose, hmmg);
      fflush(stdout);
      if (maxiterg > 0 && NULL != hmmg_files) /* otherwise NULL == hmmg */
        SaveHMMs(ncate, hmmg, hmmg_files[imotif]);
      if (maxiterd > 0) /* otherwise NULL == hmmd */
        SaveHMMs(ncate, hmmd, hmmd_files[imotif]);
      //printf("HMM saved! Aligning...\n"); fflush(stdout);

      for (k = 0; k < ncate; k++)
        {
          int idx, tpos;
          struct plan7_s *hmm = NULL;
          int nseqs = catenseq[k];
          unsigned char **dsqs = catedsqs[k];
          SQINFO *sqinfo = catesqinfo[k];
          struct matlocs_s *matlocs = (struct matlocs_s *) MallocOrDie(
                                      sizeof(struct matlocs_s) * nseqs);
          struct p7trace_s **tr_arr = (struct p7trace_s **) MallocOrDie(
                                      sizeof(struct p7trace_s *) * nseqs);
          if (maxiterd > 0) hmm = hmmd[k]; else hmm = hmmg[k];

          /* Scan sequences by Viterbi algorithm, save HMM and locations */
          ScanMatches(nseqs, dsqs, sqinfo, hmm, matlocs, tr_arr);
          SaveLoc(locfiles[k], nseqs, dsqs, sqinfo, matlocs, imotif+1);

          /* Prepare background CDF, just once */
          if (NULL == null_cdf) null_cdf = PdfToCdf(Alphabet_size, hmm->null);

          /* Mask motifs with noise */
          for (idx = 0; idx < nseqs; idx++)
            {
              for (tpos = 0; tpos < tr_arr[idx]->tlen; tpos++)
                if (STM == tr_arr[idx]->statetype[tpos])
                  dsqs[idx][tr_arr[idx]->pos[tpos]] = RandomAlphabet(null_cdf);
              assert(matlocs->n == 1); /* OOPS model */
              FreeMatlocs(matlocs + idx);
            }
          free(matlocs); free(tr_arr);
        }
      //printf("Aligned and saved! Freeing...\n"); fflush(stdout);
      for (k = 0; k < ncate; k++) FreePlan7(hmmg[k]);
      if (maxiterd > 0)
        {
          if (hmmg[0] != hmmd[0]) {for (k=0; k<ncate; k++) FreePlan7(hmmd[k]);}
          free(hmmd);
        }
    }
  free(hmmg);
}

static void
RunSearchMultiMotifs(int ncate, char **seqfiles, char **locfiles,
                     char ***hmmg_files, char ***hmmd_files, int nmotif,
                     int setlength, int ninitg, int ninitd, int maxiterg,
                     int maxiterd, float learnrt, float inslimit, float avglen,
                     int verbose)
{
  int                  *catenseq   = NULL;
  unsigned char      ***catedsqs   = NULL;
  SQINFO              **catesqinfo = NULL;
  SetAlphabet(hmmAMINO);
  catedsqs = LoadCateDsqs(ncate, seqfiles, &catenseq, &catesqinfo);
  SearchMultiMotifs(ncate, catenseq, catedsqs, catesqinfo, locfiles, hmmg_files,
                    hmmd_files, nmotif, setlength, ninitg, ninitd, maxiterg,
                    maxiterd, learnrt, inslimit, avglen, verbose);
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
}

/***********************************************
 * Misc
 ***********************************************/

static void
PrintHeader(int ncate, int setlength, int nmotif, int maxiterg, int maxiterd,
            int ninitg, int ninitd, float learnrt, int seed, int verbose,
            char *locdir, char *hmmg_dir, char *hmmd_dir, char **seqfiles)
{
  int k;
  printf("submotif -l %d -n %d -i %d -I %d -b %d -B %d -r %.3g -s %d -v %d "
         "-o %s -g %s -d %s", setlength, nmotif, maxiterg, maxiterd,
         ninitg, ninitd, learnrt, seed, verbose, locdir, hmmg_dir, hmmd_dir);
  for (k = 0; k < ncate; k++) {printf("%s ", seqfiles[k]);}
  printf("\n\n");
}
