/* subrandom.c
 *
 * Generate a random profile HMM.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: subrandom [-options] <SEQFILE>\n"
"Options:\n"
"   -n <num-motif>    : number of motifs, using random masking (default 1)\n"
"   -l <length>       : profile HMM (match-states) length (default 10)\n"
"   -a <avg-length>   : average length for background (default 530.0)\n"
"   -g <gen-hmm-dir>  : output gen-HMM directory (default current directory)\n"
"   -s <seed>         : random seed\n"
"   -r                : random amino acids instead of from sequence\n";

static struct opt_s OPTIONS[] = {
  { "-n", TRUE, sqdARG_INT    },
  { "-l", TRUE, sqdARG_INT    },
  { "-a", TRUE, sqdARG_FLOAT  },
  { "-g", TRUE, sqdARG_STRING },
  { "-s", TRUE, sqdARG_INT    },
  { "-r", TRUE, sqdARG_NONE   },
  { "-h", TRUE, sqdARG_NONE   }
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_NMOTIF      = 1;
const int   DEFAULT_PHMM_LEN    = 10;
const float DEFAULT_AVGLEN      = 530.0;

static char **HmmgFilenames(char *hmmg_dir, char *seqfile, int nmotif, int setlength);
static void SaveRandomHMM(char *seqfile, char *hmmfile, int setlength, float avglen);

int
main(int argc, char **argv)
{
  int       nmotif          = DEFAULT_NMOTIF;
  int       imotif;
  int       setlength       = DEFAULT_PHMM_LEN;
  float     avglen          = DEFAULT_AVGLEN;
  char     *seqfile         = NULL;
  char     *hmmg_dir        = "";
  char    **hmmfiles        = NULL;
  int       seed            = -1; /* default, no seed */
  int       israndseq       = 1;
  char     *optname;
  char     *optarg;
  int       optind;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-l") == 0) setlength     = atoi(optarg);
    else if (strcmp(optname, "-n") == 0) nmotif        = atoi(optarg);
    else if (strcmp(optname, "-a") == 0) avglen        = atof(optarg);
    else if (strcmp(optname, "-g") == 0) hmmg_dir      = optarg;
    else if (strcmp(optname, "-s") == 0) seed          = atoi(optarg);
    else if (strcmp(optname, "-r") == 0) israndseq     = 0;
    else if (strcmp(optname, "-h") == 0)
      {
        printf("subrandom: generate a random HMM\n\n%s\n", usage);
        exit(0);
      }
  }

  if (argc - optind < 1)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];

  if (seed >= 0) sre_srandom(seed);
  hmmfiles = HmmgFilenames(hmmg_dir, seqfile, nmotif, setlength);
  if (!israndseq) seqfile = NULL; /* indicates random amino acids */
  for (imotif = 0; imotif < nmotif; imotif++)
    {
      SaveRandomHMM(seqfile, hmmfiles[imotif], setlength, avglen);
    }
  for (imotif = 0; imotif < nmotif; imotif++) free(hmmfiles[imotif]);
  free(hmmfiles);
  return 0;
}

/* Create generative HMM filenames like hmm/YeastER4_03.hmm
 */
static char **
HmmgFilenames(char *hmmg_dir, char *seqfile, int nmotif, int setlength)
{
  char **hmmg_files;
  int slash, dot, imotif;

  hmmg_files = (char **) MallocOrDie(sizeof(char*) * nmotif);
  for (slash = strlen(seqfile)-1; slash>=0 && seqfile[slash]!='/'; slash--);
  slash++;
  for (dot = strlen(seqfile)-1; dot>=slash && seqfile[dot]!='.'; dot--);
  if (dot < slash) dot = strlen(seqfile);

  for (imotif = 0; imotif < nmotif; imotif++)
    {
      hmmg_files[imotif] = (char*) MallocOrDie(strlen(hmmg_dir)+dot-slash+12);
      strcpy(hmmg_files[imotif], hmmg_dir);
      strncat(hmmg_files[imotif], seqfile + slash, dot - slash);
      sprintf(hmmg_files[imotif] + strlen(hmmg_files[imotif]),
              "%d_%02d.hmm", setlength, imotif);
    }
  return hmmg_files;
}

static void
SaveRandomHMM(char *seqfile, char *hmmfile, int setlength, float avglen)
{
  int nseqs, idx;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = NULL;
  struct plan7_s *hmm = NULL;
  struct p7prior_s *prior = NULL;
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */

  SetAlphabet(hmmAMINO);
  prior = P7DefaultPrior();
  P7DefaultNullModel(randomseq, &p1);
  p1 = avglen / (avglen + 1.0);
  if (NULL != seqfile)
    {
      /* Load sequences and pick one randomly */
      dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo);
      idx = (int) (sre_random() * nseqs);
      assert(0 <= idx && idx < nseqs);
      /* Generate random HMM */
      hmm = P7Init(setlength, prior, randomseq, p1, 1, dsqs+idx, sqinfo+idx);
      /* Free sequences */
      for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
      free(dsqs); free(sqinfo);
    }
  else
    {
      hmm = P7Init(setlength, prior, randomseq, p1, -1, NULL, NULL);
    }

  /* Print and save random HMM */
  PrintHMM(hmm);
  SaveHMM(hmm, hmmfile);

  FreePlan7(hmm);
  P7FreePrior(prior);
}
