/* subruncv.c
 *
 * Cross-validation functions.
 * Wrapper-main functions for easier calling by the main program and unit test.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"             /* function declarations for SUBLOC     */

/***********************************************
 * Cross validation
 ***********************************************/

/* Purpose:  Split pre-permuted data into training and testing set according to
 *           cross-validation rules. Called in each fold.
 *           IMPORTANT: training and testing set holders must be allocated first.
 *
 * Args:     nfold          - number of folds
 *           fold           - which fold to test now
 *           ncate          - number of categories
 *           catenseq       - categorized number of sequences
 *           catedsqs       - categorized sequences
 *           catesqinfo     - categorized SQINFOs
 *           trainnseq      - categorized number of sequences of training set
 *                            MUST allocated first, 0 .. ncate-1
 *           traindsqs      - categorized sequences of training set
 *                            MUST allocated first, 0 .. ncate-1
 *           trainsqinfo    - categorized SQINFOs of training set
 *                            MUST allocated first, 0 .. ncate-1
 *           testnseq       - categorized number of sequences of testing set
 *                            MUST allocated first, 0 .. ncate-1
 *           testdsqs       - categorized sequences of testing set
 *                            MUST allocated first, 0 .. ncate-1
 *           testsqinfo     - categorized SQINFOs of testing set
 *                            MUST allocated first, 0 .. ncate-1
 */
static void
CvCopyPointers(int nfold, int fold, int ncate,
               int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo,
               int *trainnseq, unsigned char ***traindsqs, SQINFO **trainsqinfo,
               int *testnseq, unsigned char ***testdsqs, SQINFO **testsqinfo)
{
  int k;
  assert(NULL != trainnseq && NULL != traindsqs && NULL != trainsqinfo);
  assert(NULL != testnseq  && NULL != testdsqs  && NULL != testsqinfo );
  assert(nfold != 0);
  for (k = 0; k < ncate; k++)
    {
      int n = catenseq[k];
      int a = n * fold / nfold; //integer division intended
      int b = n * (fold + 1) / nfold; //integer division intended
      unsigned char **s = catedsqs[k];
      unsigned char **tests = (unsigned char **) MallocOrDie(
                              (b - a) * sizeof(unsigned char *));
      unsigned char **trains = (unsigned char **) MallocOrDie(
                               (n - b + a) * sizeof(unsigned char *));
      SQINFO *q = catesqinfo[k];
      SQINFO *testq = (SQINFO*) MallocOrDie((b - a) * sizeof(SQINFO));
      SQINFO *trainq = (SQINFO*) MallocOrDie((n - b + a) * sizeof(SQINFO));
      testnseq[k] = b - a; trainnseq[k] = n - b + a;
      traindsqs[k] = trains; trainsqinfo[k] = trainq;
      testdsqs[k] = tests; testsqinfo[k] = testq;

      memcpy(tests, s + a, (b - a) * sizeof(unsigned char *));
      memcpy(testq, q + a, (b - a) * sizeof(SQINFO));
      if (fold > 0)
        {
          memcpy(trains, s, a * sizeof(unsigned char *));
          memcpy(trainq, q, a * sizeof(SQINFO));
        }
      if (fold < nfold - 1)
        {
          memcpy(trains + a, s + b, (n - b) * sizeof(unsigned char *));
          memcpy(trainq + a, q + b, (n - b) * sizeof(SQINFO));
        }
    }
}

/* Purpose:  Split pre-permuted data into training and testing set according to
 *           cross-validation rules. Called in each fold.
 *           IMPORTANT: training and testing set holders must be allocated first.
 *
 * Args:     ncate          - number of categories
 *           traindsqs      - categorized sequences of training set
 *           trainsqinfo    - categorized SQINFOs of training set
 *           testdsqs       - categorized sequences of testing set
 *           testsqinfo     - categorized SQINFOs of testing set
 */
static void CvFreeCopy(int ncate, unsigned char ***traindsqs, SQINFO **trainsqinfo,
                unsigned char ***testdsqs, SQINFO **testsqinfo)
{
  int k;
  for (k = 0; k < ncate; k++)
    {
      free(traindsqs[k]); free(trainsqinfo[k]);
      free(testdsqs[k] ); free(testsqinfo[k] );
    }
}

/* Purpose:  Cross validate generative and discriminative training.
 *
 * Args:     ncate      - number of categories
 *           catenseq   - number of sequences of each category, 0..ncate
 *           catedsqs   - array of sequences of each category, 0..ncate
 *           catesqinfo - array of sequence information of each category, 0..ncate
 *           nfold      - number of fold
 *           setlength  - profile HMM length
 *           learnrt    - learning rate
 *           ninit      - number of random initialization
 *           maxiterg   - maximum iterations of generative training
 *           maxiterd   - maximum iterations of discriminative training
 *           verbose    - verbose output
 *           accgen     - pre-allocated array of accuracy of each fold (gen training)
 *           accdisc    - pre-allocated array of accuracy of each fold (disc training)
 */
void
CrossValidate(int ncate, int *catenseq, unsigned char ***catedsqs, SQINFO **catesqinfo,
              int nfold, int setlength, float learnrt, int ninit, int maxiterg,
              int maxiterd, float inslimit, float avglen, int verbose, float *accgen,
              float *accdisc)
{
  int fold, k;
  struct plan7_s **hmmg = (struct plan7_s **) MallocOrDie(
                          ncate * sizeof(struct plan7_s *));
  //struct plan7_s **hmmd;

  int *trainnseq = (int *) MallocOrDie(ncate * sizeof(int));
  int *testnseq  = (int *) MallocOrDie(ncate * sizeof(int));
  unsigned char ***traindsqs = (unsigned char ***) MallocOrDie(
                               ncate * sizeof(unsigned char **));
  unsigned char ***testdsqs  = (unsigned char ***) MallocOrDie(
                               ncate * sizeof(unsigned char **));
  SQINFO **trainsqinfo = (SQINFO **) MallocOrDie(ncate * sizeof(SQINFO *));
  SQINFO **testsqinfo  = (SQINFO **) MallocOrDie(ncate * sizeof(SQINFO *));

  for (fold = 0; fold < nfold; fold++)
    {
      /*printf("fold=%d\n", fold);*/
      CvCopyPointers(nfold, fold, ncate, catenseq, catedsqs, catesqinfo,
          trainnseq, traindsqs, trainsqinfo, testnseq, testdsqs, testsqinfo);
      for (k = 0; k < ncate; k++)
        {
          float score;
          hmmg[k] = GenTrain(catenseq[k], catedsqs[k], catesqinfo[k], setlength,
                             maxiterg, ninit, inslimit, avglen, verbose, &score);
        }
      /*hmmd = DiscTrain(ncate, trainnseq, trainseqs, trainsqinfo, setlength,
                       learnrt, ninit, maxiterg, maxiterd, verbose, hmmg);*/

      if (maxiterg > 0)
        {
          accgen[fold] = EvaluateSeqs(ncate, hmmg, testnseq, testdsqs,
                                      testsqinfo);
          FreeHMMs(ncate, hmmg);
        }
      else accgen[fold] = 0.0;

      /*if (maxiterd > 0)
        {
          accdisc[fold] = EvaluateSeqs(ncate, hmmd, testnseq, testseqs);
          FreeHMMs(ncate, hmmd); free(hmmd);
        }
      else accdisc[fold] = 0.0;*/

      CvFreeCopy(ncate, traindsqs, trainsqinfo, testdsqs, testsqinfo);
    }
  /*printf("%d-fold CV, largest category ratio %.2f\n", nfold,
         (float)maxi(ncate,catenseq) / (float)sumi(ncate,catenseq));
  printf("Generative accuracy = %.2f (", 100. * avgf(nfold, accgen));
  for (fold = 0; fold < nfold; fold++) printf(" %.2f ", 100. * accgen[fold]);
  printf(")\n");
  if (maxiterd > 0)
    {
      printf("Discriminative accuracy = %.2f (", 100. * avgf(nfold, accdisc));
      for (fold = 0; fold < nfold; fold++) printf(" %.2f ", 100. * accdisc[fold]);
      printf(")\n");
    }*/

  free(hmmg);
  free(trainnseq); free(traindsqs); free(trainsqinfo);
  free(testnseq); free(testdsqs); free(testsqinfo);
}

/***********************************************
 * Wrappers / runners: RunCrossValidate
 ***********************************************/

/* Permute sequences randomly. Called by RunCrossValidate() only. */
static void
RandomPermute(int ncate, int *catenseq, unsigned char ***catedsqs,
              SQINFO **catesqinfo)
{
  int k, i;
  for (k = 0; k < ncate; k++)
    {
      int n = catenseq[k];
      unsigned char **s = catedsqs[k];
      SQINFO *q = catesqinfo[k];
      for (i = catenseq[k] - 1; i > 0; i--)
        {
          int j = (int) (sre_random() * (i + 1));
          if (j < i)
            {
              unsigned char *s1 = s[i]; s[i] = s[j]; s[j] = s1;
              SQINFO         q1 = q[i]; q[i] = q[j]; q[j] = q1;
            }
        }
    }
}

/* Write the parameter setting. Called by RunCrossValidate() only. */
static void
WriteSettings(FILE *fp, char *title, int nfold, int nrepeat, int setlength,
              float learnrt, int ninit, int maxiterg, int maxiterd, int seed)
{
  int k;
  if (NULL == fp) return;
  /*fprintf(fp, "Classes=%s", labels[0]);
  for (k = 1; k < ncate; k++) fprintf(fp, "/%s", labels[k]);*/
  fprintf(fp, "%s, Length=%d, LearnRate=%.3f, MaxIterG=%d, MaxIterD=%d, #Fold=%d, "
          "#Repeat=%d, #RandInit=%d, Seed=%d, SeqFile=%s\n",
          title, setlength, learnrt, maxiterg, maxiterd, nfold, nrepeat, ninit, seed);
}

/* Write cross validation results. Called by RunCrossValidate() only. */
static void
WriteCvResult(FILE *fp, int nfold, float *accgen, float *accdisc)
{
  int fold;
  if (NULL == fp) return;
  fprintf(fp, "    Gen=%.2f Disc=%.2f ( ", 100. * avgf(nfold, accgen),
                                           100. * avgf(nfold, accdisc));
  for (fold = 0; fold < nfold; fold++)
    fprintf(fp, "%.2f/%.2f ", accgen[fold], accdisc[fold]);
  fprintf(fp, ")\n");
}

/* Write cross validation results, averaged over repeats with random splits.
 * Called by RunCrossValidate() only.
 */
static void
WriteRepeatAvg(FILE *fp, int nrepeat, float accgen, float accdisc)
{
  if (NULL == fp) return;
  fprintf(fp, "Average in %d permutations: Gen=%.2f Disc=%.2f\n", nrepeat,
          100. * accgen, 100. * accdisc);
}

/* A wrapper for cross validation.
 * Called by the main program SUBCV and also in unit test.
 */
void
RunCrossValidate(int ncate, int *catenseq, unsigned char ***catedsqs,
    SQINFO **catesqinfo, int nfold, int nrepeat, int setlength, float learnrt,
    int ninit, int maxiterg, int maxiterd, float inslimit, float avglen,
    int seed, int verbose, char *title, char *resultfile)
{
  int ispermute = 1, irepeat;
  float *accgen = (float *) MallocOrDie(nfold * sizeof(float));
  float *accdisc = (float *) MallocOrDie(nfold * sizeof(float));
  float sumaccgen = 0.0, sumaccdisc = 0.0;
  FILE *resultfp = NULL;

  if (seed > 0) sre_srandom(seed); else nrepeat = 0;/*no permutation*/

  if (resultfile != NULL) resultfp = fopen(resultfile, "a");
  WriteSettings(resultfp, title, nfold, nrepeat, setlength, learnrt, ninit,
                maxiterg, maxiterd, seed);
  WriteSettings(stdout  , title, nfold, nrepeat, setlength, learnrt, ninit,
                maxiterg, maxiterd, seed);

  if (nrepeat <= 0) {nrepeat = 1; ispermute = 0;}
  for (irepeat = 0; irepeat < nrepeat; irepeat++)
    {
      if (ispermute) RandomPermute(ncate, catenseq, catedsqs, catesqinfo);
      CrossValidate(ncate, catenseq, catedsqs, catesqinfo, nfold, setlength,
                    learnrt, ninit, maxiterg, maxiterd, inslimit, avglen,
                    verbose, accgen, accdisc);
      WriteCvResult(resultfp, nfold, accgen, accdisc);
      WriteCvResult(stdout  , nfold, accgen, accdisc);
      sumaccgen += avgf(nfold, accgen); sumaccdisc += avgf(nfold, accdisc);
    }
  WriteRepeatAvg(resultfp, nrepeat, sumaccgen / nrepeat, sumaccdisc / nrepeat);
  WriteRepeatAvg(stdout  , nrepeat, sumaccgen / nrepeat, sumaccdisc / nrepeat);

  if (resultfp != NULL) fclose(resultfp);
  free(accgen); free(accdisc);
}

/***********************************************
 * Wrappers / runners
 ***********************************************/

/* Perform generative training of profile HMM on the given sequences,
 * and save to a HMM file. A wrapper for only used in unit testing so far.
 */
void
RunGenTrain(char *seqfile, char *hmmfile, int setlength, int maxiter, int ninit,
            float inslimit, float avglen, int verbose)
{
  int             idx;
  int             nseqs         = 0;
  unsigned char **dsqs          = NULL;
  SQINFO         *sqinfo        = NULL;
  struct plan7_s *hmm           = NULL;
  float           score         = 0.0;

  SetAlphabet(hmmAMINO);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo); /* load sequence and digitize */
  hmm = GenTrain(nseqs, dsqs, sqinfo, setlength, maxiter, ninit, inslimit,
                 avglen, verbose, &score); /* generative training */
  if (!verbose) { printf("Best profile HMM:\n"); PrintHMM(hmm); }
  (void) SaveHMM(hmm, hmmfile);

  (void) FreePlan7(hmm);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}
