/* subscan.c
 *
 * Scan sequences for motifs represented by profile HMMs using the Viterbi algorithm.
 * Similar to hmmalign.c in HMMER 2.3.2.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "squid.h"              /* general sequence analysis library    */
#include "ssi.h"                // subio
#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "subloc.h"

static char usage[]  =
"Usage: subscan [-options] <seqfile> <locfile> <hmmfile1> <hmmfile2>...\n"
"Options:\n"
"   -p                : posterior decoding\n"
"   -v <level>        : verbose level (default 0, least information)\n";

static struct opt_s OPTIONS[] = {
  { "-p", TRUE, sqdARG_NONE   },
  { "-v", TRUE, sqdARG_INT    }
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

static void RunViterbiSave(  char *seqfile, char *locfile , int nhmm, char **hmmfiles, int verbose);
static void RunPosteriorSave(char *seqfile, char *postfile, int nhmm, char **hmmfiles, int verbose);

int
main(int argc, char **argv)
{
  char *seqfile         = NULL;
  char *locfile         = NULL;
  char**hmmfiles        = NULL;
  int   nhmm            = 0;
  int   isposterior     = 0;
  int   verbose         = 0;
  char *optname;                /* name of option found by Getopt()         */
  char *optarg;                 /* argument found by Getopt()               */
  int   optind;                 /* index in argv[]                          */

  time_t       starttime;
  int          minutes;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-v") == 0) verbose       = atoi(optarg);
    else if (strcmp(optname, "-p") == 0) isposterior   = 1;
    else if (strcmp(optname, "-h") == 0)
      {
        printf("subscan: scan sequences given a HMM\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 2)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];
  locfile = argv[optind++];
  hmmfiles = argv + optind;
  nhmm = argc - optind;

  if (isposterior)
    {
      starttime = time((time_t *) NULL);
      RunPosteriorSave(seqfile, locfile, nhmm, hmmfiles, verbose);
      minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
      printf("Finished in %d:%d\n", minutes / 60, minutes % 60);
    }
  else
    {
      starttime = time((time_t *) NULL);
      SaveLocHeader(locfile, nhmm, 0, 0, 0, 0, 0, 0);
      RunViterbiSave(seqfile, locfile, nhmm, hmmfiles, verbose);
      minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
      SaveLocBottom(locfile, minutes);
      printf("Finished in %d:%d\n", minutes / 60, minutes % 60);
    }
  return 0;
}

static void
RunViterbiSave(char *seqfile, char *locfile, int nhmm, char **hmmfiles, int verbose)
{
  int nseqs = 0, idx, ihmm;
  struct plan7_s *hmm;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = NULL;
  struct matlocs_s *matlocs;

  SetAlphabet(hmmAMINO);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo);

  for (ihmm = 0; ihmm < nhmm; ihmm++)
    {
      if (verbose) printf("Scanning motif %d in %s...\n",ihmm+1,hmmfiles[ihmm]);
      hmm = LoadHMM(hmmfiles[ihmm]);
      /* Perform Viterbi algorithm and save locations */
      P7Logoddsify(hmm, TRUE); /* TRUE for viterbi mode */
      matlocs = (struct matlocs_s *) MallocOrDie(sizeof(struct matlocs_s) * nseqs);
      ScanMatches(nseqs, dsqs, sqinfo, hmm, matlocs, NULL);
      SaveLoc(locfile, nseqs, dsqs, sqinfo, matlocs, ihmm+1);

      for (idx = 0; idx < nseqs; idx++) (void) FreeMatlocs(matlocs + idx);
      free(matlocs);
      FreePlan7(hmm);
    }

  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}

/***********************************************
 * Scanning by posterior decoding
 ***********************************************/

/* The caller will save top motif instances as a list of tuples
 * (SEQ_IDX, SEQ_ID, BEGIN, END).
 */
/* static void
TopListAdd(int maxtop, int ntop, struct matloc_s *topinst, ?)
{
  int i, j;
  for (i = 1; i <= L; i++)
    {
      score1 = mx->xmx[i][XMB];
      if (ntop < maxtop) ntop++;
      else if (topinst[ntop-1].score >= score1) continue;
      for (j = ntop-2; j >= 0 && topinst[j] < score1; j--)
        topinst[j+1] = topinst[j];
      topinst[j+1].score = score1;
      topinst[j+1].seqidx = ?
      topinst[j+1].seqname = ?
      topinst[j+1].begin = ?
      topinst[j+1].end = ?
    }
}*/

static void
SavePosterior(FILE *fp, int idx, char *name, int L, float *begin, float *end)
{
  int i;
  fprintf(fp, "%d %s ", idx, name);
  for (i = 1; i <= L; i++)
    {
      assert(0. <= begin[i] && begin[i] <= 1.);
      fprintf(fp, "%.0f", floor(10. * begin[i]));
    }
  fprintf(fp, " ");
  for (i = 1; i <= L; i++)
    {
      assert(0. <= end[i] && end[i] <= 1.);
      fprintf(fp, "%.0f", floor(10. * end[i]));
    }
  fprintf(fp, "\n");
}

/* Scan sequences using posterior decoding.
 * Use the silent state B (beginning of motif).
 */
static void
ScanPosterior(int nseqs, unsigned char **dsqs, SQINFO *sqinfo,
              struct plan7_s *hmm, float **begin, float **end, FILE *fp)
{
  int idx;
  struct dpmatrix_s *mx;
  struct matloc_s *topinst;     /* top motif instances */
  int ntop = 0;

  topinst = (struct matloc_s *) MallocOrDie(sizeof(struct matloc_s) * ntop);
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0);
  for (idx = 0; idx < nseqs; idx++)
    {
      PosteriorBE(dsqs[idx], sqinfo[idx].len, hmm, begin[idx], end[idx]);
      SavePosterior(fp, idx, sqinfo[idx].name, sqinfo[idx].len, begin[idx], end[idx]);
    }
  FreePlan7Matrix(mx);
}

static void
RunPosteriorSave(char *seqfile, char *postfile, int nhmm, char **hmmfiles, int verbose)
{
  int nseqs = 0, idx, ihmm;
  struct plan7_s *hmm;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = NULL;
  float **begin;
  float **end;
  FILE *fp;

  SetAlphabet(hmmAMINO);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo);
  begin = (float **) MallocOrDie(sizeof(float*) * nseqs);
  end   = (float **) MallocOrDie(sizeof(float*) * nseqs);
  for (idx = 0; idx < nseqs; idx++)
    {
      begin[idx] = (float *) MallocOrDie(sizeof(float) * (1+sqinfo[idx].len));
      end[idx]   = (float *) MallocOrDie(sizeof(float) * (1+sqinfo[idx].len));
    }

  fp = fopen(postfile, "w");
  for (ihmm = 0; ihmm < nhmm; ihmm++)
    {
      if (verbose) printf("Scanning motif %d in %s...\n",ihmm+1,hmmfiles[ihmm]);
      hmm = LoadHMM(hmmfiles[ihmm]);
      fprintf(fp, "Motif %d in %s\n", ihmm+1, hmmfiles[ihmm]);
      ScanPosterior(nseqs, dsqs, sqinfo, hmm, begin, end, fp);
      FreePlan7(hmm);
    }
  fclose(fp);

  for (idx = 0; idx < nseqs; idx++) {free(begin[idx]); free(end[idx]);}
  free(begin); free(end);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}
