/* subtrain.c
 *
 * Generative training an HMM from unaligned sequences
 * (not using negative examples).
 *
 * This module is based on HMMER 1.8.5 source file train_main.c.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"             /* function declarations for SUBLOC     */

/* Return index of maximum in a float array. */
int
argmaxf(int n, float *arr)
{
  int i, argmax = 0;
  if (n <= 0) (void)Die("Invalid array length (%d) in argmaxf()", n);
  for (i = 1; i < n; i++) if (arr[i] - arr[argmax] > FLT_EPSILON) argmax = i;
  return argmax;
}

/* Return index of maximum in a double array. */
int
argmaxd(int n, double *arr)
{
  int i, argmax = 0;
  if (n <= 0) (void)Die("Invalid array length (%d) in argmaxf()", n);
  for (i = 1; i < n; i++) if (arr[i] - arr[argmax] > DBL_EPSILON) argmax = i;
  return argmax;
}

/* Return index of maximum in an int array. */
int
argmaxi(int n, int *arr)
{
  int i, argmax = 0;
  if (n <= 0) (void)Die("Invalid array length (%d) in argmaxi()", n);
  for (i = 1; i < n; i++) if (arr[i] > arr[argmax]) argmax = i;
  return argmax;
}

/* Return the maximum in an int array. */
int
maxi(int n, int *arr)
{
  int optimum;
  int i;
  if (n <= 0) (void)Die("Invalid array length (%d) in maxi()", n);
  optimum = arr[0];
  for (i = 1; i < n; i++) if (arr[i] > optimum) optimum = arr[i];
  return optimum;
}

/* Return the minimum in a float array. */
float
maxf(int n, float *arr)
{
  float optimum;
  int i;
  if (n <= 0) (void)Die("Invalid array length (%d) in minf()", n);
  optimum = arr[0];
  for (i = 1; i < n; i++) if (arr[i] - optimum > FLT_EPSILON) optimum = arr[i];
  return optimum;
}

/* Return the minimum in a float array. */
float
minf(int n, float *arr)
{
  float optimum;
  int i;
  if (n <= 0) (void)Die("Invalid array length (%d) in minf()", n);
  optimum = arr[0];
  for (i = 1; i < n; i++) if (optimum - arr[i] > FLT_EPSILON) optimum = arr[i];
  return optimum;
}

/* Return the summation of an int array. */
int
sumi(int n, int *arr)
{
  int result = 0, i;
  for (i = 0; i < n; i++) result += arr[i];
  return result;
}

/* Return the summation of an int array. */
float
sumf(int n, float *arr)
{
  float result = 0;
  int i;
  for (i = 0; i < n; i++) result += arr[i];
  return result;
}

float avgi(int n, int   *arr) {return (float) sumi(n, arr) / (float) n;}
float avgf(int n, float *arr) {return sumf(n, arr) / n;}

/* Add the given float number to all members of an array. */
void
floatsadd(int n, float *f, float x)
{
  int i;
  for (i = 0; i < n; i++) f[i] += x;
  for (i = 0; i < n; i++) assert(f[i] >= -FLT_EPSILON);
}

/* Disable null model or log-odds for Baum-Welch algorithm. */
void
P7UniformNullModel(struct plan7_s *hmm)
{
  int x;
  for (x = 0; x < Alphabet_size; x++) hmm->null[x] = 1.;//test!!!
  hmm->p1 = 1.;
}

/* A simple prior mainly for testing purpose; based on P7LaplacePrior() code */
struct p7prior_s *
P7SimplePrior(void)
{
  struct p7prior_s *pri;

  pri = P7AllocPrior();
  pri->strategy = PRI_DCHLET;

  pri->tnum     = 1;
  pri->tq[0]    = 1.;
  pri->t[0][TMM] = .75; pri->t[0][TMI] = .125; pri->t[0][TMD] = .125;
  pri->t[0][TIM] = .5 ; pri->t[0][TII] = .5 ;
  pri->t[0][TDM] = .5 ; pri->t[0][TDD] = .5 ;

  pri->mnum  = 1;
  pri->mq[0] = 1.;
  FSet(pri->m[0], Alphabet_size, .05);

  pri->inum  = 1;
  pri->iq[0] = 1.;
  FSet(pri->i[0], Alphabet_size, .05);

  return pri;
}

void
P7RetractWing(struct plan7_s *hmm)
{
  hmm->tbd1 = 0.;
  float tmm_tmi = hmm->t[hmm->M-1][TMM] + hmm->t[hmm->M-1][TMI];
  hmm->t[hmm->M-1][TMM] /= tmm_tmi; hmm->t[hmm->M-1][TMI] /= tmm_tmi;
  hmm->t[hmm->M-1][TDD] = hmm->t[hmm->M-1][TMD] = 0.; hmm->t[hmm->M-1][TDM] = 1.;
  hmm->t[1][TDD] = hmm->t[1][TDM] = 0.; /* For backward algorithm */
}

/* Purpose:  Create a dynamic programming matrix for standard Forward,
 *           Backward. Keeps 2D arrays compact in RAM in an attempt
 *           to maximize cache hits. This code is based on HMMER 2.3.2.
 *
 * Args:     N     - N+1 rows are allocated, for sequence.
 *           M     - size of model in nodes
 *           padN  - over-realloc in seq/row dimension, or 0
 *           padM  - over-realloc in HMM/column dimension, or 0
 *
 * Return:   mx
 *           mx is allocated here. Caller frees with FreeMleDpMat(mx).
 */
struct mledpmat_s *
CreateMleDpMat(int N, int M, int padN, int padM)
{
  struct mledpmat_s *mx;
  int i;

  mx          = (struct mledpmat_s *) MallocOrDie (sizeof(struct mledpmat_s));
  mx->scl     = NULL;

  mx->xmx     = (float **) MallocOrDie (sizeof(float *) * (N+1));
  mx->mmx     = (float **) MallocOrDie (sizeof(float *) * (N+1));
  mx->imx     = (float **) MallocOrDie (sizeof(float *) * (N+1));
  mx->dmx     = (float **) MallocOrDie (sizeof(float *) * (N+1));
  mx->xmx_mem = (void *) MallocOrDie (sizeof(float) * ((N+1)*5));
  mx->mmx_mem = (void *) MallocOrDie (sizeof(float) * ((N+1)*(M+2)));
  mx->imx_mem = (void *) MallocOrDie (sizeof(float) * ((N+1)*(M+2)));
  mx->dmx_mem = (void *) MallocOrDie (sizeof(float) * ((N+1)*(M+2)));

  /* The indirect assignment below looks wasteful; it's actually
   * used for aligning data on 16-byte boundaries as a cache
   * optimization in the fast altivec implementation
   */
  mx->xmx[0] = (float *) mx->xmx_mem;
  mx->mmx[0] = (float *) mx->mmx_mem;
  mx->imx[0] = (float *) mx->imx_mem;
  mx->dmx[0] = (float *) mx->dmx_mem;
  for (i = 1; i <= N; i++)
    {
      mx->xmx[i] = mx->xmx[0] + (i*5);
      mx->mmx[i] = mx->mmx[0] + (i*(M+2));
      mx->imx[i] = mx->imx[0] + (i*(M+2));
      mx->dmx[i] = mx->dmx[0] + (i*(M+2));
    }

  mx->maxN = N;
  mx->maxM = M;
  mx->padN = padN;
  mx->padM = padM;

  return mx;
}

/* Purpose:  Free a dynamic programming matrix allocated by CreateMleDpMat().
 *
 * Return:   (void)
 */
void
FreeMleDpMat(struct mledpmat_s *mx)
{
  if (NULL != mx->xmx_mem) free(mx->xmx_mem);
  if (NULL != mx->mmx_mem) free(mx->mmx_mem);
  if (NULL != mx->imx_mem) free(mx->imx_mem);
  if (NULL != mx->dmx_mem) free(mx->dmx_mem);
  if (NULL != mx->xmx) free(mx->xmx);
  if (NULL != mx->mmx) free(mx->mmx);
  if (NULL != mx->imx) free(mx->imx);
  if (NULL != mx->dmx) free(mx->dmx);
  if (NULL != mx->scl) free(mx->scl);
  free(mx);
}

/* Purpose:  The Forward dynamic programming algorithm.
 *           Based on HMMER 2.3.2 code.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_mx - RETURN: dp matrix; pass NULL if it's not wanted
 *
 * Return:   Pr(Seq|Model).
 */
float
MleForward(unsigned char *dsq, int L, struct plan7_s *hmm, struct mledpmat_s **ret_mx)
{
  struct mledpmat_s *mx;
  float *scl;
  float **xmx, **mmx, **imx, **dmx;
  float sclprob;
  int   i, k, ts;

  /* t[0][0..6] Not set in ZeroPlan7 or any other places. */
  for (ts = 0; ts < 7; ts++) hmm->t[0][ts] = 0.;

  /* Allocate a DP matrix with 0..L rows, 0..M-1 columns. */
  mx = CreateMleDpMat(L+1, hmm->M, 0, 0);
  mx->scl = (float *) MallocOrDie (sizeof(float) * (L+2));
  scl = mx->scl; xmx = mx->xmx; mmx = mx->mmx; imx = mx->imx; dmx = mx->dmx;

  /* Initialization of the zero row.
   * Note that xmx[i][stN] = 0 by definition for all i,
   *    and xmx[i][stT] = xmx[i][stC], so neither stN nor stT need
   *    to be calculated in DP matrices.
   */
  xmx[0][XMN] = 1.;                              /* S->N, p=1            */
  xmx[0][XMB] = hmm->xt[XTN][MOVE];              /* S->N->B, no N-tail   */
  xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = 0.;  /* need seq to get here */
  for (k = 0; k <= hmm->M; k++)
    mmx[0][k] = imx[0][k] = dmx[0][k] = 0.;      /* need seq to get here */
  scl[0] = 1.;

  /* Recursion. Done as a pull.
   * Note some slightly wasteful boundary conditions:
   *    t[0] = 0 for all eight transitions (no node 0)
   */
  for (i = 1; i <= L; i++)
    {
      mmx[i][0] = imx[i][0] = dmx[i][0] = 0.;
      for (k = 1; k <= hmm->M; k++)
        {
          mmx[i][k]  = hmm->mat[k][dsq[i]] * ( mmx[i-1][k-1] * hmm->t[k-1][TMM]
                                             + imx[i-1][k-1] * hmm->t[k-1][TIM]
                                             + dmx[i-1][k-1] * hmm->t[k-1][TDM]
                                             + xmx[i-1][XMB] * hmm->begin[k]  );
          if (k == hmm->M) break;
          dmx[i][k]  = mmx[i][k-1] * hmm->t[k-1][TMD]
                     + dmx[i][k-1] * hmm->t[k-1][TDD];
          imx[i][k]  = hmm->ins[k][dsq[i]] * ( mmx[i-1][k] * hmm->t[k][TMI]
                                             + imx[i-1][k] * hmm->t[k][TII]);
        }
      /* Now the special states.
       * remember, C and J emissions are zero score by definition
       */
      xmx[i][XMN] = hmm->null[dsq[i]] *   xmx[i-1][XMN] * hmm->xt[XTN][LOOP];
      xmx[i][XME] = 0.;
      for (k = 1; k <= hmm->M; k++) xmx[i][XME] += mmx[i][k] * hmm->end[k];
      xmx[i][XMJ] = hmm->null[dsq[i]] * ( xmx[i-1][XMJ] * hmm->xt[XTJ][LOOP]
                                        + xmx[i-1][XME] * hmm->xt[XTE][LOOP]);
      xmx[i][XMB] =                       xmx[i  ][XMN] * hmm->xt[XTN][MOVE]
                                        + xmx[i  ][XMJ] * hmm->xt[XTJ][MOVE];
      xmx[i][XMC] = hmm->null[dsq[i]] * ( xmx[i-1][XMC] * hmm->xt[XTC][LOOP]
                                        + xmx[i-1][XME] * hmm->xt[XTE][MOVE]
                                                        * hmm->xt[XTC][LOOP]);

      /* Scale this row */
      /*scl[i] = xmx[i][XMN] + xmx[i][XMJ] + xmx[i][XMC] + mmx[i][hmm->M];
      for (k = 1; k < hmm->M; k++) scl[i] += mmx[i][k] + imx[i][k];*/
      scl[i] = 0.;
      if (xmx[i][XMN] > scl[i]) scl[i] = xmx[i][XMN];
      if (xmx[i][XMC] > scl[i]) scl[i] = xmx[i][XMC];
      if (xmx[i][XMB] > scl[i]) scl[i] = xmx[i][XMB];
      if (xmx[i][XME] > scl[i]) scl[i] = xmx[i][XME];
      if (xmx[i][XMJ] > scl[i]) scl[i] = xmx[i][XMJ];
      if (mmx[i][hmm->M] > scl[i]) scl[i] = mmx[i][hmm->M];
      for (k = 1; k < hmm->M; k++)
        {
          if (mmx[i][k] > scl[i]) scl[i] = mmx[i][k];
          if (imx[i][k] > scl[i]) scl[i] = imx[i][k];
          if (dmx[i][k] > scl[i]) scl[i] = dmx[i][k];
        }

      scl[i] = 1.0 / scl[i];
      /*if (!(scl[i] >= 0.))
        {
          printf("\nscl[%d] = %f, N J C = %f %f %f\n", i, scl[i], xmx[i][XMN], xmx[i][XMJ], xmx[i][XMC]);
          printf("mmx: "); for (k = 1; k <= hmm->M; k++) printf(" %f", mmx[i][k]);
          printf("\nimx: "); for (k = 1; k < hmm->M; k++) printf(" %f", imx[i][k]);
          printf("\n");
          printf("hmm->xt[XTN][MOVE]=%f\n", hmm->xt[XTN][MOVE]);
          printf("xmx[*][XMB]=[");
          int i1;
          for (i1 = 1; i1 <= i; i1++) printf("%f ", xmx[i][XMB]);
          printf("]\n");
          fflush(stdout);
        }*/
      //assert(scl[i] >= 0.);
      if (!(scl[i] >= 0.)) {printf("ASSERTION FAILED: scl[i] >= 0.\n");}

      xmx[i][XMN] *= scl[i]; xmx[i][XME] *= scl[i]; xmx[i][XMJ] *= scl[i];
      xmx[i][XMB] *= scl[i]; xmx[i][XMC] *= scl[i]; mmx[i][hmm->M] *= scl[i];
      for (k = 1; k < hmm->M; k++)
        {
          mmx[i][k] *= scl[i]; dmx[i][k] *= scl[i]; imx[i][k] *= scl[i];
        }
    }
  /* HMMER 1.8 Forward have one more row like mmx[L+1][k], but unlike Backward,
   * this row is NOT required in Baum-Welch reestimation ForbackCount().
   */

  sclprob = xmx[L][XMC] * hmm->xt[XTC][MOVE]
          + xmx[L][XME] * hmm->xt[XTE][MOVE] * hmm->xt[XTC][MOVE];

  if (ret_mx != NULL) *ret_mx = mx; else FreeMleDpMat(mx);

  return sclprob;           /* the total Forward score. */
}

/* Purpose:  The Backward dynamic programming algorithm.
 *           Note that the last state is restricted to be the terminal state.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_mx - RETURN: dp matrix; pass NULL if it's not wanted
 *
 * Return:   Pr(Seq|Model).
 */
float
MleBackward(unsigned char *dsq, int L, struct plan7_s *hmm, float *scl,
            struct mledpmat_s **ret_mx)
{
  struct mledpmat_s *mx;
  float **xmx, **mmx, **imx, **dmx;
  float sclprob;
  int   i,k;
  float tec = hmm->xt[XTE][MOVE] * hmm->xt[XTC][LOOP];
  float tej = hmm->xt[XTE][LOOP] * hmm->xt[XTJ][LOOP];
  float teb = hmm->xt[XTE][LOOP] * hmm->xt[XTJ][MOVE];

  /* Allocate a DP matrix with 0..L rows, 0..M-1 columns. */
  mx = CreateMleDpMat(L+1, hmm->M, 0, 0);
  xmx = mx->xmx; mmx = mx->mmx; imx = mx->imx; dmx = mx->dmx;

  /* Initialization of the L+1 and L row. Only mmx and imx need to have a L+1 row.
   * Note that xmx[i][stS] = xmx[i][stN] by definition for all i,
   *    so stS need not be calculated in backward DP matrices.
   */
  xmx[L+1][XMC] = xmx[L+1][XME] = 1.;            /* No difference in ForbackCount */
  xmx[L+1][XMJ] = xmx[L+1][XMB] = xmx[L+1][XMN] = 0.;
  for (k = hmm->M; k >= 1; k--) mmx[L+1][k] = dmx[L+1][k] = imx[L+1][k] = 0.;

  xmx[L][XMC] = hmm->xt[XTC][MOVE];              /* C<-T                          */
  xmx[L][XME] = xmx[L][XMC] * hmm->xt[XTE][MOVE];/* E<-C, no C-tail               */
  xmx[L][XMJ] = xmx[L][XMB] = xmx[L][XMN] = 0.;  /* need seq to get out from here */
  for (k = hmm->M; k >= 1; k--) {
    mmx[L][k] = xmx[L][XME] * hmm->end[k];       /* M<-E */
    imx[L][k] = dmx[L][k] = 0.;                  /* need seq to get out from here */
  }

  xmx[L][XMN] *= scl[L]; xmx[L][XME] *= scl[L]; xmx[L][XMJ] *= scl[L];
  xmx[L][XMB] *= scl[L]; xmx[L][XMC] *= scl[L]; mmx[L][hmm->M] *= scl[L];
  for (k = 1; k < hmm->M; k++)
    {
      mmx[L][k] *= scl[L]; dmx[L][k] *= scl[L]; imx[L][k] *= scl[L];
    }

  /* Recursion. Done as a pull.
   * Note slightly wasteful boundary conditions:
   *    M_M precalculated, D_M set to -INFTY,
   *    D_1 wastefully calculated.
   * Scores for transitions to D_M also have to be hacked to -INFTY,
   * as Plan7Logoddsify does not do this for us (I think? - ihh).
   */
  assert(hmm->t[hmm->M-1][TDD] < 1e-20);  /* Assert no D_M state */
  assert(hmm->t[hmm->M-1][TMD] < 1e-20);  /* Assert no D_M state */
  for (i = L-1; i >= 0; i--)
    {
      /* Do the special states first.
       * remember, C, N and J emissions are zero score by definition
       */
      xmx[i][XMC] = xmx[i+1][XMC] * hmm->xt[XTC][LOOP] * hmm->null[dsq[i+1]];
      xmx[i][XMB] = 0.;
      for (k = hmm->M; k >= 1; k--)
        xmx[i][XMB] += mmx[i+1][k] * hmm->begin[k] * hmm->mat[k][dsq[i+1]];
      xmx[i][XMJ] = xmx[i  ][XMB] * hmm->xt[XTJ][MOVE]
                  + xmx[i+1][XMJ] * hmm->xt[XTJ][LOOP] * hmm->null[dsq[i+1]];
      xmx[i][XME] = xmx[i+1][XMC] * tec                * hmm->null[dsq[i+1]]
                  + xmx[i+1][XMJ] * tej                * hmm->null[dsq[i+1]]
                  + xmx[i  ][XMB] * teb;
      xmx[i][XMN] = xmx[i  ][XMB] * hmm->xt[XTN][MOVE]
                  + xmx[i+1][XMN] * hmm->xt[XTN][LOOP] * hmm->null[dsq[i+1]];

      /* Now the main states. Note the boundary conditions at M.
       */
      if (i>0) {
        mmx[i][hmm->M] = xmx[i][XME] * hmm->end[hmm->M];
        dmx[i][hmm->M] = imx[i][hmm->M] = 0.; /* imx should not matter */
        for (k = hmm->M-1; k >= 1; k--)
          {
            imx[i][k]  = imx[i+1][k  ] * hmm->t[k][TII] * hmm->ins[k  ][dsq[i+1]]
                       + mmx[i+1][k+1] * hmm->t[k][TIM] * hmm->mat[k+1][dsq[i+1]];
            dmx[i][k]  = dmx[i  ][k+1] * hmm->t[k][TDD]
                       + mmx[i+1][k+1] * hmm->t[k][TDM] * hmm->mat[k+1][dsq[i+1]];
            mmx[i][k]  = xmx[i  ][XME] * hmm->end[k]
                       + mmx[i+1][k+1] * hmm->t[k][TMM] * hmm->mat[k+1][dsq[i+1]]
                       + imx[i+1][k  ] * hmm->t[k][TMI] * hmm->ins[k  ][dsq[i+1]]
                       + dmx[i  ][k+1] * hmm->t[k][TMD];
          }
      }

      /* Scale the current row. */
      xmx[i][XMN] *= scl[i]; xmx[i][XME] *= scl[i]; xmx[i][XMJ] *= scl[i];
      xmx[i][XMB] *= scl[i]; xmx[i][XMC] *= scl[i]; mmx[i][hmm->M] *= scl[i];
      for (k = 1; k < hmm->M; k++)
        {
          mmx[i][k] *= scl[i]; dmx[i][k] *= scl[i]; imx[i][k] *= scl[i];
        }
    }

  sclprob = xmx[0][XMN];

  if (ret_mx != NULL) *ret_mx = mx; else FreeMleDpMat(mx);

  return sclprob;           /* the total Backward score. */
}

/* Purpose:  Count probabilities of state transitions and symbol emissions from a
 *           forward-backward calculation on a single sequence. Results are
 *           stored in a counts-based HMM which will be normalized later.
 *           The full Baum-Welch equivalent to TraceCount() for Viterbi.
 *           Closer to HMMER 1.8.5, since scaling instead of log-odds is taken
 *           to prevent underflow. Log-odds may not converge and log-in-integer
 *           results in integer underflow and instability.
 *
 * Notes:    The boundaries of forward and backward matrices are used,
 *           0 = bck->mmx[L+1][2..M] = bck->imx[L+1][1..M-1] (= bck->dmx[L+1][1..M-1])
 *           bck->dmx[1..L][M] = 0
 *
 * Args:     hmm   - model. Probability form.
 *           prior - Used only for alphabet information.
 *           dsq   - sequence, 1..L
 *           L     - length of sequence
 *           score - scaled probability of the sequence
 *           weight- external sequence weight
 *           fwd   - forward matrix produced by Forward()
 *           bck   - backward matrix produced by Backward()
 *           scl   - scale factors used by forward and backward
 *           count - counts-based HMM to accumulate counts in.
 *                   Caller is responsible for allocating this structure
 *                   and initializing it properly.
 *
 * Return:   (void)
 *           Some counts are added to the count-based HMM.
 */
void
MleForbackCount(struct plan7_s *hmm, unsigned char *dsq, int L, float score,
                float weight, struct mledpmat_s *fwd, struct mledpmat_s *bck,
                struct plan7_s *count)
{
  int first = 1;
  int i, k, ts;
  float tmp_m1, tmp_d1, tmp_i0, t_m1, t_d1, t_i0, mat, ins;
  float prob;         /* overall scaled probability of the seq | model */
  float wt;
  float *scl = fwd->scl;

  /* "prob" is the total P(seq | model). We have to normalize by this.
   * We want to know, how much of the total alignment probability
   * flows through a particular state or state transition?
   */
  prob = score / weight;
  if (prob == 0.0) {
    printf("Warning: scaled probability is still too small: %e\n", prob);
    return;      /* silent! we have precision problems. */
  }
  /* assert(prob > 0.0); not true in MMIE training */
  for (i = 1; i <= L; i++) bck->dmx[i][1] = 0.; /* FIXME */
  for (i = 1; i <= L; i++)
    {
      bck->dmx[i][1] = 0.;
      for (k = 1; k < hmm->M; k++)
        {
          if (i < L) {mat = hmm->mat[k+1][dsq[i+1]]; ins = hmm->ins[k][dsq[i+1]];}
          else {mat = 0.; ins = 0.;}

          tmp_d1 = fwd->dmx[i][k] * hmm->t[k][TDD] * bck->dmx[i][k+1] / scl[i];
          tmp_m1 = fwd->dmx[i][k] * hmm->t[k][TDM] * bck->mmx[i+1][k+1] * mat;
          count->t[k][TDD] += tmp_d1 / prob; /* t[1][TDD]=0 later; t[M-1][TDD]=0 by bckd[1:L][M]=0 */
          count->t[k][TDM] += tmp_m1 / prob; /* t[1][TDD]=0 later; bckm[L+1][2:M]=0 */

          tmp_d1 = fwd->mmx[i][k] * hmm->t[k][TMD] * bck->dmx[i][k+1] / scl[i];
          tmp_m1 = fwd->mmx[i][k] * hmm->t[k][TMM] * bck->mmx[i+1][k+1] * mat;
          tmp_i0 = fwd->mmx[i][k] * hmm->t[k][TMI] * bck->imx[i+1][k] * ins;
          count->t[k][TMD] += tmp_d1 / prob; /* t[M-1][TMD]=0 by bckd[1:L][M]=0 */
          count->t[k][TMM] += tmp_m1 / prob; /* bckm[L+1][2:M]=0 */
          count->t[k][TMI] += tmp_i0 / prob; /* bcki[L+1][1:M-1]=0 */
          wt = (tmp_m1 + tmp_i0 + tmp_d1) / prob;
          P7CountSymbol(count->mat[k], dsq[i], wt);

          tmp_m1 = fwd->imx[i][k] * hmm->t[k][TIM] * bck->mmx[i+1][k+1] * mat;
          tmp_i0 = fwd->imx[i][k] * hmm->t[k][TII] * bck->imx[i+1][k] * ins;
          count->t[k][TIM] += tmp_m1 / prob; /* bckm[L+1][2:M]=0 */
          count->t[k][TII] += tmp_i0 / prob; /* bcki[L+1][1:M-1]=0 */
          wt = (tmp_m1 + tmp_i0) / prob;
          P7CountSymbol(count->ins[k], dsq[i], wt);
        }
      wt = fwd->mmx[i][hmm->M] * bck->mmx[i][hmm->M] / (scl[i] * prob);
      P7CountSymbol(count->mat[hmm->M], dsq[i], wt);
    }
}

/* Purpose:  Given a forward matrix and the scalefactors,
 *           return the log likeihood log P(S | M) in base 2 (bits)
 */
double
MleFwdScore(struct mledpmat_s *fwd, int L, int M, double sclprob)
{
  int    i;
  double score = sreLOG2(sclprob);
  for (i = 1; i <= L; i++)
    score -= sreLOG2(fwd->scl[i]); /* Note that scales are reversed */
  return score;
}

static void
ForBackNaN(unsigned char *dsq1, SQINFO *sqinfo1, struct plan7_s *hmm,
           struct mledpmat_s *fwd, struct mledpmat_s *bck, float t10,
           float sclprob)
{
  int i;
  printf("Sequence %s: t10=%.2f, sclprob=%f\nscl: ",sqinfo1->name,t10,sclprob);
  for (i = 1; i < sqinfo1->len; i++) printf("%.2f ",fwd->scl[i]);
  printf("\nFWD:\n"); PrintMleDpMat(fwd, sqinfo1->len, hmm->M);
  printf("BCK:\n");   PrintMleDpMat(bck, sqinfo1->len, hmm->M);
  fflush(stdout);
  FILE *seqfp = fopen("assert.fa", "w");
  char *seq = DedigitizeSequence(dsq1, sqinfo1->len);
  WriteSeq(seqfp, SQFILE_FASTA, seq, sqinfo1);
  fclose(seqfp);
  free(seq);
  SaveHMM(hmm, "assert.hmm");
}

/* Count the sufficient statistics (alpha, beta, gamma matrix)
 * and update a count-based HMM structure.
 * This covers the E-step and part of the M-step.
 */
float
TrainCountSuffStat(int nseqs, unsigned char **dsq, SQINFO *sqinfo, float *wt,
                   struct plan7_s *hmm, struct plan7_s *newhmm)
{
  int idx;
  float sclprob = 0., loglik = 0.;
  struct mledpmat_s *fwd; //dpmatrix_s *fwd;
  struct mledpmat_s *bck; //dpmatrix_s *bck;
  for (idx = 0; idx < nseqs; idx++)
    {
      sclprob = MleForward(dsq[idx], sqinfo[idx].len, hmm, &fwd);
      //printf("FWD:\n"); WarnMleDpMat(fwd, sqinfo[idx].len, hmm->M);
      MleBackward(dsq[idx], sqinfo[idx].len, hmm, fwd->scl, &bck);
      //printf("BCK:\n"); WarnMleDpMat(bck, sqinfo[idx].len, hmm->M);
      loglik += MleFwdScore(fwd, sqinfo[idx].len, hmm->M, sclprob);

      if (NULL != newhmm) /* if NOT the last run (calculating log likelihood) */
        MleForbackCount(hmm, dsq[idx], sqinfo[idx].len, sclprob, wt[idx],
                        fwd, bck, newhmm);
      if (NULL != newhmm && !(newhmm->t[1][0]>=0.)) /* Infinity; terminate */
        {
          ForBackNaN(dsq[idx],sqinfo+idx,hmm,fwd,bck,newhmm->t[1][0],sclprob);
          FreeMleDpMat(fwd); FreeMleDpMat(bck); return 0.0;
        }
      FreeMleDpMat(fwd); FreeMleDpMat(bck);
    }
  return loglik / nseqs;
}

void
TrainLimitIns(struct plan7_s *newhmm, float inslimit)
{
  int k;
  float tim_min = 1.0 / inslimit;
  for (k = 1; k < newhmm->M; k++)
    if (newhmm->t[k][TIM] < tim_min)
      {
        newhmm->t[k][TIM] = tim_min; newhmm->t[k][TII] = 1 - tim_min;
      }
}

/* Function: P7Distance()
 *
 * Purpose:  Test two models for how different they are, using
 *           a simple squared difference measure on all homologous
 *           parameters. They must have the same architecture:
 *           i.e. check that newhmm->M == oldhmm->M before calling.
 *
 * Args:     newhmm  - new HMM, probability form
 *           oldhmm  - old HMM, probability form
 *
 * Return:   distance.
 */
float
P7Distance(struct plan7_s *newhmm, struct plan7_s *oldhmm)
{
  int    k, x, ts;
  float  distance = 0.0;

  for (k = 1; k < newhmm->M; k++) /* main model transitions */
    for (ts = 0; ts < 7; ts++)
      if (k != newhmm->M - 1 || (ts != TDD && ts != TMD))
        distance += SQR( 100. * (newhmm->t[k][ts] - oldhmm->t[k][ts]) );
  for (k = 1; k <= newhmm->M; k++) /* match emissions */
    for (x = 0; x < Alphabet_size; x++)
      distance += SQR( 100. * (newhmm->mat[k][x] - oldhmm->mat[k][x]) );
  for (k = 1; k < newhmm->M; k++) /* insert emissions  */
    for (x = 0; x < Alphabet_size; x++)
      distance += SQR( 100. * (newhmm->ins[k][x] - oldhmm->ins[k][x]) );

  distance = sqrt(distance) / newhmm->M;
  return distance;
}

/* In generative training, optimize a HMM based on a MAP criteria,
 * by Baum-Welch or EM algorithm.
 */
struct plan7_s *
TrainOptimize(struct plan7_s *hmm, int nseqs, unsigned char **dsq,
              SQINFO *sqinfo, int setlength, struct p7prior_s *prior,
              int maxiter, float inslimit, int verbose, float *ret_score)
{
  int    idx, iter;
  float  *wt;                   /* array of sequence weights */
  struct plan7_s *newhmm;       /* re-estimated hmm */
  float  score;
  float  hmmdistance;           /* measure of parameter change between two models */
  double distance_threshold = 0.001;/* convergence criterion */

  /* Set up per-sequence weights. */
  wt = (float *) MallocOrDie(nseqs * sizeof(float));
  for (idx = 0; idx < nseqs; idx++) wt[idx] = 1.0;

  if (verbose >= 2)
    printf("Iteration      length      change      score    \n---------      ------      ------      -----    \n");

  for (iter = 1; iter <= maxiter && !killnow; iter++)
    {
      if (verbose >= 4) PrintHMM(hmm);
      /* E-step: collect full forward-backward matrices (sufficient stat) */
      if ((newhmm = AllocPlan7(hmm->M)) == NULL) /* counts-based HMM */
        Die("Failed to allocate for a new HMM\n");
      ZeroPlan7(newhmm);
      Plan7SetNullModel(newhmm, hmm->null, hmm->p1);
      //P7UniformNullModel(hmm); P7Logoddsify(hmm, 0);
      score = TrainCountSuffStat(nseqs, dsq, sqinfo, wt, hmm, newhmm);
      if (score == 0.0) /* if overflow, break without freeing hmm */
        {
          FreePlan7(newhmm); newhmm = hmm;
          printf("Terminate due to overflow\n"); break;
        }

      /* M-step: normalize an existing counts-based model (parameters) */
      P7PriorifyHMM(newhmm, prior);
      P7RetractWing(newhmm);
      Plan7Renormalize(newhmm);
      TrainLimitIns(newhmm, inslimit);
      Plan7GlobalConfig(newhmm);

      if (newhmm->M != hmm->M) hmmdistance = FLT_MAX;
      else hmmdistance = P7Distance(newhmm, hmm);

      if (verbose >= 2)
        {
          printf(" %4d ::    ", iter);            /* what iteration               */
          printf("   %6d   ", newhmm->M);         /* new model length             */
          if (hmm->M == newhmm->M)
            printf("%9.4f   ", hmmdistance);      /* change from old model        */
          else printf("     **     ");            /* (architecture changed)       */
          printf("%f\n", score);                  /* average score of old model   */
        }
      else if (verbose) printf("%d:%.2f(%.1f) ", iter, score, hmmdistance);

      if (verbose && !(hmmdistance >= 0.)) PrintHMM(newhmm);

      /* Switch new model to current model */
      FreePlan7(hmm);
      hmm = newhmm;

      /* Check for convergence. */
      if (hmmdistance < distance_threshold) break;
    }
  *ret_score = TrainCountSuffStat(nseqs, dsq, sqinfo, wt, hmm, NULL);
  if (verbose >= 2) PrintHMM(hmm);
  else if (verbose) printf("\n");
  fflush(stdout);
  free(wt);
  return hmm;
}

/* Initialize a count-based profile HMM based on randomly selected segments. */
static void
P7InitRandomPos(struct plan7_s *hmm, int nseqs, unsigned char **dsqs, SQINFO *sqinfo)
{
  int idx, k, pos;
  for (idx = 0; idx < nseqs; idx++)
    {
      pos = (int) (sre_random() * (sqinfo[idx].len - hmm->M));
      for (k = 1; k <= hmm->M; k++)
        P7CountSymbol(hmm->mat[k], dsqs[idx][pos+k-1], 1.0);
    }
}

/* Initialize a count-based profile HMM based on random amino acids. */
static void
P7InitRandomAA(struct plan7_s *hmm)
{
  int k;
  for (k = 1; k <= hmm->M; k++)
    P7CountSymbol(hmm->mat[k], (int) (sre_random() * Alphabet_size), 10.);
}

/* Purpose:  Initialize a profile HMM, using just the prior,
 *           or randomly selected segments, or random amino acids.
 *
 * Args:     nseqs  - number of sequences for randomly selected segments,
 *                    0 for prior, -1 for random amino acids.
 *
 * Return:   Allocated HMM.
 */
struct plan7_s *
P7Init(int setlength, struct p7prior_s *prior, float *randomseq, float p1,
       int nseqs, unsigned char **dsqs, SQINFO *sqinfo)
{
  struct plan7_s *hmm = AllocPlan7(setlength);
  Plan7SetNullModel(hmm, randomseq, p1);
  ZeroPlan7(hmm);
  if (nseqs > 0) P7InitRandomPos(hmm, nseqs, dsqs, sqinfo);
  else if (nseqs == -1) P7InitRandomAA(hmm);
  P7PriorifyHMM(hmm, prior);
  P7RetractWing(hmm);
  Plan7GlobalConfig(hmm);
  return hmm;
}

/* Perform generative training of a HMM by Baum-Welch or EM algorithm.
 * Initial parameters are estimated by random positions in the sequences
 * (init by average canceled due to empirical tests).
 */
struct plan7_s *
GenTrain(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, int setlength,
         int maxiter, int ninit, float inslimit, float avglen, int verbose,
         float *ret_score)
{
  struct plan7_s *hmm = NULL;
  struct plan7_s *tophmm = NULL;
  struct p7prior_s *prior = P7DefaultPrior();
  float score;                  /* log2 likelihood of each HMM          */
  float topscore;               /* top log-likelihood                   */
  int   iinit;                  /* init index                           */
  int   topinit;                /* the top init index (zero-start best?)*/
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */

  P7DefaultNullModel(randomseq, &p1);
  p1 = avglen / (avglen + 1.0);

  for (iinit = 0; iinit < ninit; iinit++)
    {
      if (verbose) printf("Init set %d/%d for generative training...\n", iinit, ninit);
      hmm = P7Init(setlength, prior, randomseq, p1, nseqs, dsqs, sqinfo);
      hmm = TrainOptimize(hmm, nseqs, dsqs, sqinfo, setlength, prior, maxiter,
                          inslimit, verbose, &score);//init hmm freed
      if (0 == iinit || score > topscore)
        {
          if (NULL != tophmm) FreePlan7(tophmm);
          tophmm = hmm; topscore = score; topinit = iinit;
        }
      else FreePlan7(hmm);
    }

  if (verbose)
    {
      printf("Best profile HMM (init number %d out of %d):\n",
              topinit + 1, ninit);
      PrintHMM(tophmm);
    }
  P7FreePrior(prior);
  *ret_score = topscore;
  return tophmm;
}

/***********************************************
 * Classify
 ***********************************************/

/* Purpose:  Return log likelihood of sequences based on given HMMs
 *           (using the forward algorithm) for classification.
 *           Caller typically use argmaxf() to find the maximum.
 *
 * Args:     nhmm  - number of HMMs
 *           hmms  - array of HMMs, 0..nhmm
 *           seq   - sequence, 0..L-1
 *
 * Return:   Allocated array of log likelihood.
 */
float *
Classify(int nhmm, struct plan7_s **hmms, unsigned char *dsq, int L)
{
  int ihmm;
  float *scores = (float *) MallocOrDie(nhmm * sizeof(float));
  for (ihmm = 0; ihmm < nhmm; ihmm++)
    scores[ihmm] = P7Forward(dsq, L, hmms[ihmm], NULL);
  return scores;
}

/* Purpose:  Evaluates trained HMMs on testing sequences and true answers.
 *
 * Args:     nhmm  - number of HMMs
 *           hmms  - array of HMMs, 0..nhmm
 *           nseqs - number of sequences
 *           seqs  - sequences
 *           cates - true categories
 *
 * Return:   (float) accuracy
 */
float
EvaluateSeqs(int ncate, struct plan7_s **hmms, int *catenseq,
             unsigned char ***catedsqs, SQINFO **catesqinfo)
{
  int ncorrect = 0, nseqs = 0, k, idx;
  for (k = 0; k < ncate; k++)
    {
      nseqs += catenseq[k];
      for (idx = 0; idx < catenseq[k]; idx++)
        {
          float *scores = Classify(ncate, hmms, catedsqs[k][idx],
                                   catesqinfo[k][idx].len);
          if (argmaxf(ncate, scores) == k) ncorrect++;
          free(scores);
        }
    }
  return ((float) ncorrect) / nseqs;
}

/***********************************************
 * Predict motif locations by Viterbi
 ***********************************************/

/* Free the content of matlocs_s structure, but NOT the structure itself.
 * The caller need to call free(matlocs) afterwards.
 * This make it easier for array of strucutres (not array of pointers).
 */
void
FreeMatlocs(struct matlocs_s *matlocs)
{
  assert(NULL != matlocs->begin);
  assert(NULL != matlocs->end);
  free(matlocs->begin);
  free(matlocs->end);
}

/* Purpose:  Allocate and construct match-location arrays from Viterbi trace.
 *
 * Note:     Structure matlocs should already be allocated.
 *           This design is due to matlocs is usually an array
 *           corresponding to all sequences.
 */
void
Trace2Loc(struct matlocs_s *matlocs, struct p7trace_s *tr, int M)
{
  int ibegin = 0, iend = 0;     /* two index for assert */
  int tpos;
  assert(NULL != matlocs);
  //assert(NULL == matlocs->begin); assert(NULL == matlocs->end);
  matlocs->n = 0;
  for (tpos = 0; tpos < tr->tlen; tpos++)
    if (STM == tr->statetype[tpos] && 1 == tr->nodeidx[tpos]) (matlocs->n)++;
  matlocs->begin = (int *) MallocOrDie(sizeof(int) * matlocs->n);
  matlocs->end   = (int *) MallocOrDie(sizeof(int) * matlocs->n);
  for (tpos = 0; tpos < tr->tlen; tpos++)
    {
      if (STM == tr->statetype[tpos] && 1 == tr->nodeidx[tpos])
        {
          assert(ibegin == iend);
          matlocs->begin[ibegin++] = tr->pos[tpos];
        }
      else if (STM == tr->statetype[tpos] && M == tr->nodeidx[tpos])
        {
          assert(ibegin == iend+1);
          matlocs->end[iend++] = tr->pos[tpos];
        }
    }
  assert(ibegin == iend);
}

/* Perform Viterbi algorithm and save locations.
 * Note that matlocs is an allocated array of size nseqs,
 * but individual elements are allocated and need to be freed.
 */
void
ScanMatches(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, struct plan7_s *hmm,
            struct matlocs_s *matlocs, struct p7trace_s **tr_arr)
{
  int idx;
  struct p7trace_s *tr;
  struct dpmatrix_s *mx;

  P7Logoddsify(hmm, TRUE); /* TRUE for viterbi mode */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0);
  for (idx = 0; idx < nseqs; idx++)
    {
      if (P7ViterbiSpaceOK(sqinfo[idx].len, hmm->M, mx))
        (void) P7Viterbi(dsqs[idx], sqinfo[idx].len, hmm, mx, &tr);
      else
        (void) P7SmallViterbi(dsqs[idx], sqinfo[idx].len, hmm, mx, &tr);
      Trace2Loc(matlocs + idx, tr, hmm->M);
      if (NULL != tr_arr) tr_arr[idx] = tr; else P7FreeTrace(tr);
    }
  FreePlan7Matrix(mx);
}

/***********************************************
 * Predict motif locations by posterior decoding
 ***********************************************/

void
PosteriorBE(unsigned char *dsq, int L, struct plan7_s *hmm, float *begin, float *end)
{
  int i;
  float sclprob = 0., loglik = 0.;
  struct mledpmat_s *fwd; //dpmatrix_s *fwd;
  struct mledpmat_s *bck; //dpmatrix_s *bck;

  sclprob = MleForward(dsq, L, hmm, &fwd);
  MleBackward(dsq, L    , hmm, fwd->scl, &bck);
  for (i = 1; i <= L; i++)
    {
      begin[i] = fwd->xmx[i][XMB] * bck->xmx[i][XMB] / (fwd->scl[i] * sclprob);
      end[  i] = fwd->xmx[i][XME] * bck->xmx[i][XME] / (fwd->scl[i] * sclprob);
    }
  FreeMleDpMat(fwd); FreeMleDpMat(bck);
}
